@implementation RemoteInvocationServer {
 - init {} {
     $super init
     set logfile RemoteInvocationServer.log
     set pool [AutoReleasePool new]
     set useAllowedCommands 0
     set allowedCommands(normal) {}
     set allowedCommands(regexp) {}
   }

 - handleRequest:fromSocket:address:andPort: \
     {aRequest aSocket anAddress aPort} {
      switch -regexp -- [string tolower $aRequest] {
        {^@rmi } {
          set cmd [join [lrange [split $aRequest] 1 end]]
          set err [$self executeCommand: $cmd \
                             fromSocket: $aSocket \
                                address: $anAddress \
                                andPort: $aPort]
          $self send: $err to: $aSocket
        }
        default {
          $self send: "unknown request" to: $aSocket
          $self closeSocket: $aSocket
        }
      }
    }

  - executeCommand:fromSocket:address:andPort: \
      {aCommand aSocket anAddress aPort} {
      set msg "$anAddress:$aPort -> "
      append msg $aCommand
      @log $logfile $msg
      if {$useAllowedCommands} {
        puts stderr "using allowed commands"
        puts stderr "\"$allowedCommands(normal)\" contains: \"$aCommand\" ?"
        if {[allowedCommands(normal) contains: $aCommand]} {
          puts stderr "allowedCommands(normal) contains: $aCommand"
          catch {eval $aCommand} result
        } elseif [string length $allowedCommands(regexp)] {
          puts stderr "allowedCommands(normal) dosn't contain: $aCommand"
          foreach cmd $allowedCommands(regexp) {
            puts stderr "cmd:$cmd / regexp $cmd $aCommand ?"
            if [regexp $cmd $aCommand] {
              puts stderr "yes"
              catch {eval $aCommand} result
              return $result
            } else {
              puts stderr "no"
            }
          }
        }
      } else {
        set result "command not allowed"
      }
      return $result
    }

  - dealloc {} {
      $pool release
      return [$super dealloc]
    }

  - useAllowedCommands {} { set useAllowedCommands 1 }
  - dontUseAllowedCommands {} { set useAllowedCommands 0 }
  - doesUseAllowedCommands {} { return $useAllowedCommands } 

  - allowedCommands {} { return $allowedCommands(normal) }
  - allowedCommands: commands { set allowedCommands(normal) $commands }

  - regexpAllowedCommands {} { return $allowedCommands(regexp) }
  - regexpAllowedCommands: commands { set allowedCommands(regexp) $commands }
}
