@implementation Tree {
  - init {} {
      $super init 
      $self root: /
      $self separator: /
    }

  - root: rootName {
      #- changing existing child info
      set root $rootName
    }

  - root {} {
      return $root
    }

  - separator: aCharacter {
      #- changing existing info
      foreach array {nodes childs properties} {
        foreach field [array names $array] {
          if {[llength [split $field $separator]] > 1} {
            set ${array}([join [split $field $separator] $aCharacter]) \
              [set ${array}($field)]
            unset ${array}($field)
          }
        }
      }
      set separator $aCharacter
    }

  - separator {} {
      return $separator
    }

  - add: aNode {
      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }
      #- already exist ?
      if {[$self contains: $aNode]} {
        return [error "([info level 0]) \"$aNode\" already exist"]
      }
     
      set nodeName [$self nodeNameOf: $aNode]
      set parent [$self parentOf: $aNode]

      #- checking parent existence
      if {"$parent" == "$root"} {
        lappend rootChilds $nodeName 
      } else {
        if {![$self contains: $parent]} {
          return [error "([info level 0]) \"$parent\" doesn't exist"]
        } else {
          lappend childs($parent) $nodeName
        }
      } 
      set nodes($aNode) $nodeName
    }

  - remove: aNode {
      if {"$aNode" == "$root"} { $self clear ; $self root: / ; return $self }

      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }

      #- already gone ?
      if {![$self contains: $aNode]} { 
        return [error "([info level 0]) \"$aNode\" doesn't exist"]
      }
      
      set nodeName [$self nodeNameOf: $aNode]
      set parent [$self parentOf: $aNode]

      #- removing node link from parent childs list
      if {"$parent" != "$root"} {
        if [string length $parent] {
           childs($parent) remove: $nodeName
        }
      } else {
        rootChilds remove: $nodeName
      }

      #- removing node
      unset nodes($aNode)

      #- removing childs
      if [info exists childs($aNode)] {
        foreach node $childs($aNode) {
          $self remove: [join [list $aNode $node] $separator]
        }
        unset childs($aNode)
      }

      #- removing properties
      foreach field [array names properties $aNode,*] {
        unset properties($field)
      }
    }

  - parentOf: aNode {
      if {"$aNode" == "$root"} { return }

      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }

      set parent [join [lreplace [split $aNode $separator] end end] $separator]
      if {![string length $parent]} {
        return $root
      } else {
        return $parent
      }
    } 

  - childsOf: aNode {
      if {"$aNode" == "$root"} {
        return $rootChilds
      }

      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }
      if {![$self contains: $aNode]} {
        return [error "([info level 0]) \"$aNode\" doesn't exist"]
      }
      set childsOfNode {}
      if [info exists childs($aNode)] {
        foreach node $childs($aNode) {
          lappend childsOfNode [join [list $aNode $node] $separator]
        }
      }
      return $childsOfNode
    }

  - allChildsOf: aNode {
      if {"$aNode" == "$root"} {
        return [array names nodes]
      }

      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }
      if [$self contains: $aNode] {
        if {"$aNode" != "$separator"} {
           return [array names nodes [join [list ${aNode} *] $separator]]
        } else {
          return [array names nodes ${aNode}?*]
        }
      }
      return
    }

  - contains: aNode {
      if {"$aNode" == "$root"} { return 1 }
      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }
      if [info exists nodes($aNode)] { return 1 } else { return 0 }
    }

  - loadFromFile: aFile {
      set f [open $aFile r]
      while {![eof $f]} {
        set line [string trim [gets $f]]
        if [string length $line] {
          $self add: $line
        }
      }
      close $f
    }

  - saveToFile: aFile {
      set f [open $aFile w]
      foreach node [lsort [$self allChildsOf: $root]] {
        puts $f $node
      }
      close $f
    }

  - clear {} {
      foreach array {nodes childs properties} {
        foreach field [array names $array] {
          unset ${array}($field)
        }
      }
      $self root: $root
    }

  - nodeNameOf: aNode {
      if {"$aNode" == "$separator" && "$aNode" == "$root"} {
        return $aNode
      }
      return [lindex [split $aNode $separator] end]
    }

  - levelOf: aNode {
      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }
      if {"$aNode" == "$root"} { return 0 }
      set nodeLevel [llength [split $aNode $separator]]
      incr nodeLevel -1
      return $nodeLevel
    }

  - property:of: {propertyName aNode} {
      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }
      if [info exists properties($aNode,$propertyName)] {
        return $properties($aNode,$propertyName)
      } else {
        return [error "([info level 0]) \"$aNode\" doesn't have any \"$propertyName\" property"]
      }
    }

  - property:of:to: {propertyName aNode aValue} {
      #- removing root name if in the path
      set aNode [string trimleft $aNode $separator]
      if {"[lindex [split $aNode $separator] 0]" == "$root"} {
        set aNode [join [lreplace [split $aNode $separator] 0 0] $separator]
      }
      set properties($aNode,$propertyName) $aValue
    }
}
