@implementation CanvasBox {
  - init {} {
      $super init
      $super hideBorder 
      set zoom 1.0

      set bg [[entry $path._e] cget -background]
      destroy $path._e

      set canvas [canvas $frame.canvas -width 1 -height 1 -bd 1 -relief sunken]
      $canvas configure -bg $bg

      set xscroll [scrollbar $frame.xscroll -orient horizontal \
        -command "$canvas xview"]
      set yscroll [scrollbar $frame.yscroll -orient vertical \
        -command "$canvas yview"]

      grid rowconfigure $frame 0 -weight 1
      grid columnconfigure $frame 0 -weight 1
      grid $canvas $yscroll -sticky snew -padx 1 -pady 1
      grid $xscroll -sticky snew -padx 1 -pady 1

      $canvas configure -xscrollcommand "$xscroll set" \
        -yscrollcommand "$yscroll set"

      $self enableEvent: Configure on: $canvas
      $self on: Configure do: {
        $canvas configure -scrollregion [$canvas bbox all]
      }

      $self enableEvent: {Button-4 Button-5} on: $canvas
      $self on: Button-4 do: {$canvas yview scroll -1 units}
      $self on: Button-5 do: {$canvas yview scroll +1 units}
      
      $self enableEvent: {
        Shift-ButtonPress-1 Shift-ButtonPress-3 
        Shift-ButtonRelease-1 Shift-ButtonRelease-3
      } on: $canvas

      $self on: Shift-ButtonPress-1 do: {$self startZoomingUp}
      $self on: Shift-ButtonPress-3 do: {$self startZoomingDown}
      $self on: Shift-ButtonRelease-1 do: {$self stopZooming}
      $self on: Shift-ButtonRelease-3 do: {$self stopZooming}
    }

  - canvas {} {
      return $canvas
    }

  - zoom: factor {
      if {[set f [expr 1.0 * $factor]] != 0.0} {
        $canvas scale all 0 0 $f $f
        $canvas configure -scrollregion [$canvas bbox all]
        set zoom $f
      }
    }

  - zoom {} {
      return $zoom
    }

  - startZoomingUp {} {
      $self stopZooming
      $self zoomUp
    }

  - zoomUp {} {
      $self zoom: 2
      set cmd "$self zoomUp"
      set zoomingId [after 200 $cmd]
    }

  - startZoomingDown {} {
      $self stopZooming
      $self zoomDown
    }

  - zoomDown {} {
      $self zoom: 0.5
      set cmd "$self zoomDown"
      set zoomingId [after 200 $cmd]
    }

  - stopZooming {} {
      if {[string length $zoomingId]} {
        after cancel $zoomingId
        set zoomingId {}
      }
    }

  - canvasBackground: aColor {
      $canvas configure -bg $aColor
      $frame configure -bg $aColor
    }

  - canvasBackground {} {
      return [$canvas cget -background]
    }
}
