#------------------------------------------------------------------------------
#         file: oo.tcl
#------------------------------------------------------------------------------
#      project: myObjectiveTcl
#  description: a pure tcl implementation of an objective-C like framework
#      version: 0.1
# release date: 01/01/2003
#      licence: BSD licence
#------------------------------------------------------------------------------
#       author: Frederic Colfs
#      contact: frederic.colfs@belgacom.net
#------------------------------------------------------------------------------
#
# procedures :
#
# -- objective-c oriented --
#
# @selector
# @args
# @interface
# @implementation
# @class
# @protocol
#
# -- runtime implementation --
#
# @maxInteger
# @nextId
# @newInstanceFromClass
# @sendMessageToObject
# @sendMessageToSuper
# @getCallerLevel
# @getAccessToClassVariables
# @getAccessToClassAndInstanceVariables
#
# -- dynamic loading --
#
# @path
# @proto
# @package
# @unload
# @reload
#
# -- utils --
#
# @date
# @time
# @now
# @log
# @debug
# @error
# @syntax
#
# -- new syntax --
#
# className message 
# list|array dump
# list each: { puts stderr "$each" }
# number !
# text encrypt
# text decrypt
# list sort
# list addIfNotYetIn: newElement
# fieldValueList sortByValue
# fieldValueList sortByField
# fieldValueList fields
# fieldValueList values
# fieldValueList set: fieldName to: newValue
# fieldValueList get: fieldName
# fieldValueList add: newfield newvalue
# fieldValueList removeField: fieldName
# list remove: value
# fieldValueList contains: fieldName
# list indexOf: value
# variableName += number
# -- variableName
# ++ variableName
# variableName = value
# string stdout
# string stderr
# string length
# list llength
#
# -- tk specific --
# 
# overrides default Tk scrollbar behaviour to keep relief unchanged:
#   tkScrollButtonDown
#   tkScrollButtonUp
#   tkScrollTopBottom
#   tkScrollButton2Down
# 
# widgets ressources
#
# @image
# @bitmapCopy
#
# images: 
#   down_arrow
#   up_arrow
#   left_arrow
#   right_arrow
#   return
#   readioOn
#   radioOff
#   empty16x16
#   swithOff
#   switchOn
#   empty9x9
#   empty
#   plus
#   minus
#   circle
#   white_circle
#   blue_circle
#   red_circle
#   green_circle
#   yellow_circle
#   orange_circle
#   black_circle
#   violet_circle
#   brown_circle
#   grey_circle
#   directory
#   file
#   bullet
#   plus_25x25
#   green_plus
#   minus_25x25
#   red_minus
#
# @eventData
# @removeBindingsOnDestroyEvent   
#
# -- doc generation --
#
# @makedoc (v1)
#
# @makedoc.new (v2)
# @superClassesOf
# @docClassPage
# @docPackagePage
# @docProtocolPage
# @docHtmlHead
# @docNavigationBar
#
#------------------------------------------------------------------------------

global @RUNTIME

set @RUNTIME(directory) [file dirname [info script]]

# -- debugging --
if {[info exist DEBUG]} {
  rename proc _proc
  _proc proc {name args body} {
    if {"$name" != "unknown"} {
      append newBody "puts stderr \"* \[info level 0\]\""
    }
    append newBody $body
    _proc $name $args $newBody
  }
}

# -- objective-c oriented --

#------------------------------------------------------------------------------
# extract selector from message
#------------------------------------------------------------------------------
proc @selector { {message {}} } {
  set selector {}
  if [string length [string trim $message { }]] {
    if {![regexp : $message]} {
      set selector [lindex $message 0]
    } else {
      foreach {field value} $message {
        if [regexp : $field] { append selector $field }
      }
    }
  }
  set selector
}

#------------------------------------------------------------------------------
# extract arguments from message
#------------------------------------------------------------------------------
proc @args { {message {}} } {
  set args {}
  if {![regexp : $message]} {
    set args [lrange $message 1 end]
  } elseif [string length [string trim $message { }]] {
    foreach {field value} $message {
      if {![regexp : $field]} { lappend args $field }
      lappend args $value
    }
  }
  set args
}

#------------------------------------------------------------------------------
# parser for the class interface
#------------------------------------------------------------------------------
proc @interface { {args {}} } {
  global @RUNTIME

  foreach v {
    class superclass protocols category
    class,variables class,selectors
    instance,variables instance,selectors
  } {
    set $v {}
  }

  #- splitting args into realArgs and code
  if [regexp {^([^\{]+)[\{](.+)[\}]$} $args -> realArgs code] {
    #- parsing realArgs to get class, superclass, protocols, category if any
    if [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*<([^>]+)>[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class superclass protocols category] {
      #-1- class + superclass + protocols + category
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*<([^>]+)>[ ]*$} \
        $realArgs -> class superclass protocols] {
      #-2- class + superclass + protocols
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class superclass category] {
      #-3- class + superclass + category
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*$} \
        $realArgs -> class superclass] {
      #-4- class + superclass
    } elseif [regexp {^([^ ]+)[ ]*<([^>]+)>[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class protocols category] {
      #-5- class + protocols + category
    } elseif [regexp {^([^ ]+)[ ]*<([^>]+)>[ ]*$} \
        $realArgs -> class protocols] {
      #-6- class + protocols
    } elseif [regexp {^([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class category] {
      #-7- class + category
    } elseif [regexp {^([^ ]+)[ ]*$} $realArgs -> class] {
      #-8- class
    } else {
      return [error "malformed interface"]
    }
  } elseif [regexp {^([^\{]*)[\{](.*)[\}]$} $args] {
    #- empty interface
  } else {
    return [error "malformed interface"]
  }

  set code [string trimright $code "\r\n "]

  #- removing comments from code
  set newcode ""
  foreach line [split $code \n] {
    if [regexp {^[ ]*[#](.*)$} $line] { continue }
    append newcode $line\n
  }
  set code $newcode

  #- parsing code
  if [regexp {[\{][\r\n ]*(.*)[\}](.*)} $code -> variables code] {
    set variables [string trimright $variables "\n\r "]
    set instance,variables $variables
    if {[string length $category] && [string length ${instance,variables}]} {
      return [error "class:$class category:$category \
        -> can't add instance variables with a category !"]
    }
  }

  foreach line [split $code \;] {
    set line [string trim $line "\r\n "]
    if {![string length $line]} { continue }
    if [regexp {^([+-])[ ]*([^$]+)$} $line -> sign message] {
      set selector [@selector $message]
      set args [@args $message]
      switch -- $sign {
        + {
           set type class
        }
        - {
          set type instance
        }
        default { }
      }
      lappend ${type},selectors $selector
    } else {
      set classVariables $line
      foreach v $classVariables {
        lappend class,variables $v
      }
    }
  }
  #----------------------------------
  #- creating interface into runtime
  #----------------------------------
  #- adding <class> to @RUNTIME(classes)
  #- setting or adding to the following
  #- @<class>(instance,variables)
  #- @<class>(instance,selectors)
  #- @<class>(instance,methodFor,<selector>)
  #- @<class>(class,variables)
  #- @<class>(class,selectors)
  #- @<class>(class,methodFor,<selector>)
  #- @<class>(superclass)
  #- @<class>(protocols)
  #- @<class>(categories)
  #- @<class>(class)

  upvar #0 @DOC DOC
  upvar #0 @${class} SELF
  foreach v {
    class superclass protocols categories
    class,variables class,selectors
    instance,variables instance,selectors
  } {
    if {![info exists SELF($v)]} {
      set SELF($v) {}
    }
  }

  set SELF(class) $class
  set SELF(superclass) $superclass

  if [string length $superclass] {
    if {![@class $superclass]} {
      return [error \
        "couldn't import \"$superclass\" superclass (for \"$class\" class)"]
    }
    upvar #0 @${superclass} SUPER
    foreach type {instance class} {
      foreach field {variables selectors} {
        foreach value $SUPER(${type},${field}) {
          if {"$field" == "variables"} {
            if {![SELF(${type},${field}) contains: $value]} {
              lappend SELF(${type},${field}) $value
            }
          } else {
            lappend SELF(${type},${field}) $value
          }
        }
      }
      foreach field [array names SUPER ${type},methodFor,*] {
        set SELF($field) $SUPER($field)
      }
    }
    foreach proto $SUPER(protocols) {
      if {![SELF(protocols) contains: $proto]} {
        lappend SELF(protocols) $proto
      }
    }
  }
  if [string length $category] {
    #- importing original class if it's a category
    if {![@class $class]} {
      return [error "couldn't import original $class definition"]
    }
    if {![SELF(categories) contains: $category]} {
      lappend SELF(categories) $category
    }

    #- doc
    set DOC(category,interface,$category) [info script]
  }

  foreach protocol [split $protocols ,] {
    set protocol [string trim $protocol]
    if {![SELF(protocols) contains: $protocol]} {
      lappend SELF(protocols) $protocol
    }
  }
  foreach type1 {class instance} {
    foreach type2 {variables selectors} {
      foreach value [set ${type1},${type2}] {
        if {![SELF($type1,$type2) contains: $value]} {
          lappend SELF($type1,$type2) $value
        }
      }
    }
  }
  #- creating class variables
  foreach var $SELF(class,variables) {
    if [regexp {^@([^$]+)$} $var -> arrayName] {
      global @${class}@${arrayName}
      array set @${class}@${arrayName} {}
    } else {
      set SELF($var) {}
    }
  }

  if {![@RUNTIME(classes) contains: $class]} {
    lappend @RUNTIME(classes) $class
  }

  #- for the doc
  if {![info exists DOC(class,interface,$class)]} {
    set DOC(class,interface,$class) [info script]
  }

  #- package
  if {![info exists SELF(package)]} {
    set dir [file split [info script]]
    if {[dir contains: interface]} {
      set dir [lreplace $dir [dir indexOf: interface] end]
    }
    set libDir [set @RUNTIME(directory)]
    if {[dir contains: $libDir]} {
      set dir [lreplace $dir [set i [dir indexOf: $libDir]] $i]
    }
    set dir [file join $dir]
    set dir [string trimleft $dir \\]
    set dir [string trimleft $dir /]
    set SELF(package) $dir
    if {![@RUNTIME(packages) contains: $dir]} {
      lappend @RUNTIME(packages) $dir
    }
  }

}

#------------------------------------------------------------------------------
# parser for the class implementation
#------------------------------------------------------------------------------
proc @implementation { {args {}} } {
  global @RUNTIME

  set category {}

  #- splitting args into real args and code
  if [regexp {^([^\{]+)[\{](.+)[\}]$} $args -> realArgs code] {
    #- parsing realArgs
    if [regexp {^([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
      $realArgs -> class category] {
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*$} \
        $realArgs -> class superclass] {
    } elseif [regexp {^([^ ]+)[ ]*$} $realArgs -> class] {
    }

    set code [string trimright $code "\r\n "]

    #- removing comments from code
    set newcode ""
    foreach line [split $code \n] {
      if [regexp {^[ ]*[#](.*)$} $line] { continue }
      append newcode $line\n
    }
    set code $newcode

    if [string length $category] {
      #- importing original class if it's a category
      if {![@class $class]} {
        return [error "couldn't import original $class definition"]
      }
    }

    upvar #0 @DOC DOC
    upvar #0 @${class} SELF
    if [string length $SELF(superclass)] {
      upvar #0 @$SELF(superclass) SUPER
    }


    #- parsing code and creating instance/class methods into runtime
    #- proc @<class>,instance,<selector>
    #- proc @<class>,class,<selector>

    foreach {char selector args body} $code {
      switch -- $char {
        - { 
          set type instance
        }
        + {
          set type class
        }
        default { return [error "unknown char:$char"] }
      }
      #- args
      if {![string length $args]} { set args "{args {}}" }
      set args [linsert $args 0 self]

      if {"$type" == "instance"} {
        append newbody {
          @getAccessToClassAndInstanceVariables
        }
      } else {
        append newbody {
          @getAccessToClassVariables
        }
      }
      append newbody $body
      set body $newbody
      unset newbody
      append body {
        set self
      }

      #- method
      proc @${class},$type,$selector $args $body

      #- adding method in class description if not yet in
      if {![SELF($type,selectors) contains: $selector]} {
        lappend SELF($type,selectors) $selector
      }
      set SELF(${type},methodFor,${selector}) \
        @${class},$type,$selector
    }

    #- check if all methods are implemented
    foreach type {class instance} {
      #- getting only selectors that are not inherited
      set notInheritedSelectors {}
      if [info exists SUPER] {
        foreach selector $SELF($type,selectors) {
          if {![SUPER($type,selectors) contains: $selector]} {
            lappend notInheritedSelectors $selector
          }
        }
      } else {
        set notInheritedSelectors $SELF($type,selectors)
      }
      foreach selector $notInheritedSelectors {
        if {![string length [info proc @${class},$type,$selector]]} {
          return [error "uncompleted implementation for \"$class\" class:\n\
            \"$selector\" selector is not implemented !\n"]
        }
      }
    }

    #- check if protocol methods are implemented
    if [string length $SELF(protocols)] {
      foreach protocol $SELF(protocols) {
        if {![@proto $protocol]} {
          return [error "unknown protocol $protocol for class $class"]
        } 
        upvar #0 @@${protocol} PROTOCOL
        foreach type {instance class} {
          foreach selector $PROTOCOL(${type},selectors) {
            if {![SELF(${type},selectors) contains: $selector]} {
              return [error \
                "$class class must conform to $protocol protocol and \
                \"$selector\" selector is not implemented!"]
            }
          }
        }
      }
    }

    #- creating class callback
    proc $class { {args {}} } {
      return [@sendMessageToObject [lindex [info level 0] 0] $args]
    }

    #- adding class name to runtime
    if {![@RUNTIME(classes) contains: $class]} {
      lappend @RUNTIME(classes) $class
    }

    #- doc
    if {![info exists DOC(class,implementation,$class)]} {
      set DOC(class,implementation,$class) [info script]
    }

    if [string length $category] {
      if {![SELF(categories) contains: $category]} {
        lappend SELF(categories) $category
      }
      set DOC(category,interface,$category) [info script]
    }

    #- package
    if {![info exists SELF(package)]} {
      set dir [file split [info script]]
      if {[dir contains: implementation]} {
        set dir [lreplace $dir [dir indexOf: implementation] end]
      }
      set libDir [set @RUNTIME(directory)]
      if {[dir contains: $libDir]} {
        set dir [lreplace $dir [set i [dir indexOf: $libDir]] $i]
      }
      set dir [file join $dir]
      set dir [string trimleft $dir \\]
      set dir [string trimleft $dir /]
      set dir [string trimleft $dir \\]
      set SELF(package) $dir
      if {![@RUNTIME(packages) contains: $dir]} {
        lappend @RUNTIME(packages) $dir
      }
    }
  } else {
    return [error "malformed implementation"]
  }

}

#------------------------------------------------------------------------------
# parser for the protocol
#------------------------------------------------------------------------------
proc @protocol { {args {}} } {
  global @RUNTIME

  foreach v {protocols instance,selectors class,selectors} {
    set $v {}
  }

  if [regexp {^([^\{]+)[\{](.+)[\}]$} $args -> realArgs code] {
    if [regexp {^([^ ]+)[ ]*<([^>]+)>[ ]*$} \
      $realArgs -> protocol protocols] { 
    } elseif [regexp {^([^ ]+)[ ]*$} \
      $realArgs -> protocol] {
    }
  } else {
    return [error "malformed protocol declaration"]
  }

  set code [string trimright $code "\r\n "]

  #- removing comments from code
  set newcode ""
  foreach line [split $code \n] {
    if [regexp {^[ ]*[#](.*)$} $line] { continue }
    append newcode $line\n
  }
  set code $newcode

  #- parsing code
  foreach line [split $code \;] {
    set line [string trim $line "\r\n "]
    if {![string length $line]} { continue }
    if [regexp {^([+-])[ ]*([^$]+)$} $line -> sign message] {
      set selector [@selector $message]
      set args [@args $message]
      switch -- $sign {
        + {
           set type class
        }
        - {
          set type instance
        }
        default { }
      }
      lappend ${type},selectors $selector
    } else {
      return [error "unknow statement \"$line\" for $protocol protocol"]
    }
  }


  if [string length $protocol] {
    foreach proto $protocols {
      #- loads parent protocol definitions
      if {![@proto $proto]} {
        return [error "unknown protocol: $proto"]
      }
      foreach type {instance class} {
        upvar #0 @@${proto} PROTO
        foreach selector $PROTO(${type},selectors) {
          if {![${type},selectors contains: $selector]} {
            lappend ${type},selectors $selector
          }
        }
      }
    }
  }

  upvar #0 @@${protocol} PROTOCOL
  set PROTOCOL(protocols) $protocols

  foreach type {instance class} {
    if {![info exists PROTOCOL(${type},selectors)]} { 
      set PROTOCOL(${type},selectors) {}
    }
    foreach selector [set ${type},selectors] {
      if {![PROTOCOL(${type},selectors) contains: $selector]} {
        lappend PROTOCOL(${type},selectors) $selector
      }
    }
  }
  if {![@RUNTIME(protocols) contains: $protocol]} {
    lappend @RUNTIME(protocols) $protocol
  }
  upvar #0 @DOC DOC
  set DOC(protocol,$protocol) [info script]
  return
}
if {![info exists @RUNTIME(protocols)]} {
  set @RUNTIME(protocols) {}
}


#----------------------------
# loads a class into runtime
#----------------------------
proc @class {class} {
  global @RUNTIME

  if [regexp {\.} $class] {
    #- package specified
    set splitted [split $class .]
    set class [lindex $splitted end]
    set package [file join [lreplace $splitted end end]]
    set path [file join [set @RUNTIME(directory)] $package]
  } elseif {[llength [file split $class]] > 1} {
    #- path specified
    set splitted [file split $class]
    set class [lindex $splitted end]
    set package [file join [lreplace $splitted end end]]
    set path $package
  } else {
    set package {}
    set path [set @RUNTIME(path)]
  }

  if [@RUNTIME(classes) contains: $class] { 
    if [string length $package] {
      upvar #0 @$class CLASS
      if {"$CLASS(package)" == "$package"} {
        return 1
      }
    }
    return 1 
  }
 
  foreach dir $path {
    set interface [file join $dir interface ${class}.tcl]
    set implementation [file join $dir implementation ${class}.tcl]
    if {[file exists $interface] && [file exists $implementation]} {
      @path $dir
      foreach f [list $interface $implementation] {
        if [catch {source $f} err] {
          return [error "source $f : $err"]
        } 
      }
      upvar #0 @$class CLASS
      #- removing lib directory from package path
      set libDir [set @RUNTIME(directory)]
      set dir [file split $dir]
      if {[dir contains: $libDir]} {
        set dir [lreplace $dir [set i [dir indexOf: $libDir]] $i]
      }
      set dir [file join $dir]
      set dir [string trimleft $dir /]
      set dir [string trimleft $dir \\]
      set CLASS(package) $dir
      return 1
    }
  }
  return 0
}
if {![info exists @RUNTIME(classes)]} {
  set @RUNTIME(classes) {}
  set @RUNTIME(packages) {}
}

#------------------------------------------------------------------------------
# unload a class or a protocol from the runtime
#------------------------------------------------------------------------------
proc @unload {class} {
  global @RUNTIME

  if [@RUNTIME(classes) contains: $class] {
    @RUNTIME(classes) remove: $class
    global @${class}
    unset @${class}
    rename $class ""
    foreach proc [info proc] {
      if [regexp ^@${class}, $proc] {
        rename $proc ""
      }
    }
  } elseif [@RUNTIME(protocols) contains: $class] {
    @RUNTIME(protocols) remove: $class
    global @@${class}
    unset @@${class}
  }
}

#------------------------------------------------------------------------------
# reload a class or a protocol into the runtime
#------------------------------------------------------------------------------
proc @reload {class} {
  global @RUNTIME

  if [@RUNTIME(classes) contains: $class] {
    @unload $class
    @class $class
  } elseif [@RUNTIME(protocols) contains: $class] {
    @unload $class
    @proto $class
  }
}

# -- runtime implementation --

#------------------------------------------------------------------------------
# gets integer maximum 
#------------------------------------------------------------------------------
proc @maxInteger {} {
  set i 1
  set index 1

  while {$i > 0} {
    set i [expr 1 << [incr index]]
  }
  expr $i - 1
}
set @RUNTIME(maxInteger) [@maxInteger]

#------------------------------------------------------------------------------
# gets next available object Id
#------------------------------------------------------------------------------
proc @nextId {} {
  upvar #0 @RUNTIME(counter) counter
  upvar #0 @RUNTIME(lastId) lastId
  upvar #0 @RUNTIME(maxInteger) maxInteger

  set maxIntegerLength [string length $maxInteger]
  set len [expr $maxIntegerLength -1]
  set incrementing 1
  set index [expr [llength $counter] -1]

  while {$incrementing} {
    #- incrementing the current counter
    set count [expr [string trimleft [lindex $counter $index] 0] + 1]

    #- have we reached the limit ?
    if {[string length $count] == $maxIntegerLength} {
      #- reset it + report to next counter
      set counter [lreplace $counter $index $index [format %0${len}i 0]]
      incr index -1
      if {$index == -1} {
        #- next counter doesn't exist yet
        set counter [linsert $counter 0 0]
        set index 0
      }
    } else {
      #- no limit reached so, updating counter
      set counter [lreplace $counter $index $index [format %0${len}i $count]]
      set incrementing 0
    }
  }
  set lastId [string trimleft [join $counter {}] 0]
}
if {![info exists @RUNTIME(lastId)]} {
  set @RUNTIME(lastId) 0
  set @RUNTIME(counter) 0
}

#------------------------------------------------------------------------------
# creates new instance from a class template
#------------------------------------------------------------------------------
proc @newInstanceFromClass { class } {
  upvar #0 @$class CLASS

  set id id[@nextId]
  upvar #0 @$id ID

  foreach var $CLASS(instance,variables) {
    if [regexp {^@([^$]+)$} $var -> arrayName] {
      upvar #0 @${id}@${arrayName} myArray

      #- workaround to set the array under Windows
      set myArray(foo) bar
      unset myArray(foo)
    } else {
      set ID($var) {}
    }
  }
  set ID(class) $class

  proc $id { {args {}} } {
    @sendMessageToObject [lindex [info level 0] 0] $args
  }

  set id
}

#------------------------------------------------------------------------------
# callback used when a message is sent to a class or an object
#------------------------------------------------------------------------------
proc @sendMessageToObject {self {message {}}} {
  upvar #0 @${self}(class) class
  upvar #0 @${class} SELF

  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }

  set selector [@selector $message]
  set args [@args $message]

  if {[info exists SELF(${type},methodFor,${selector})]} {
    eval $SELF(${type},methodFor,${selector}) $self $args
  } elseif {[info exists SELF(${type},methodFor,forwardInvocation:)]} {
    $SELF(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists SELF(${type},methodFor,forward:)]} {
    $SELF(${type},methodFor,forward:) $self $message
  } elseif {[info exists SELF(${type},methodFor,doesNotRecognize:)]} {
    $SELF(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SELF(${type},methodFor,error:)]} {
    $SELF(${type},methodFor,error:) $self \
      "$self cannot respond to \"$message\" message !"
  } else {
    error "$self cannot respond to \"$message\" message !"
  }

}

#------------------------------------------------------------------------------
# callback used when a message has to be sent to super
#------------------------------------------------------------------------------
proc @sendMessageToSuper { {args {}} } {
  #- get caller proc
  if [catch {set callerFullProcName [info level -1]} err] {
    return [error "cannot call super outside of a instance/class method ($err)"]
  }
  set callerProcName [lindex $callerFullProcName 0]
  set self [lindex $callerFullProcName 1]

  upvar #0 @${self}(class) class
  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }
  set message $args
  set selector [@selector $message]
  set args [@args $message]

  set _class [string trimleft [lindex [split $callerProcName ,] 0] @]
  upvar #0 @${_class} CLASS
  set superclass $CLASS(superclass)

  if {![string length $superclass]} {
    return [error "cannot call super method when superclass doesn't exist !"]
  }

  upvar #0 @$superclass SUPERCLASS
  if {[info exists SUPERCLASS(${type},methodFor,${selector})]} {
    eval $SUPERCLASS(${type},methodFor,${selector}) $self $args
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forwardInvocation:)]} {
    $SUPERCLASS(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forward:)]} {
    $SUPERCLASS(${type},methodFor,forward:) $self $message
  } elseif {[info exists SUPERCLASS(${type},methodFor,doesNotRecognize:)]} {
    $SUPERCLASS(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SUPERCLASS(${type},methodFor,error:)]} {
    $SUPERCLASS(${type},methodFor,error:) $self \
      "super of $self cannot respond to \"$message\" message !"
  } else {
    error "super of $self cannot respond to \"$message\" message !"
  }
}

#------------------------------------------------------------------------------
# gets level of caller procedure (NB: works only when message without aruments)
#------------------------------------------------------------------------------
proc @getCallerLevel {} {
  set cmd "uplevel info level"
  set currentLevel [eval $cmd]
  set currentMessage [eval $cmd 0]
  if [regexp {^@} $currentMessage] {
    set selector [lindex [split [lindex $currentMessage 0] ,] end]
    set object [lindex $currentMessage end]
    set msg "$object $selector"
    set level $currentLevel
    while {$level > 0} {
      if [catch {
        set level [eval $cmd]
        set message [eval $cmd 0]
      } err] {
        return 0
      } elseif {"$msg" == "$message"} {
        #- maybe called by unknown
        set cmd "uplevel $cmd"
        if [catch {
          set level [eval $cmd]
          set message [eval $cmd 0]
        } err] {
          return $level
        } elseif {"unknown $msg" == "$message" \
           || "::unknown $msg" == "$message"} {
          return [expr $level -1]
        } else {
          return $level
        }
      }
      set cmd "uplevel $cmd"
      incr level -1
    }
  }
  return
}

#------------------------------------------------------------------------------
# gets access to class variables (used in class methods)
#------------------------------------------------------------------------------
proc @getAccessToClassVariables {} {
  uplevel {
    upvar #0 @${self} SELF
    upvar #0 @$SELF(class) CLASS
    foreach field $CLASS(class,variables) {
      if {![regexp {^@([^$]+)$} $field -> arrayName]} {
        upvar #0 @${SELF(class)}($field) $field
      } else {
        upvar #0 @${SELF(class)}@{$arrayName} $arrayName
      }
    }
    set super @sendMessageToSuper
  }
}

#------------------------------------------------------------------------------
# gets access to class and instance variables (used in instance methods)
#------------------------------------------------------------------------------
proc @getAccessToClassAndInstanceVariables {} {
  uplevel {
    upvar #0 @${self} SELF
    upvar #0 @$SELF(class) CLASS
    foreach field $CLASS(class,variables) {
      if {![regexp {^@([^$]+)$} $field -> arrayName]} {
        upvar #0 @${SELF(class)}($field) $field
      } else {
        upvar #0 @${SELF(class)}@{$arrayName} $arrayName
      }
    }
    set super @sendMessageToSuper
    foreach field $CLASS(instance,variables) {
      if {![regexp {^@([^$]+)$} $field -> arrayName]} {
        upvar #0 @${self}($field) $field
      } else {
        upvar #0 @${self}@${arrayName} $arrayName
      }
    }
  }
}

# -- dynamic loading --

#-------------------------------
# extends or return search path 
#-------------------------------
proc @path {{args {}}} {
  upvar #0 @RUNTIME RUNTIME

  if {![string length $args]} { return $RUNTIME(path) }
  set directories [join $args]

  foreach directory $directories {
    set dir1 $directory
    set dir2 [file join $RUNTIME(directory) $directory]

    foreach i {1 2} {
      if [file isdirectory [set dir$i]] {
        set interface [file join [set dir$i] interface]
        set implementation [file join [set dir$i] implementation]
        if {[file isdirectory $interface] && \
          [file isdirectory $implementation]} {
          if {![RUNTIME(path) contains: [set dir$i]]} {
            lappend RUNTIME(path) [set dir$i]
          }
          break
        }
      } 
    }
  }
  set RUNTIME(path)
}
if {![info exists @RUNTIME(path)]} {
  set @RUNTIME(path) [set @RUNTIME(directory)]
}

#-------------------------------
# loads a protocol into runtime
#-------------------------------
proc @proto {protocol} {
  global @RUNTIME

  set path [set @RUNTIME(path)]
  if [regexp {\.} $protocol] {
    #- package specified
    set splitted [split $protocol .]
    set protocol [lindex $splitted end]
    set package [file join [lreplace $splitted end end]]
    set path [file join [set @RUNTIME(directory)] $package]
  } elseif {[llength [file split $protocol]] > 1} {
    #- path specified
    set splitted [file split $protocol]
    set protocol [lindex $splitted end]
    set package [file join [lreplace $splitted end end]]
    set path $package
  }

  if [@RUNTIME(protocols) contains: $protocol] { return 1 }

  foreach dir $path {
    set proto [file join $dir protocol ${protocol}.tcl]
    if {[file exists $proto] && [file exists $proto]} {
      if [catch {
        source $proto
      } error] {
        return [error "$error"]
      } else {
        upvar #0 @@$protocol PROTOCOL
        set dir [file split $dir]
        set libDir [set @RUNTIME(directory)]
        if {[dir contains: $libDir]} {
          set dir [lreplace $dir [set i [dir indexOf: $libDir]] $i]
        }
        set dir [file join $dir]
        set dir [string trimleft $dir /]
        set dir [string trimleft $dir \\]
        set PROTOCOL(package) $dir
        return 1
      }
    }
  }
  return 0
}

#-------------------------------
# loads a package into runtime
#-------------------------------
proc @package {{args {}}} {
  upvar #0 @RUNTIME RUNTIME

  if {![string length $args]} { return $RUNTIME(packages) }
  set packages [join $args]

  foreach package $packages {
    if {![file isdirectory $package]} {
      if {![file isdirectory [file join $RUNTIME(directory) $package]]} {
        foreach path $RUNTIME(path) {
          if {"[file tail $path]" == "$package"} {
            set dir $package
          }
        }
      } else {
        set dir [file join $RUNTIME(directory) $package]
      }
    } else {
      set dir $package
    }
    if {![info exists dir]} {
      return [error "couldn't find $package package"]
    } else {
      @path $dir
      if {![catch {set files [glob \
        [file join $package interface *.tcl]]} err]} {
         regsub -all -- {.tcl} $files {} classes
         foreach class $classes {
           @class $class
         }
      }
    }
  }
}

# -- utils --

#------------------------------------------------------------------------------
# date getter
#------------------------------------------------------------------------------
proc @date {} {
  return [clock format [clock seconds] -format {%d/%m/%Y}]
}

#------------------------------------------------------------------------------
# time getter
#------------------------------------------------------------------------------
proc @time {} {
  return [clock format [clock seconds] -format {%H:%M:%S}]
}

#------------------------------------------------------------------------------
# date and time getter
#------------------------------------------------------------------------------
proc @now {} {
  return [clock format [clock seconds] -format {%d/%m/%Y %H:%M:%S}]
}

#------------------------------------------------------------------------------
# log ouput to file
#------------------------------------------------------------------------------
proc @log {file message} {
  set f [open $file a+]
  puts $f "\[[@date] [@time]\] " nonewline
  puts $f $message
  close $f
  return
}

#------------------------------------------------------------------------------
# common debug output
#------------------------------------------------------------------------------
proc @debug {message} {
  puts stderr "\[[@date] [@time]\] \[DEBUG\] $message"
}

#------------------------------------------------------------------------------
# common error output
#------------------------------------------------------------------------------
proc @error {message} {
  puts stderr "\[[@date] [@time]\] \[ERROR\] $message"
}

#------------------------------------------------------------------------------
# add a new tcl syntax
#------------------------------------------------------------------------------
proc @syntax {body} {
  append b $body\n[info body unknown]
  proc unknown args $b
}

# -- new syntax --

#------------------------------------------------------------------------------
# dynamic class loading
#------------------------------------------------------------------------------
@syntax {
  if {[string length [info proc @class]] && [@class [lindex $args 0]] == 1} {
    return [eval $args]
  }
}

#------------------------------------------------------------------------------
# myArray dump
# myFieldValueList dump
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "dump"} {
    upvar [lindex $args 0] list
    set result {}
    set length 0
    if [array exists list] {
      foreach field [lsort [array names list]] {
        if {[set len [string length $field]] > $length} { set length $len }
        lappend result [list $field $list($field)]
      }
    } elseif {[info exists list] && ![expr [llength $list] % 2]} {
      foreach {field value} $list {
        if {[set len [string length $field]] > $length} { set length $len }
        lappend result [list $field $value]
      }
    } else {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set result2 {}
    foreach line $result {
      append result2 [format %${length}s [lindex $line 0]]
      append result2 { = }
      append result2 [lrange $line 1 end]\n
    }
    return [string trimright $result2 \n]
  }
}

#------------------------------------------------------------------------------
# set list {34 56 32}
#
# list each: { puts stderr "in \"$self\" list, index $index = $each" }
#   -> in "34 56 32" list, index 0 = 34
#      in "34 56 32" list, index 1 = 56
#      in "34 56 32" list, index 2 = 32
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "each:"} {
    upvar [lindex $args 0] _self
    set cmd {
      foreach var {self index each} {
        if [info exists $var] {
          set @@@@@${var} [set $var]
        }
      }
    }
    append cmd "set self "
    append cmd [list $_self]
    append cmd \n
    append cmd "set index 0\n"
    append cmd "foreach each \$self {\n"
    append cmd "  [join [lrange $args 2 end]]\n"
    append cmd "  incr index\n"
    append cmd "}"

    append cmd {
      foreach var {self index each} {
        if [info exists @@@@@${var}] {
          set $var [set @@@@@${var}]
        } else {
          unset $var
        }
      }
    }
    uplevel $cmd
    return
  }
}

#------------------------------------------------------------------------------
# factorial
#
# 3 ! ->  3 * 2 * 1
#
# set a 4
# a ! -> set a [expr 4 * 3 * 2 * 1]
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "!"} {
    if {![regexp {^([+\-0-9\.e]+)$} [lindex $args 0] -> nb]} {
      upvar [lindex $args 0]  nb
      if {![info exists nb]} {
        return [error "\"[lindex $args 0]\" doesn't exist"]
      }
    }
    set i [expr $nb - 1]
    set tmpnb $nb
    while {$i > 1} {
      set tmpnb [expr $nb * $i]
      if {$tmpnb > 0} {
        set nb $tmpnb
        incr i -1
      } else {
        return [error "stopped factorial to [expr $i -1] (to big value)"]
      }
    }
    return $nb
  }
}

#------------------------------------------------------------------------------
# mytext encrypt
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "encrypt"} {
    upvar [lindex $args 0] txt
    if {![info exists txt]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set encrypted {}
    foreach c [split $txt {}] {
      scan $c %c ascii
      if [expr $ascii % 2] {
        incr ascii 15
      } else {
        incr ascii 17
      }
      if {$ascii > 255} { 
        set ascii [expr ($ascii - 255)] 
      }
      append encrypted [format %c $ascii]
    }
    set txt $encrypted
    return
  }
}

#------------------------------------------------------------------------------
# mytext decrypt
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "decrypt"} {
    upvar [lindex $args 0] txt
    if {![info exists txt]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set decrypted {}
    foreach c [split $txt {}] {
      scan $c %c ascii
      if [expr $ascii % 2] {
        incr ascii -17
      } else {
        incr ascii -15
      }
      if {$ascii < 0} { 
        set ascii [expr (0 - $ascii)] 
      }
      append decrypted [format %c $ascii]
    }
    set txt $decrypted
    return
  }
}

#------------------------------------------------------------------------------
# set myList {1 2 3 3 2}
# myList sort -> {1 2 3}
#------------------------------------------------------------------------------
@syntax { 
  if {"[lindex $args 1]" == "sort"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set newlist {}
    foreach element [lsort $list] {
      if {[lsearch -exact $newlist $element] == -1} {
        lappend newlist $element
      }
    }
    set list $newlist
    return $list
  }
}

#------------------------------------------------------------------------------
# set myList {12 43 5}
#
# myList addIfNotYetIn: 43 -> {12 43 5}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "addIfNotYetIn:"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set element [lindex $args 2]
    if {[lsearch -exact $list $element] == -1} {
      lappend list $element
    }
    return $list
  }
}

#------------------------------------------------------------------------------
# set fieldValueList {Marc 23 Luc 43 Tom 24}
#
# fieldValueList sortByValue
#   -> {Marc 23 Tom 24 Luc 43}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "sortByValue"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    foreach {field value} $list {
      lappend values $value
      set fields($value,$field) $field
    }
    values sort
    set list {}
    foreach value $values {
      foreach f [array names fields ${value},*] {
        lappend list $fields($f)
        lappend list $value
      }
    }
    unset fields values field value
    return $list
  }
}

#------------------------------------------------------------------------------
# set fieldValueList {Marc 23 Luc 43 Tom 24}
#
# fieldValueList sortByField
#   -> {Luc 43 Marc 23 Tom 24}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "sortByField"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    foreach {field value} $list {
      lappend fields $field
      set values($field,$value) $value
    }
    fields sort
    set list {}
    foreach field $fields {
      foreach f [array names values ${field},*] {
        lappend list $field
        lappend list $values($f)
      }
    }
    unset fields values field value
    return $list
  }
}

#------------------------------------------------------------------------------
# set fieldValueList {Marc 23 Luc 43 Tom 24}
#
# fieldValueList fields
#   -> {Marc Luc Tom}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "fields"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set fields {}
    foreach {field value} $list {
      lappend fields $field
    }
    return $fields
  }
}

#------------------------------------------------------------------------------
# set fieldValueList {Marc 23 Luc 43 Tom 24}
#
# fieldValueList values
#   -> {23 43 24}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "values"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set values {}
    foreach {field value} $list {
      lappend values $value
    }
    return $values
  }
}

#------------------------------------------------------------------------------
# set fieldValueList {Marc 23 Luc 43 Tom 24}
#
# fieldValueList set: Marc to: 43 
#   -> $fieldValueList = {Marc 43 Luc 43 Tom 24}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "set:" && "[lindex $args 3]" == "to:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    set value [lindex $args 4]
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    if {[set index [lsearch -exact $list $field]] != -1} {
      incr index
      set list [lreplace $list $index $index $value]
    } elseif [string length $field] {
      return \
        [error "couldn't find \"$field\" field in \"[lindex $args 0]\" list"]
    }
    return $list
  }
}

#------------------------------------------------------------------------------
# set fieldValueList {Marc 23 Luc 43 Tom 24}
#
# fieldValueList get: Marc 
#   -> 23
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "get:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    if {[set index [lsearch -exact $list $field]] != -1} {
      incr index
      return [lindex $list $index]
    } elseif [string length $field] {
      return \
        [error "couldn't find \"$field\" field in \"[lindex $args 0]\" list"]
    }
  }
}

#------------------------------------------------------------------------------
# set fieldValueList {Marc 23 Luc 43 Tom 24}
#
# fieldValueList add: {John 2}
#   -> {Marc 23 Luc 43 Tom 24 John 2}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "add:"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set list [join [list $list [lindex $args 2]]]
    return $list
  }
}

#------------------------------------------------------------------------------
# set myFieldValueList {marc 3 luc 4}
#
# myFieldValueList removeField: marc 
#   -> $myFieldValueList = {luc 4}
#------------------------------------------------------------------------------
@syntax { 
  if {"[lindex $args 1]" == "removeField:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    while {[set index [lsearch -exact $list $field]] != -1} {
      set list [lreplace $list $index [expr $index + 1]]
    }
    return $list
  }
}

#------------------------------------------------------------------------------
# set myList {12 45 2 34}
#
# myList remove: 45
#   -> $myList = {12 2 34}
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "remove:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    while {[set index [lsearch -exact $list $field]] != -1} {
      set list [lreplace $list $index $index]
    }
    return $list
  }
}

#------------------------------------------------------------------------------
# set myList {1 2 34 54}
#
# myList contains: 34
#  -> 1
# myList contains: 35
#  -> 0
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "contains:"} {
    upvar [lindex $args 0] listname
    set field [lindex $args 2]
    if {![info exists listname]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    } elseif {[set index [lsearch -exact $listname $field]] != -1} {
      return 1
    } else {
      return 0
    }
  }
}

#------------------------------------------------------------------------------
# set myList {1 2 34 54}
#
# myList indexOf: 34
#  -> 2
# myList indexOf: 23
#  ->
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "indexOf:"} {
    upvar [lindex $args 0] listname
    set field [lindex $args 2]
    if {![info exists listname]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    } elseif {[set index [lsearch -exact $listname $field]] != -1} {
      return $index
    } else { 
      return
    }
  }
}

#------------------------------------------------------------------------------
# set a [expr $a +  2] -> a += 2
#
# nb: will work for +,-,*,/,%,>>,<<
#------------------------------------------------------------------------------
@syntax {
  if {[regexp {^([+\-\\\*%/><]+)=$} [lindex $args 1] -> sign]} {
    uplevel "set [lindex $args 0] \
      \[expr \$[lindex $args 0] $sign [lrange $args 2 end]\]"
    return
  }
}

#------------------------------------------------------------------------------
# incr a -1 -> --a
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "--"} { 
    uplevel "incr [lindex $args 0] -1"
    return 
  }
}

#------------------------------------------------------------------------------
# incr a 1  -> ++a
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "++"} { 
    uplevel "incr [lindex $args 0]"
    return
  }
}

#------------------------------------------------------------------------------
# set a 1 -> a = 1
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "="} {
    set cmd "set [lindex $args 0] [lrange $args 2 end]"
    uplevel "set [lindex $args 0] [lrange $args 2 end]"
    return
  }
}

#------------------------------------------------------------------------------
# print string to standard output
#
# "Hello world" stdout
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "stdout"} {
    puts stdout [lindex $args 0]
    return
  }
}

#------------------------------------------------------------------------------
# print string to standard error
#
# "Hello world" stderr
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "stderr"} {
    puts stderr [lindex $args 0]
    return
  }
}

#------------------------------------------------------------------------------
# returns string length
#
# "Hello world" length
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "length"} {
    return [string length [lindex $args 0]]
  }
}

#------------------------------------------------------------------------------
# returns list length
#
# "Hello world" llength
#------------------------------------------------------------------------------
@syntax {
  if {"[lindex $args 1]" == "llength"} {
    return [llength [lindex $args 0]]
  }
}

#------------------------------------------------------------------------------
# tk specific
#------------------------------------------------------------------------------
if {![catch {winfo exists .} err]} {
  #-----------------------------------------------------------------------------
  # overriding library scrollbar procedures 
  # to keep the scrollbar relief unchanged when clicked
  #-----------------------------------------------------------------------------
  proc tkScrollButtonDown {w x y} {
    set element [$w identify $x $y]
    if {[string match slider $element]} {
        tkScrollStartDrag $w $x $y
    } else {  
        tkScrollSelect $w $element initial
    }
  }
  proc tkScrollButtonUp {w x y} {
    tkCancelRepeat
    tkScrollEndDrag $w $x $y
    $w activate [$w identify $x $y]
  }
  proc tkScrollTopBottom {w x y} {
    set element [$w identify $x $y]
    if {[string match *1 $element]} {
        tkScrollToPos $w 0
    } elseif {[string match *2 $element]} {
        tkScrollToPos $w 1
    }
  }
  proc tkScrollButton2Down {w x y} {
    set element [$w identify $x $y]
    if {[string match {arrow[12]} $element]} {
        tkScrollButtonDown $w $x $y
        return
    }
    tkScrollToPos $w [$w fraction $x $y]
    update idletasks
    $w activate slider
    tkScrollStartDrag $w $x $y
  }

  #----------------------------------------------------------------------------
  # ressources
  #----------------------------------------------------------------------------
  option add *highlightThickness 0
  option add *selectBorderWidth 0
  foreach widget {Button Scrollbar Listbox Entry Canvas Text Menu MenuButton} {
    option add *${widget}.borderWidth 1 
  }
  option add *Text.wrap none

  #- colors
  foreach w {Listbox Text Entry} {
    option add *$w.background #fdfdfd
    option add *$w.foreground black 
    option add *$w.selectBackground #d0d0e0
    option add *$w.selectForeground black
  }

  #- removing bold font
  foreach widget {text listbox label button entry} {
    set font [[$widget ._test] cget -font]
    destroy ._test
    set class \
      [string toupper [string index $widget 0]][string range $widget 1 end]
    if {[set index [lsearch -exact $font bold]] != -1} {
      option add *${class}.Font [lreplace $font $index $index]
    }
  }

  if {![info exists @RUNTIME(images)]} {
    set @RUNTIME(images) {}
  }

  proc @image {{name {}} {image {}}} {
    upvar #0 @RUNTIME Runtime
    if [string length $image] {
      Runtime(images) add: [list $name $image]
      return $image
    } elseif {![string length $name]} {
      return $Runtime(images)
    } elseif [Runtime(images) contains: $name] {
      return [Runtime(images) get: $name]
    } else {
      set file [file join $Runtime(directory) images ${name}.*]
      if [catch {set files [glob $file]} err] {
        return
      } else {
        set file [lindex $files 0]
        Runtime(images) add: \
          [list $name [set img [image create photo -file $file]]]
        return $img
      }
    }
  }

  proc @bitmapCopy {image} {
    if {![string length [set img [@image $image]]]} {
      set img $image 
    }
    set options {}
    if [catch {
      foreach option [$img configure] {
        options add: [list [lindex $option 0] [lindex $option end]]
      }
    } err] {
      return [error "couldn't get bitmap configuration ($err)"]
    }
    return [eval image create bitmap $options]
  }

  @image down_arrow [image create bitmap \
    -data {
      #define down_arrow_width 11
      #define down_arrow_height 11
      static unsigned char down_arrow_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x03, 0xfc, 0x01,
        0xf8, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]
 
  @image up_arrow [image create bitmap \
    -data {
       #define up_arrow_width 11
       #define up_arrow_height 11
       static unsigned char up_arrow_bits[] = {
         0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x70, 0x00, 0xf8, 0x00, 0xfc, 0x01,
         0xfe, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
       };
     }
   ]

  @image left_arrow [image create bitmap \
    -data {
      #define left_arrow_width 11
      #define left_arrow_height 11
      static unsigned char left_arrow_bits[] = {
        0x00, 0x00, 0x40, 0x00, 0x60, 0x00, 0x70, 0x00, 0x78, 0x00, 0x7c, 0x00,
        0x78, 0x00, 0x70, 0x00, 0x60, 0x00, 0x40, 0x00, 0x00, 0x00, 
      };
    }
  ]

  @image right_arrow [image create bitmap \
    -data {
      #define right_arrow_width 11
      #define right_arrow_height 11
      static unsigned char right_arrow_bits[] = {
        0x00, 0x00, 0x10, 0x00, 0x30, 0x00, 0x70, 0x00, 0xf0, 0x00, 0xf0, 0x01,
        0xf0, 0x00, 0x70, 0x00, 0x30, 0x00, 0x10, 0x00, 0x00, 0x00, 
      };
    }
  ]

  @image return [image create bitmap \
    -data {
      #define return_width 15
      #define return_height 10
      static unsigned char return_bits[] = {
        0x10, 0x3c, 0x18, 0x04, 0x2c, 0x04, 0xe6, 0x07, 0x03, 0x00, 0x01, 0x00,
        0x02, 0x00, 0x04, 0x00, 0x08, 0x00, 0x10, 0x00, 
      };
    } -maskdata {
        #define returnmask_width 15
        #define returnmask_height 10
        static unsigned char returnmask_bits[] = {
          0x10, 0x7c, 0x18, 0x44, 0x2c, 0x44, 0xe6, 0x47, 0x03, 0x40, 0x01, 0x40,
          0xe2, 0x7f, 0x24, 0x00, 0x28, 0x00, 0x30, 0x00, 
        };
    } -background white -foreground black 
  ]

  @image radioOn [image create bitmap \
    -data {
      #define radioOn_width 16
      #define radioOn_height 16
      static unsigned char radioOn_bits[] = {
        0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
        0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x04, 0x00,
        0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
    } -maskdata {
        #define radioOnmask_width 16
        #define radioOnmask_height 16
        static unsigned char radioOnmask_bits[] = {
          0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0xc4, 0x07, 0xe4, 0x0f,
          0xf2, 0x1f, 0xf2, 0x9f, 0xf2, 0x9f, 0xf2, 0x9f, 0xf2, 0x9f, 0xe4, 0x4f,
          0xc4, 0x47, 0x00, 0x20, 0x20, 0x18, 0xc0, 0x07, };
    } -background white -foreground black
  ]

  @image radioOff [image create bitmap \
    -data {
      #define radioOff_width 16
      #define radioOff_height 16
      static unsigned char radioOff_bits[] = {
        0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
        0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x04, 0x00,
        0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    } -maskdata {
        #define radioOffmask_width 16
        #define radioOffmask_height 16
        static unsigned char radioOffmask_bits[] = {
          0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
          0x02, 0x00, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x04, 0x40,
          0x04, 0x40, 0x00, 0x20, 0x10, 0x18, 0xe0, 0x07, 
        };
    } -background white -foreground black
  ]

  @image empty16x16 [image create bitmap \
    -data {
      #define switchOff_width 16
      #define switchOff_height 16
      static unsigned char switchOff_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]

  @image switchOff [@bitmapCopy empty16x16]
  @image switchOn [image create bitmap \
    -data {
      #define switchOn_width 16
      #define switchOn_height 16
      static unsigned char switchOn_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x0c,
        0x20, 0x06, 0x20, 0x03, 0xa0, 0x01, 0xc0, 0x00, 0x60, 0x00, 0x30, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    } -maskdata {
        #define switchOnmask_width 16
        #define switchOnmask_height 16
        static unsigned char switchOnmask_bits[] = {
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x00, 0x0e,
          0x30, 0x07, 0xb0, 0x03, 0xf0, 0x01, 0xf0, 0x00, 0x70, 0x00, 0x30, 0x00,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        };
    } -background white -foreground black
  ]
   
  @image empty9x9 [image create bitmap \
    -data {
      #define empty_width 9
      #define empty_height 9
      static unsigned char empty_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00
      };
    }
  ]

  @image empty [@bitmapCopy empty9x9]

  @image plus [image create bitmap \
    -data {
      #define plus_width 9
      #define plus_height 9
      static unsigned char plus_bits[] = {
        0xff, 0x01, 0x01, 0x01, 0x11, 0x01, 0x11, 0x01, 0x7d, 0x01, 0x11, 0x01,
        0x11, 0x01, 0x01, 0x01, 0xff, 0x01
      };
    }
  ]

  @image minus [image create bitmap \
    -data {
      #define minus_width 9
      #define minus_height 9
      static unsigned char minus_bits[] = {
        0xff, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x7d, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0xff, 0x01
      };
    }
  ]

  set circleData {
    -data {
      #define round_width 10
      #define round_height 10
      static unsigned char round_bits[] = {
        0x78, 0x00, 0x84, 0x00, 0x02, 0x01, 0x01, 0x02, 0x01, 0x02,
        0x01, 0x02, 0x01, 0x02, 0x02, 0x01, 0x84, 0x00, 0x78, 0x00
      }
    } -maskdata {
      #define round_mask_width 10
      #define round_mask_height 10
      static unsigned char round_mask_bits[] = {
        0x78, 0x00, 0xfc, 0x00, 0xfe, 0x01, 0xff, 0x03, 0xff, 0x03,
        0xff, 0x03, 0xff, 0x03, 0xfe, 0x01, 0xfc, 0x00, 0x78, 0x00
      }
    }
  }

  @image circle [eval image create bitmap $circleData]

  foreach color {white blue red green yellow orange black violet brown grey} {
    @image ${color}_circle [set img [@bitmapCopy circle]]
    $img configure -background $color
  }

  @image big_directory [image create bitmap \
    -data {
      #define dir_width 40
      #define dir_height 36
      static unsigned char dir_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0xff,
        0x00, 0x00, 0x00, 0x04, 0x00, 0x01, 0x00, 0x00, 0x04, 0x00, 0xff, 0xff,
        0x1f, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04,
        0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00,
        0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20,
        0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00,
        0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00,
        0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04,
        0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00,
        0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20,
        0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00,
        0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00,
        0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0x04, 0x00, 0x00, 0x00, 0x20, 0xfc,
        0xff, 0xff, 0xff, 0x3f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      };
    } -maskdata {
      #define dirmask_width 40
      #define dirmask_height 36
      static unsigned char dirmask_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0xff,
        0x00, 0x00, 0x00, 0xfc, 0xff, 0x01, 0x00, 0x00, 0xfc, 0xff, 0xff, 0xff,
        0x1f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc,
        0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff,
        0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f,
        0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff,
        0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff,
        0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc,
        0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff,
        0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f,
        0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff,
        0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff,
        0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc, 0xff, 0xff, 0xff, 0x3f, 0xfc,
        0xff, 0xff, 0xff, 0x3f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      };
    }
  ]

  @image big_file [image create bitmap \
    -data {
      #define file_width 40
      #define file_height 36
      static unsigned char file_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff,
        0xff, 0x07, 0x00, 0x00, 0x01, 0x00, 0x0c, 0x00, 0x00, 0x01, 0x00, 0x14,
        0x00, 0x00, 0x01, 0x00, 0x24, 0x00, 0x00, 0x01, 0x00, 0x44, 0x00, 0x00,
        0x01, 0x00, 0x84, 0x00, 0x00, 0x01, 0x00, 0xfc, 0x01, 0x00, 0x01, 0x00,
        0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01,
        0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01,
        0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00,
        0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00,
        0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00,
        0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01,
        0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01,
        0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00,
        0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00,
        0x01, 0x00, 0x00, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      };
    } -maskdata {
      #define filecache_width 40
      #define filecache_height 36
      static unsigned char filecache_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff,
        0xff, 0x07, 0x00, 0x00, 0xff, 0xff, 0x0f, 0x00, 0x00, 0xff, 0xff, 0x1f,
        0x00, 0x00, 0xff, 0xff, 0x3f, 0x00, 0x00, 0xff, 0xff, 0x7f, 0x00, 0x00,
        0xff, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff,
        0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01,
        0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff,
        0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff,
        0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00,
        0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff,
        0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01,
        0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff,
        0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff,
        0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00,
        0xff, 0xff, 0xff, 0x01, 0x00, 0xff, 0xff, 0xff, 0x01, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      };
    }
  ]

  @image directory [image create bitmap \
    -data {
#define small_dir_width 16
#define small_dir_height 12
static unsigned char small_dir_bits[] = {
  0xfe, 0x00, 0x01, 0x01, 0x01, 0x7f, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80,
  0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xfe, 0x7f,
  };
    } -maskdata {
#define small_dir_cache_width 16
#define small_dir_cache_height 12
static unsigned char small_dir_cache_bits[] = {
  0xfe, 0x00, 0xff, 0x01, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x7f,
  };
    }
  ]

  @image opened_directory [image create bitmap \
    -data {
#define opened_small_dir_width 16
#define opened_small_dir_height 12
static unsigned char opened_small_dir_bits[] = {
  0xfe, 0x00, 0x01, 0x01, 0x01, 0x7f, 0x01, 0x80, 0xff, 0x81, 0x01, 0xff,
  0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xfe, 0x7f,
  };
    } -maskdata {
#define small_dir_cache_width 16
#define small_dir_cache_height 12
static unsigned char small_dir_cache_bits[] = {
  0xfe, 0x00, 0xff, 0x01, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x7f,
  };
    }
  ]

  @image file [image create bitmap \
    -data {
#define small_file_width 16
#define small_file_height 12
static unsigned char small_file_bits[] = {
  0xf8, 0x03, 0x08, 0x06, 0x08, 0x0a, 0x08, 0x1e, 0x08, 0x10, 0x08, 0x10,
  0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0xf8, 0x1f,
  };
    } -maskdata {
#define small_file_cache_width 16
#define small_file_cache_height 12
static unsigned char small_file_cache_bits[] = {
  0xf8, 0x03, 0xf8, 0x07, 0xf8, 0x0f, 0xf8, 0x1f, 0xf8, 0x1f, 0xf8, 0x1f,
  0xf8, 0x1f, 0xf8, 0x1f, 0xf8, 0x1f, 0xf8, 0x1f, 0xf8, 0x1f, 0xf8, 0x1f,
  };
    }
  ]

  @image bullet [image create bitmap \
    -data {
      #define bullet_width 6
      #define bullet_height 6
      static unsigned char bullet_bits[] = {
        0x1e, 0x3f, 0x07, 0x03, 0x03, 0x02, 
      };
    } -maskdata {
      #define bulletmask_width 6
      #define bulletmask_height 6
      static unsigned char bulletmask_bits[] = {
        0x1e, 0x3f, 0x3f, 0x3f, 0x3f, 0x1e, 
      };
    } -background white -foreground black
  ]

  @image plus_25x25 [image create bitmap \
    -data {
      #define cross_width 25
      #define cross_height 25
      static unsigned char cross_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00,
        0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00,
        0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00,
        0xfe, 0x83, 0xff, 0x00, 0x02, 0x00, 0x80, 0x00, 0x02, 0x00, 0x80, 0x00,
        0x02, 0x00, 0x80, 0x00, 0x02, 0x00, 0x80, 0x00, 0x02, 0x00, 0x80, 0x00,
        0xfe, 0x83, 0xff, 0x00, 0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00,
        0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00,
        0x00, 0x82, 0x00, 0x00, 0x00, 0x82, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, };
    } -maskdata {
      #define crossmask_width 25
      #define crossmask_height 25
      static unsigned char crossmask_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
        0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
        0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
        0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
        0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]

  @image green_plus [set img [@bitmapCopy [@image plus_25x25]]]
  $img configure -foreground black -background #22aa22

  @image minus_25x25 [image create bitmap \
    -data {
      #define minus_width 25
      #define minus_height 25
      static unsigned char minus_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0x02, 0x00, 0x80, 0x00, 0x02, 0x00, 0x80, 0x00,
        0x02, 0x00, 0x80, 0x00, 0x02, 0x00, 0x80, 0x00, 0x02, 0x00, 0x80, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 
       };
     } -maskdata {
      #define minusmask_width 25
      #define minusmask_height 25
      static unsigned char minusmask_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00, 0xfe, 0xff, 0xff, 0x00,
        0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]

   @image red_minus [@bitmapCopy [@image minus_25x25]]
   [@image red_minus] configure -foreground black -background #aa2222

   #- returns a more suitable list of available event data (from a bind command)
   proc @eventData {} {
     return [list [list \
       path %W focus %f width %w height %h keyCode %k x %x y %y X %X Y %Y \
       button %b char %A event %E keySym %K keySymNumber %N \
       type %T delta %D
     ]]
   }

  #- return the code to execute for a Destroy event to remove bindings
  proc @removeBindingsOnDestroyEvent {} {
    return {<Destroy> {
        foreach event [bind %W] { bind %W $event {} }
      } 
    }
  }

  proc @showBitmaps {} {
    global @RUNTIME

    toplevel .toplevel 
    set i 0
    foreach {name img} [set @RUNTIME(images)] {
      grid [label .toplevel.l$i -text $name -relief sunken -bd 1] [label .toplevel.i$i -image $img -relief sunken -bd 1] -sticky snew -padx 1 -pady 1
      incr i
    }
  }
}

#------------------------------------------------------------------------------
# build the doc from the current runtime
#------------------------------------------------------------------------------
proc @makedoc {} {
  upvar #0 @RUNTIME RUNTIME
  upvar #0 @DOC DOC

  #- creating directories
  if [catch {
    file mkdir doc
    file mkdir [file join doc]
  } err] {
    puts stderr "$err" ; return
  }

  #- index.html - frameset
  if [catch {
    set f [open [file join doc index.html] w]
    puts $f {
      <HTML>
      <HEAD>
      <TITLE>myObjectiveTcl</TITLE>
      <META NAME="Description" CONTENT="Objective-C like framework in pure tcl">
      <META NAME="Keywords" 
        CONTENT="myObjectiveTcl Objective-C tcl tk object-oriented OO">
      </HEAD>
       <FRAMESET COLS="300,*" FRAMEBORDER=1 FRAMESPACING=1 BORDER=1 
         BORDERCOLORR=BLACK>
        <FRAME FRAMEBORDER=1 AUTOSCROLLING=1 SRC=packages.html NAME=left
          MARGINWIDTH=10 MARGINHEIGHT=10>
        <FRAME FRAMEBORDER=0 AUTOSCROLLING=1 SRC=main.html NAME=right
          MARGINWIDTH=40 MARGINHEIGHT=10>
       </FRAMESET>
      </HTML>
    }
    close $f
  } err] {
    puts stderr "$err" ; return
  }

  #- main.html - empty file for first
  set f [open [file join doc main.html] w]
  puts $f "<HTML><BODY BGCOLOR=WHITE></BODY></HTML>"
  close $f

  #- classes.html - empty file for first
  set f [open [file join doc classes.html] w]
  puts $f "<HTML><BODY BGCOLOR=WHITE></BODY></HTML>"
  close $f

  #- collecting informations about classes and sorting them
  foreach class $RUNTIME(classes) {
    upvar #0 @$class @$class

    #- sort classes into packages
    lappend classes([set @${class}(package)]) $class

    #- sort selectors/variables into class specific, overridden, inherited
    foreach type1 {instance class} {
      foreach type2 {selectors variables} {
        set field ${type1},${type2}
        set value [set @${class}($field)]
        if {![string length [set superclass [set @${class}(superclass)]]]} {
          #- no superclass -> all is specific
          set specific($class,$field) [set @${class}($field)]
        } else {
          foreach subvalue $value { 
            #- doesn't exists in superclass -> class specific
            if {![@${superclass}($field) contains: $subvalue]} {
              lappend specific($class,$field) $subvalue
            } elseif [regexp variables $field] {
              #- exist in superclass -> inherited for variables
              lappend inherited($class,$field) $subvalue
            } elseif [string length [info proc @$class,$type1,$subvalue]] {
              #- the selector is overridden by the current class
              lappend overridden($class,$field) $subvalue
            } else {
              #- just simple inherited
              lappend inherited($class,$field) $subvalue
            }
          }
        }
      }
    }
    set inheritedFrom($class) ""
    set superExists 1
    set super $class
    while {$superExists} {
      if [string length [set super [set @${super}(superclass)]]] {
        lappend inheritedFrom($class) $super
        if {![info exists @${super}]} {
          upvar #0 @${super} @${super}
        }
      } else {
        set superExists 0
        break
      }
    }
    if [string length [set inheritedFrom($class)]] {
      set i 0
      foreach c [set inheritedFrom($class)] {
        set pkg [set @${c}(package)] 
        if {"[set @${class}(package)]" == "."} { set pkg $pkg }
        if {$i} {
          append inheritedFromInHtml($class) \
            ",&nbsp;<A HREF=../../$pkg/classes/$c.html>$c</A>" 
        } else {
          append inheritedFromInHtml($class) \
            "<A HREF=../../$pkg/classes/$c.html>$c</A>" 
        }
        incr i
      }
      set inheritedFrom($class) [set inheritedFromInHtml($class)]
      unset inheritedFromInHtml($class)
    }
  }

  #- collecting informations about protocols
  foreach proto $RUNTIME(protocols) {
    upvar #0 @@$proto @@$proto

    #- sort classes into packages
    lappend protocols([set @@${proto}(package)]) $proto

    #- sort selectors into protocol specific and inherited
    foreach type1 {instance class} {
     foreach type2 {selectors} {
        set field ${type1},${type2}
        set value [set @@${proto}($field)]

        if {![string length [set parents [set @@${proto}(protocols)]]]} {
          #- no parent protocols -> all is specific
          set specific(protocol,$proto,$field) [set @@${proto}($field)]
        } else {
          #- parent protocols exists, checking if there is inheritence
          foreach subvalue $value {
            set count 0
          
            foreach parent $parents {
              #- doesn't exists in superclass -> class specific
              if [@@${parent}($field) contains: $subvalue] {
                incr count
              }
            }
            if { $count > 0 } {
              lappend inherited(protocol,$proto,$field) $subvalue
            } else {
              lappend specific(protocol,$proto,$field) $subvalue
            } 
          }
        }
      }
    }
    set inheritedFromProtocol($proto) [set @@${proto}(protocols)]
    if [string length $inheritedFromProtocol($proto)] {
      set i 0
      foreach p $inheritedFromProtocol($proto) {
        set pkg [set @@${p}(package)]
        if {"[set @@${proto}(package)]" == "."} { set pkg $pkg }
        if {$i} {
          append inheritedFromInHtml($proto) \
            ",&nbsp;<A HREF=../../$pkg/protocols/$p.html>$p</A>"
        } else {
          append inheritedFromInHtml($proto) \
            "<A HREF=../../$pkg/protocols/$p.html>$p</A>"
        }
        incr i
      }
      set inheritedFromProtocol($proto) [set inheritedFromInHtml($proto)]
      unset inheritedFromInHtml($proto)
    }
   
  }
 
  #- packages.html - navigation throught packages
  set fd(left) [open [file join doc packages.html] w]
  puts $fd(left) {<HTML>}
  puts $fd(left) [@docHtmlHead]
  puts $fd(left) {
     <BODY BGCOLOR=#DEDEDE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR>
       <H1>myObjectiveTcl</H1><UL>
  }

  #- opening all packages file descriptor
  foreach package [set packages [lsort [array names classes]]] {
    #- creating package file 
    if [catch {
      set fd($package) [open [file join doc $package.html] w]
    } err] {
      puts stderr "$err" ; return
    }
    puts $fd($package) {<HTML>}
    puts $fd($package) [@docHtmlHead]
    puts $fd($package) {
     <BODY BGCOLOR=#DEDEDE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR>
       <H1>myObjectiveTcl</H1><UL>
    }
  }

  #- handlig html for each package
  foreach package $packages {
    #- adding link to main packages index
    puts $fd(left) "<LI><A HREF=$package.html>$package</A>"

    #- adding link to others package pages
    foreach p $packages {
      #- not for the current one
      if {"$p" != "$package"} {
        puts $fd($p) "<LI><A HREF=$package.html>$package</A>"
      }
    }

    puts $fd($package) "<LI><A HREF=packages.html>$package</A><UL>"
  
    if [catch {
      file mkdir [file join doc $package]
      file mkdir [file join doc $package classes]
      file mkdir [file join doc $package classes interfaces]
      file mkdir [file join doc $package classes implementations]
      file mkdir [file join doc $package protocols]
      file mkdir [file join doc $package protocols sources]
      file mkdir [file join doc $package categories]
      file mkdir [file join doc $package categories interfaces]
      file mkdir [file join doc $package categories implementations]
      file mkdir [file join doc $package images]
      file mkdir [file join doc $package samples]
    } err] {
      puts stderr "$err" ; return
    }
    #- classes
    if {[info exists classes($package)] && [string length $classes($package)]} {
      puts $fd($package) "<LI>Classes"
      puts $fd($package) "<UL>"
      foreach class [lsort $classes($package)] {
        if {![string length [set @${class}(superclass)]]} {
          #- root class because of no superclass -> in bold
          puts $fd($package) "<LI>\
            <A HREF=$package/classes/$class.html TARGET=right>\
            <B>$class</B></A>"
        } else {
          puts $fd($package) "<LI>\
            <A HREF=$package/classes/$class.html TARGET=right>\
            $class</A>"
        }
      }
      puts $fd($package) "</UL>"
    }

    #- protocols
    if {[info exists protocols($package)] && \
      [string length $protocols($package)]} {
      puts $fd($package) "<LI>Protocols"
      puts $fd($package) "<UL>"
      foreach protocol [lsort $protocols($package)] {
        puts $fd($package) "<LI>\
          <A HREF=$package/protocols/$protocol.html TARGET=right>\
          $protocol</A>"
      }
      puts $fd($package) "</UL>"
    }
    puts $fd($package) "</UL>"
  }
  puts $fd(left) "</UL>"

  puts $fd(left) {
     </BODY>
    </HTML>
  }
  close $fd(left)

  #- closing all packages
  foreach p $packages {
    puts $fd($p) "</UL></BODY></HTML>"
    close $fd($p)
  }


  #- classes/<class>.html 
  foreach class $RUNTIME(classes) {
    set package [set @${class}(package)]
    set f [open [file join doc $package classes $class.html] w]
    puts $f <HTML>
    puts $f [@docHtmlHead]
    puts $f {
       <BODY BGCOLOR=WHITE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR><BR>
    }
    puts $f "<H1>$class class</H1>\n\
      <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->\n\
      <TABLE BORDER=0>\n"
    if [string length [set inheritedFrom($class)]] {
      puts $f "<TR>\n\
         <TD ALIGN=RIGHT>inherited from: </TD>\n\
         <TD ALIGN=LEFT>[set inheritedFrom($class)]</TD>\n\
       </TR>"
    }


    if [string length [set @${class}(protocols)]] {
      puts $f "<TR><TD ALIGN=RIGHT>protocols:</TD><TD ALIGN=LEFT>"
      set h ""
      foreach p [set @${class}(protocols)] {
        if {![info exists @@${p}(package)]} { upvar #0 @@${p} @@${p} }
        append h \
          "<A HREF=../../[set @@${p}(package)]/protocols/${p}.html>$p</A>, "
      }
      puts $f "[string trimright $h {, }]</TD></TR>"
    }

    puts $f "\
       <TR>\n\
         <TD ALIGN=RIGHT>package:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=../../[set @${class}(package)].html TARGET=left>\
             [set @${class}(package)]</A></TD>\n\
       </TR>\n\
    "

    puts $f "<TR><TD></TD><TD></TD></TR>"

    puts $f "<TR>\n\
         <TD ALIGN=RIGHT>declared in:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=interfaces/$class.html>$DOC(class,interface,$class)</A>\n\
         </TD>\n\
       </TR>\n\
       <TR>\n\
         <TD ALIGN=RIGHT>implemented in:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=implementations/$class.html>\
             $DOC(class,implementation,$class)</A></TD>\n\
       </TR>\n"

    puts $f "\
      </TABLE><BR><BR>\n\
      <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"

   #- sample code ?
   if [file exists [set sample \
       [file join $RUNTIME(directory) $package samples ${class}.tcl]]] {
      #- sample code ?
        set dest [open [file join doc $package samples ${class}.html] w]
        set src [open $sample r]
        puts $dest {
          <HTML>
        }
        puts $dest [@docHtmlHead]
        puts $dest {
          <BODY BGCOLOR=WHITE><BR><BR>
        }
        puts $dest "<H1>$class sample</H1>"
        puts $dest "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>--><BR>"
        puts $dest "<FONT color=#808080><PRE>\n"
        puts $dest [read $src]
        puts $dest "</PRE></BODY></HTML>"
        close $dest
        close $src
        if {![file exists [file join doc $package images sample.jpg]]} {
           file copy [file join $RUNTIME(directory) images sample.jpg] \
             [file join doc $package images sample.jpg]
        }
        puts $f "<A ALIGN=RIGHT HREF=../samples/${class}.html>\
          <IMG ALIGN=RIGHT BORDER=0 SRC=../images/sample.jpg></A>"
    }

    #- screenshots ?
    if [file exists [set img [file join $RUNTIME(directory) $package screenshots ${class}.jpg]]] {
      #- code ?
      if [file exists [set sample [file join $RUNTIME(directory) $package samples ${class}.tcl]]] {
        set dest [open [file join doc $package samples ${class}.html] w]
        set src [open $sample r]
        puts $dest {
        <HTML>
        }
        puts $dest [@docHtmlHead]
        puts $dest {
       <BODY BGCOLOR=WHITE>
       <BR><BR>
       }
        puts $dest "<H1>$class sample</H1>"
        puts $dest "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>--><BR>"
        puts $dest "<IMG BORDER=0 ALIGN=CENTER SRC=../images/${class}.jpg><BR><BR>"
        puts $dest "<FONT color=#808080><PRE>\n"
        puts $dest [read $src]
        puts $dest "</PRE></BODY></HTML>"
        close $dest 
        close $src
        puts $f "<A ALIGN=RIGHT HREF=../samples/${class}.html>\
          <IMG ALIGN=RIGHT BORDER=0 SRC=../images/${class}.jpg></A>"
      } else {
        #- no code
        puts $f "<IMG BORDER=0 ALIGN=RIGHT SRC=../images/${class}.jpg>"
      }

      if [file exists [file join doc $package images ${class}.jpg]] {
        file delete [file join doc $package images ${class}.jpg]
      }
      file copy $img [file join doc $package images ${class}.jpg]
    }

    foreach type1 {specific overridden inherited} \
      text1 { Specific Overridden Inherited} {
      foreach type3 {selectors variables} {
        foreach type2 {instance class} text2 {Instance Class} {
           if {[info exists ${type1}($class,$type2,$type3)] && \
             [string length [set ${type1}($class,$type2,$type3)]]} {
             set title "$text1 $type2 $type3"
             #puts $f "<U><B>$title</B></U>"
             puts $f "<U>$title</U>"
             puts $f "<UL>"
             foreach value [set ${type1}($class,$type2,$type3)] {
               puts $f "<LI>$value"
             }
             puts $f "</UL><!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
           }
        }
      }
    }
    puts $f {
       </BODY>
      </HTML>
    }
    close $f

    #- interface & implementation sources
    foreach type {interface implementation} {
      if [catch {
        set f [open [file join doc $package classes \
          ${type}s $class.html] w]
        puts $f {
          <HTML>
        }
        puts $f [@docHtmlHead]
        puts $f {
          <BODY BGCOLOR=WHITE><BR><BR>
        }
        puts $f "<H1>$DOC(class,$type,$class)</H1>"
        puts $f "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
        puts $f "<FONT color=#808080><PRE>\n"
        set ff [open $DOC(class,$type,$class) r]
        set txt [read $ff]
        close $ff
        #- replacing < > with &lt; &gt;
        foreach {src dest} {< {\&lt;} > {\&gt;}} {
          regsub -all -- $src $txt $dest txt
        }
        puts $f $txt
        puts $f "</PRE></FONT></BODY></HTML>"
        close $f
      } err] {
        puts stderr "$err" ; return
      }
    }
  }

  foreach protocol $RUNTIME(protocols) {
    set package [set @@${protocol}(package)]

    #- protocols/<protocol>.html
    if [catch {
      set f [open [file join doc $package protocols $protocol.html] w]
      puts $f {
        <HTML>
      }
      puts $f [@docHtmlHead]
      puts $f {
        <BODY BGCOLOR=WHITE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR><BR>
      }
      puts $f "<H1>$protocol protocol</H1>\n\
        <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->\n\
        <TABLE BORDER=0>\n"
      if [string length [set inheritedFromProtocol($protocol)]] {
        puts $f "<TR>\n\
          <TD ALIGN=RIGHT>inherited from:</TD>\n\
          <TD ALIGN=LEFT>[set inheritedFromProtocol($protocol)]</TD>\n\
          </TR>"
      }

    if [string length [set @@${protocol}(protocols)]] {
      puts $f "<TR><TD ALIGN=RIGHT>protocols:</TD><TD ALIGN=LEFT>"
      set h ""
      foreach p [set @@${protocol}(protocols)] {
        if {![info exists @@${p}(package)]} { upvar #0 @@${p} @@${p} }
        append h \
          "<A HREF=../../[set @@${p}(package)]/protocols/${p}.html>$p</A>, "
      }
      puts $f "[string trimright $h {, }]</TD></TR>"
    }

    puts $f "\
       <TR>\n\
         <TD ALIGN=RIGHT>package:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=../../[set @@${protocol}(package)].html TARGET=left>\
             [set @@${protocol}(package)]</A></TD>\n\
       </TR>\n\
    "

      puts $f "<TR><TD></TD><TD></TD></TR>"
      puts $f "<TR>\n\
          <TD ALIGN=RIGHT>declared in:</TD>\n\
          <TD ALIGN=LEFT>\
            <A HREF=sources/$protocol.html>$DOC(protocol,$protocol)</A>\n\
           </TD>\n\
          </TR>\n"


      puts $f "\
         </TABLE><BR><BR>\n\
        <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"

      foreach type1 {specific inherited} text1 { {} Inherited} {
        foreach type3 {selectors} {
          foreach type2 {instance class} text2 {Instance Class} {
            if {[info exists ${type1}(protocol,$protocol,$type2,$type3)] && \
               [string length \
                 [set ${type1}(protocol,$protocol,$type2,$type3)]]} {
              if [string length $text1] {
                set title "$text1 $type2 $type3"
              } else {
                set title "$text2 $type3"
              }
              #puts $f "<U><B>$title</B></U>"
              puts $f "<U>$title</U>"
              puts $f "<UL>"
              foreach value [set ${type1}(protocol,$protocol,$type2,$type3)] {
                puts $f "<LI>$value"
              }
              puts $f "</UL><!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
            }
          }
        }
      }
      puts $f {
         </BODY>
        </HTML>
      }
      close $f
    } err] {
      puts stderr "$err" ; return
    }

    #- protocol source
    if [catch {
      set f [open [file join doc $package \
        protocols sources $protocol.html] w]
        puts $f "<HTML><BODY BGCOLOR=WHITE><BR><BR>"
        puts $f "<H1>$DOC(protocol,$protocol)</H1>"
        puts $f "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
        puts $f "<FONT color=#808080><PRE>\n"
        set ff [open $DOC(protocol,$protocol) r]
        set txt [read $ff]
        close $ff
        #- replacing < > with &lt; &gt;
        foreach {src dest} {< {\&lt;} > {\&gt;}} {
          regsub -all -- $src $txt $dest txt
        }
        puts $f $txt
        puts $f "</PRE></FONT></BODY></HTML>"
        close $f
    } err] {
      puts stderr "$err" ; return
    }
  }
}

#- new version -- taking HtmlPage into account + new menu design -- IN PROGRESS
proc @makedoc.new { {args {}} } {
  upvar #0 @RUNTIME RUNTIME

  #- loading package if any specified
  set packages $args
  if [string length $packages] {
    @package $packages
  }

  #- init
  set docdir doc.new
 
  #- directories and images
  if {![file exists $docdir]} { file mkdir $docdir }
  foreach img {samples grey.jpg} {
    if [file exists [set src [file join $RUNTIME(directory) images grey.jpg]]] {
      catch {file copy $src [file join $docdir $img]}
    }
  }

  #- main navigation bar
  set navigation [@docNavigationBar {
    packages packages.html packages 
    classes classes.html classes
    protocols protocols.html classes
    samples samples.html classes 
    screenshots screenshots.html classes
  }]

  #- main index (navigation + package + classes + center)
  set f [open [file join $docdir index.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<FRAMESET BORDERCOLOR=BLACK ROWS=\"80,*\" \
    FRAMEBORDER=1 FRAMESPACING=1 BORDER=1>"
  puts $f "<FRAME SRC=navigation.html BORDERCOLOR=BLACK \
    NAME=navigation MARGINWIDTH=4 MARGINHEIGHT=1 BORDER=1 FRAMESPACING=1>"
  puts $f "<FRAMESET BORDERCOLOR=BLACK COLS=\"220,*\" \
    FRAMEBORDER=1 FRAMESPACING=1 BORDER=1>"
  puts $f "<FRAMESET BORDERCOLOR=BLACK ROWS=\"20%,*\" \
    FRAMEBORDER=1 FRAMESPACING=1 BORDER=1>"
  puts $f "<FRAME SRC=packages.html BORDERCOLOR=BLACK NAME=packages \
    BORDER=1 FRAMESPACING=1>"
  puts $f "<FRAME SRC=classes.html BORDERCOLOR=BLACK NAME=classes \
    BORDER=1 FRAMESPACING=1>"
  puts $f "</FRAMESET>"
  puts $f "<FRAME SRC=blank.html BORDERCOLOR=BLACK NAME=center \
    BORDER=1 FRAMESPACING=1 MARGINWIDTH=40 MARGINHEIGHT=40>"
  puts $f "</FRAMESET>"
  puts $f "</FRAMESET>"
  puts $f "</HTML>"
  close $f

  #- navigation page
  set f [open [file join $docdir navigation.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#d5e5e5>"
  #puts $f "<BODY BGCOLOR=white>"
  #puts $f "<U>myObjectiveTcl</U>"
  puts $f "<BR><BR>"
  puts $f $navigation
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f

  #- center page
  set f [open [file join $docdir blank.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#d5e5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f

  #- packages page
  if {![file exists [set dir [file join $docdir]]]} {
    file mkdir $dir
  }
  set f [open [file join $docdir packages.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#d5d5e5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U>Packages</U>"
  puts $f "<UL>"
  foreach package [lsort $RUNTIME(packages)] {
    puts $f "<LI><A HREF=${package}.html TARGET=classes>$package</A>"
    if {![file exists [set dir [file join $docdir $package]]]} {
      file mkdir $dir
    }
  }
  puts $f "</UL>"
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f

  #- packages
  foreach package [lsort $RUNTIME(packages)] {
    foreach dir {classes interfaces samples protocols images} {
      catch {file mkdir [file join $docdir $package $dir]}
    }
    catch {file mkdir [file join $docdir $package protocols sources]}
    @docPackagePage $docdir $package
  }

  #- classes 
  set f [open [file join $docdir classes.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#e5d5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U>Classes</U>"
  puts $f "<UL>"

  set samples {}
  set screenshots {}

  foreach class [lsort $RUNTIME(classes)] {
    if {![info exists @$class]} { upvar #0 @$class @$class }
    set package [set @${class}(package)]
    if [string length [set @${class}(superclass)]] {
      puts $f "<LI><A HREF=$package/classes/$class.html TARGET=center>$class</A>"
    } else {
      puts $f "<LI><A HREF=$package/classes/$class.html TARGET=center><B>$class</B></A>"
    }
    @docClassPage $docdir $class
    #- sample ?
    if [file exists \
      [file join $RUNTIME(directory) $package samples ${class}.tcl]] {
      lappend samples $class
    }
    #- screenshots ?
    if [string length [set file [glob -nocomplain \
      [file join $RUNTIME(directory) $package screenshots ${class}.*]]]] {
      lappend screenshots $class
    }
  }

  puts $f "</UL>"
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f

  #- protocols
  set f [open [file join $docdir protocols.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#e5d5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U>Protocols</U>"
  puts $f "<UL>"
  foreach protocol [lsort $RUNTIME(protocols)] {
    if {![info exists @@$protocol]} { upvar #0 @@$protocol @@$protocol }
    set package [set @@${protocol}(package)]
    puts $f "<LI><A HREF=$package/protocols/$protocol.html \
      TARGET=center> $protocol</A>"
    @docProtocolPage $docdir $protocol
  }
  puts $f "</UL>"
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f


  #- categories

  #- samples
  set f [open [file join $docdir samples.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#e5d5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U>Samples</U>"
  puts $f "<UL>"
  foreach class [lsort $samples] {
    if {![info exists @${class}(package)]} { upvar #0 @$class @$class }
    set package [set @${class}(package)]
    puts $f "<LI><A HREF=$package/samples/$class.html TARGET=center>$class</A>"
  }
  puts $f "</UL>"
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f

  #- screenshots
  set f [open [file join $docdir screenshots.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#e5d5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U>Screenshots</U>"
  puts $f "<UL>"
  foreach class [lsort $screenshots] {
    if {![info exists @${class}(package)]} { upvar #0 @$class @$class }
    set package [set @${class}(package)]
    puts $f "<LI><A HREF=$package/samples/$class.html TARGET=center>"
    puts $f "$class</A>"
  }
  puts $f "</UL>"
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f

}

proc @superClassesOf {class} {
  upvar #0 @$class @$class
  if {![info exists @${class}(superclass)]} { return }
  set classes {}
  while 1 {
    if {[info exists @${class}(superclass)] && \
      [string length [set superclass [set @${class}(superclass)]]]} {
      lappend classes $superclass
      upvar #0 @$superclass @$superclass
      set class $superclass
    } else {
      break
    }
  }
  return $classes
}

proc @docClassPage {docdir class} {
  upvar #0 @RUNTIME RUNTIME
  upvar #0 @DOC DOC
  upvar #0 @$class @$class
  set package [set @${class}(package)]
  set f [open [file join $docdir $package classes $class.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#d5e5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U><B>$class</B> class</U><BR><BR>"

  #- identification
  puts $f "<TABLE>"
  if [string length [set classes [@superClassesOf $class]]] {
    set html {}
    puts $f "<TD ALIGN=RIGHT>superclasses:</TD>"
    puts $f "<TD ALIGN=LEFT>"
    foreach c $classes { 
      if {![info exists @${c}(package)]} { upvar #0 @$c @$c }
      append html "<A HREF=../../[set @${c}(package)]/classes/$c.html>$c</A> , "
    }
    puts $f [string trimright $html {, }]
    puts $f "</TD>"
    puts $f "<TR>"
  }
  puts $f "<TD ALIGN=RIGHT>declared in:</TD>"
  puts $f "<TD ALIGN=LEFT>"
  puts $f "<A HREF=../interfaces/$class.html>$DOC(class,interface,$class)</A>"
  @docTxt2Html $DOC(class,interface,$class) \
    [file join $docdir $package interfaces $class.html]
  puts $f "</TD>"
  puts $f "</TR><TR>"
  puts $f "<TD ALIGN=RIGHT>package:</TD>"
  puts $f "<TD ALIGN=LEFT>"
  puts $f "<A HREF=../../$package.html TARGET=classes>$package</A>"
  puts $f "</TD>"
  if [string length [set @${class}(protocols)]] {
    puts $f "<TR>"
    puts $f "<TD ALIGN=RIGHT>protocols:</TD>"
    puts $f "<TD ALIGN=LEFT>"
    set txt ""
    foreach protocol [set @${class}(protocols)] {
      if {![info exists @@${protocol}(package)]} {
        upvar #0 @@$protocol @@$protocol
      }
      append txt "<A HREF=../../[set @@${protocol}(package)]/protocols/"
      append txt "${protocol}.html>$protocol</A>"
      append txt " , "
    }
    puts $f [string trimright $txt { ,}]
    puts $f "</TD>"
    puts $f "</TR>"
  }
  puts $f "</TR>"

  puts $f "</TABLE><BR>"

  foreach t1 {instance class} {
    foreach t2 {selectors variables} {
      set superclass [set @${class}(superclass)]
      if {![info exists @${superclass}($t1,$t2)]} { 
        upvar #0 @$superclass @$superclass
      }
      if [string length $superclass] {
        set superclass_${t1}_${t2} [set @${superclass}($t1,$t2)]
      }
      foreach v [set @${class}($t1,$t2)] {
        #- if in superclass 
        if {[string length $superclass] && \
          [superclass_${t1}_${t2} contains: $v]} {
          #- overridden ?
          if {"$t2" == "selectors"} {
            if [string length [info proc @${class},$t1,$v]] {
              lappend overridden_${t1}_${t2} $v
            } else {
              lappend inherited_${t1}_${t2} $v
            }
          } else {
            lappend inherited_${t1}_${t2} $v
          }
        } else {
          #- class specific
          lappend specific_${t1}_${t2} $v
        }
      }
    }
  }
 
  #- sample ?
  set sample ""
  foreach dir $RUNTIME(path) {
    if [file exists [file join $dir $package samples ${class}.tcl]] {
      set sample [file join $dir $package samples ${class}.tcl]
    }
  }
  if [string length $sample] {
    set link ""
    foreach dir $RUNTIME(path) {
      if [string length [set img [glob -nocomplain \
        [file join $dir $package screenshots ${class}.*]]]] {
        set ext [file extension $img]
        catch {file copy $img [file join $docdir $package images ${class}$ext]}
        set link "<A ALIGN=RIGHT HREF=../samples/${class}.html>\
          <IMG ALIGN=RIGHT MARGINWIDTH=10 BORDER=0 \
          SRC=../images/${class}$ext></A>"
        set img ../images/${class}$ext
        break
      }
    }
    if {![string length $link]} {
      set link "<TABLE WIDTH=100% ALIGN=RIGHT><TR><TD WIDTH=80%></TD><TD>\
      <A ALIGN=RIGHT HREF=../samples/${class}.html><I>see sample</I></A>\
      </TD></TR></TABLE>"
    }
    puts $f $link
    @docSamplePage $docdir $class $sample $img
  }

  set html {} 
  foreach t1 {specific overridden inherited} {
    foreach t2 {instance class} {
      foreach t3 {selectors variables} {
        if {[info exists ${t1}_${t2}_${t3}] && \
          [string length [set ${t1}_${t2}_${t3}]]} {
          lappend links ${t1}_${t2}_${t3}
          puts $f "<A NAME=#${t1}_${t2}_${t3}>&nbsp;</A>"
          puts $f "<U>$t1 $t2 $t3</U><UL>"
          foreach v [lsort [set ${t1}_${t2}_${t3}]] {
            puts $f "<LI>$v"
          }
          puts $f "</UL>"
        }
      }
    }
  }

  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f
}

proc @docPackagePage {docdir package} {
  upvar #0 @RUNTIME RUNTIME
  upvar #0 @DOC DOC

  set f [open [file join $docdir $package.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#e5d5d5>"
  #puts $f "<BODY BGCOLOR=white>"

  #- classes
  set classes {}
  foreach class $RUNTIME(classes) {
    upvar #0 @$class @$class
    if {"$package" == "[set @${class}(package)]"} {
      lappend classes $class
    }
  }
  if [llength $classes] {
    puts $f "<U><B>$package</B> classes</U><BR>"
    puts $f "<UL>"
    foreach class [lsort $classes] {
      if [string length [set @${class}(superclass)]] {
        puts $f "<LI><A HREF=$package/classes/$class.html TARGET=center>$class</A>"
      } else {
        puts $f "<LI><A HREF=$package/classes/$class.html TARGET=center><B>$class</B></A>"
      }
    }
    puts $f "</UL>"
  }

  #- protocols
  set protocols {}
  foreach protocol $RUNTIME(protocols) {
    upvar #0 @@$protocol @@$protocol
    if {"$package" == "[set @@${protocol}(package)]"} {
      lappend protocols $protocol
    }
  }
  if [llength $protocols] {
    puts $f "<U><B>$package</B> protocols</U><BR>"
    puts $f "<UL>"
    foreach protocol $protocols {
      puts $f "<LI><A HREF=$package/protocols/$protocol.html TARGET=center>"
      puts $f "$protocol</A>"
    }
    puts $f "</UL>"
  }

  #- categories

  puts $f "</BODY></HTML>"
  close $f
}

proc @docProtocolPage {docdir protocol} {
  upvar #0 @DOC DOC
  upvar #0 @@$protocol @@$protocol
  set package [set @@${protocol}(package)]
  set f [open [file join $docdir $package protocols $protocol.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#d5e5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U><B>$protocol</B> protocol</U><BR><BR>"

  #- identification
  puts $f "<TABLE>"
  puts $f "<TD ALIGN=RIGHT>declared in:</TD>"
  puts $f "<TD ALIGN=LEFT>"
  puts $f "<A HREF=sources/$protocol.html>$DOC(protocol,$protocol)</A>"
  @docTxt2Html $DOC(protocol,$protocol) \
    [file join $docdir $package protocols sources $protocol.html]
  puts $f "</TD>"
  puts $f "</TR><TR>"
  puts $f "<TD ALIGN=RIGHT>package:</TD>"
  puts $f "<TD ALIGN=LEFT>"
  puts $f "<A HREF=../../$package.html TARGET=classes>$package</A>"
  puts $f "</TD>"
  if [string length [set @@${protocol}(protocols)]] {
    puts $f "<TR>"
    puts $f "<TD ALIGN=RIGHT>protocols:</TD>"
    puts $f "<TD ALIGN=LEFT>"
    set txt ""
    foreach protocol [set @@${protocol}(protocols)] {
      append txt "<A HREF=protocols/sources/$protocol.html>"
      append txt "$DOC(protocol,$protocol)</A>"
      append txt " , "
    }
    puts $f [string trimright $txt { ,}]
    puts $f "</TD>"
    puts $f "</TR>"
  }
  puts $f "</TR>"

  puts $f "</TABLE><BR>"

  foreach t1 {instance class} {
    foreach t2 {selectors} {
      foreach v [set @@${protocol}($t1,$t2)] {
          #- protocol specific
          lappend specific_${t1}_${t2} $v
      }
    }
  }
  
  set html {} 
  foreach t1 {specific inherited} {
    foreach t2 {instance class} {
      foreach t3 {selectors} {
        if {[info exists ${t1}_${t2}_${t3}] && \
          [string length [set ${t1}_${t2}_${t3}]]} {
          lappend links ${t1}_${t2}_${t3}
          puts $f "<A NAME=#${t1}_${t2}_${t3}>&nbsp;</A>"
          puts $f "<U>$t1 $t2 $t3</U><UL>"
          foreach v [lsort [set ${t1}_${t2}_${t3}]] {
            puts $f "<LI>$v"
          }
          puts $f "</UL>"
        }
      }
    }
  }

  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f
}
proc @docNavigationBar {links} {
  set separator "<TR><TD COLSPAN=[expr [llength $links] / 3]>"
  append separator "<IMG SRC=grey.jpg WIDTH=100% HEIGHT=1 BORDER=0></TD></TR>"
  set navigation "<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0 WIDTH=100%>"
  append navigation $separator
  append navigation "<TR>"
  foreach {txt lnk tgt} $links {
    append navigation "<TD><A HREF=${lnk} TARGET=$tgt>$txt</A></TD>"
  }
  append navigation "</TR>"
  append navigation $separator
  append navigation "</TABLE>"
  return $navigation
}

proc @docHtmlHead {} {
  return {
    <HEAD>
     <TITLE>myObjectiveTcl</TITLE>
      <META NAME="Description" CONTENT="Objective-C like framework in pure tcl">
      <META NAME="Keywords" 
        CONTENT="myObjectiveTcl Objective-C tcl tk object-oriented OO">
     <STYLE>
      BODY, TABLE, TR, TD, A:link, A:visited {
       font-size: 13 ;
       font-weight: normal ;
       font-family: "arial, helvetica" ;
       text-decoration: none ;
      }
      PRE {
        font-size: 12 ;
        color: #555555 ;
      }
      A:link, A:visited {
        text-decoration: none ;
        color: #8080B0
      }
      A:hover, A:active {
       text-decoration: underline ;
       color: #8080B0
      }
      LI {
       list-style-type: circle ;
      }
     </STYLE>
    </HEAD>
  }
}

proc @docTxt2Html {src dest} {
  set f [open $dest w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#d5e5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<PRE>"
  set txt [read [set ff [open $src r]]]
  close $ff
  regsub -all -- {<} $txt {\&lt;} txt
  regsub -all -- {>} $txt {\&gt;} txt
  puts $f $txt
  puts $f "</PRE>"
  puts $f "</BODY>"
  puts $f "</HTML>"
  close $f
}

proc @docSamplePage {docdir class src img} {
  upvar #0 @RUNTIME RUNTIME
  upvar #0 @$class @$class
 
  set package [set @${class}(package)]
 
  set f [open [file join $docdir $package samples ${class}.html] w]
  puts $f "<HTML>"
  puts $f [@docHtmlHead]
  puts $f "<BODY BGCOLOR=#d5e5d5>"
  #puts $f "<BODY BGCOLOR=white>"
  puts $f "<U><B>$class</B> sample</U><BR>"
  if [string length $img] {
    puts $f "<BR><IMG BORDER=0 BORDERCOLOR=BLACK SRC=$img><BR>"
  }
  puts $f "<PRE>"
  set ff [open $src r]
  puts $f [read $ff]
  close $ff
  puts $f "</PRE>"
  puts $f "</BODY></HTML>"
  close $f
}

@path {core gui}
