
if {![info exists @RUNTIME(classes)]} {
  set @RUNTIME(classes) {}
  set @RUNTIME(categories) {}
  set @RUNTIME(protocols) {}
}

proc @getPathFromClass class {
  if [llength [file split $class]] {
    set exp [file tail $class]
    if [regexp [file join (.*) [file tail $class]] $class -> path] {
      set path
    } 
  }
}

proc @getClassFromClass class {
  set class [lindex [split [file tail $class] .] end]
}

#----------------------------
# loads a class into runtime
#----------------------------
proc @class {class} {
  upvar #0 @RUNTIME RUNTIME

  set code 0

  #- don't allow class name beginning with a @ character
  if [regexp {^@} $class] { return $code }

  set path [@getPathFromClass $class]
  set package [@getPackageFromClass $class]
  set class [@getClassFromClass $class]

  if [RUNTIME(classes) contains: ${class}] {
    return 1
  } elseif [RUNTIME(protocols) contains: ${class}] {
    return 1
  }

  #- if directory is specified, adding it to paths list if not yet in
  if {[string length $path] && ![RUNTIME(path) contains: $path]} {
    @path $path
  }

  if {[string length $package]} {
    @package $package 
  } else {
    if {[llength $RUNTIME(packages)] == 0} {
      lappend packages {}
    }
  }
  set paths {}
 
  foreach path $RUNTIME(path) {
    foreach p $RUNTIME(packages) {
      lappend paths [file join $RUNTIME(directory) $p]
      lappend paths [file join $RUNTIME(directory) $path $p]
      lappend paths [file join $path $p]
      set pkg([file join $RUNTIME(directory) $p]) $p
      set pkg([file join $RUNTIME(directory) $path $p]) $p
      set pkg([file join $path $p]) $p
    }
  }

  set files {}

  foreach path $paths {
    set _pkg [set pkg($path)]
    set i 0
    foreach v {interface implementation} {
      set $v [file join $path $v ${class}.tcl]
      if [file exists [set $v]] {
        if [catch {source [set $v]} err] {
          return [error $err]
        } else {
          incr i
        }
      }
    }
    if {$i == 0} { 
      #- no source found in interface/implementation folders 
      #- -> can be in only one file and outside
      #- the interface/implementation folders
      if [file exists [set f [file join $path ${class}.tcl]]] {
        if [catch {source $f} err] {
          return [error $err]
        } else {
          incr i
        }
      }
    }
    if {$i} {
      upvar #0 @$class CLASS
      if {![string length $CLASS(package)]} {
        set CLASS(package) $_pkg
      } else {
        set _pkg $CLASS(package)
      }
      if [string length $CLASS(implementation)] {
        interp alias "" "${_pkg}.$class" "" $class
      } else {
        #RUNTIME(protocols) addIfNotYetIn: $class
      } 
      return 1
    }
  }

  return $code
}

