

#------------------------------------------------------------------------------
# build the doc from the current runtime
#------------------------------------------------------------------------------
proc @makedoc {} {
  upvar #0 @RUNTIME RUNTIME

  #- creating directories
  if [catch {
    file mkdir doc
  } err] {
    puts stderr "$err" ; return
  }

  #- index.html - frameset
  if [catch {
    set f [open [file join doc index.html] w]
    puts $f {
      <HTML>
      <HEAD>
      <TITLE>myObjectiveTcl</TITLE>
      <META NAME="Description" CONTENT="Objective-C like framework in pure tcl">
      <META NAME="Keywords" 
        CONTENT="myObjectiveTcl Objective-C tcl tk object-oriented OO">
      </HEAD>
       <FRAMESET COLS="300,*" FRAMEBORDER=1 FRAMESPACING=1 BORDER=1 
         BORDERCOLOR=BLACK>
        <FRAME FRAMEBORDER=1 AUTOSCROLLING=1 SRC=packages.html NAME=left
          MARGINWIDTH=10 MARGINHEIGHT=10>
        <FRAME FRAMEBORDER=0 AUTOSCROLLING=1 SRC=main.html NAME=right
          MARGINWIDTH=40 MARGINHEIGHT=10>
       </FRAMESET>
      </HTML>
    }
    close $f
  } err] {
    puts stderr "$err" ; return
  }

  #- main.html - empty file for first
  set f [open [file join doc main.html] w]
  puts $f "<HTML><BODY BGCOLOR=WHITE></BODY></HTML>"
  close $f

  #- classes.html - empty file for first
  set f [open [file join doc classes.html] w]
  puts $f "<HTML><BODY BGCOLOR=WHITE></BODY></HTML>"
  close $f

  #- collecting informations about classes and sorting them
  foreach class $RUNTIME(classes) {
    upvar #0 @$class @$class

    #- sort classes into packages
    lappend classes([set @${class}(package)]) $class

    #- sort selectors/variables into class specific, overridden, inherited
    foreach type1 {instance class} {
      foreach type2 {selectors variables} {
        set field ${type1},${type2}
        set value [set @${class}($field)]
        if {![string length [set superclass [set @${class}(superclass)]]]} {
          #- no superclass -> all is specific
          set specific($class,$field) [set @${class}($field)]
        } else {
          foreach subvalue $value { 
            #- doesn't exists in superclass -> class specific
            if {![@${superclass}($field) contains: $subvalue]} {
              lappend specific($class,$field) $subvalue
            } elseif [regexp variables $field] {
              #- exist in superclass -> inherited for variables
              lappend inherited($class,$field) $subvalue
            } elseif [string length [info proc @$class,$type1,$subvalue]] {
              #- the selector is overridden by the current class
              lappend overridden($class,$field) $subvalue
            } else {
              #- just simple inherited
              lappend inherited($class,$field) $subvalue
            }
          }
        }
      }
    }
    
    #- going deeply in the class hierarchy    
    if {0} {
    set inheritedFrom($class) ""
    set superExists 1
    set super $class
    while {$superExists} {
      if [string length [set super [set @${super}(superclass)]]] {
        lappend inheritedFrom($class) $super
        if {![info exists @${super}]} {
          upvar #0 @${super} @${super}
        }
      } else {
        set superExists 0
        break
      }
    }
    }

    set inheritedFrom($class) [set @${class}(superclass)]
    #- multiple inheritence
    if [string length @${class}(superclasses)] {
      foreach c [set @${class}(superclasses)] {
        lappend inheritedFrom($class) $c
      }
    }

    if [string length [set inheritedFrom($class)]] {
      set i 0
      foreach c [set inheritedFrom($class)] {
        set pkg [set @${c}(package)] 
        if {"[set @${class}(package)]" == "."} { set pkg $pkg }

        if {$i} {
          append inheritedFromInHtml($class) \
            "\n,&nbsp;<A HREF=../../$pkg/classes/$c.html>$c</A>"
        } else {
          append inheritedFromInHtml($class) \
            "\n<A HREF=../../$pkg/classes/$c.html>$c</A>" 
        }
        incr i
      }
      set inheritedFrom($class) [set inheritedFromInHtml($class)]
      unset inheritedFromInHtml($class)
    }
  }

  #- collecting informations about protocols
  foreach proto $RUNTIME(protocols) {
    upvar #0 @$proto @$proto

    #- sort classes into packages
    lappend protocols([set @${proto}(package)]) $proto

    #- sort selectors into protocol specific and inherited
    foreach type1 {instance class} {
     foreach type2 {selectors} {
        set field ${type1},${type2}
        set value [set @${proto}(interface,$field)]

        if {![string length [set parents [set @${proto}(protocols)]]]} {
          #- no parent protocols -> all is specific
          set specific(protocol,$proto,$field) [set @${proto}(interface,$field)]
        } else {
          #- parent protocols exists, checking if there is inheritence
          foreach subvalue $value {
            set count 0
          
            foreach parent $parents {
              #- doesn't exists in superclass -> class specific
              if [@${parent}($field) contains: $subvalue] {
                incr count
              }
            }
            if { $count > 0 } {
              lappend inherited(protocol,$proto,$field) $subvalue
            } else {
              lappend specific(protocol,$proto,$field) $subvalue
            } 
          }
        }
      }
    }
    set inheritedFromProtocol($proto) [set @${proto}(protocols)]
    if [string length $inheritedFromProtocol($proto)] {
      set i 0
      foreach p $inheritedFromProtocol($proto) {
        set pkg [set @${p}(package)]
        if {"[set @${proto}(package)]" == "."} { set pkg $pkg }
        if {$i} {
          append inheritedFromInHtml($proto) \
            ",&nbsp;<A HREF=../../$pkg/protocols/$p.html>$p</A>"
        } else {
          append inheritedFromInHtml($proto) \
            "<A HREF=../../$pkg/protocols/$p.html>$p</A>"
        }
        incr i
      }
      set inheritedFromProtocol($proto) [set inheritedFromInHtml($proto)]
      unset inheritedFromInHtml($proto)
    }
   
  }
 
  #- packages.html - navigation throught packages
  set fd(left) [open [file join doc packages.html] w]
  puts $fd(left) {<HTML>}
  puts $fd(left) [@docHtmlHead]
  puts $fd(left) {
     <BODY BGCOLOR=#DEDEDE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR>
       <H1>myObjectiveTcl</H1><UL>
  }

  #- opening all packages file descriptor
  foreach package [set packages [lsort [array names classes]]] {
    #- creating package file 
    if [catch {
      file mkdir [file join doc $package]
      set fd($package) [open [file join doc $package index.html] w]
    } err] {
      puts stderr "$err" ; return
    }
    puts $fd($package) {<HTML>}
    puts $fd($package) [@docHtmlHead]
    puts $fd($package) {
     <BODY BGCOLOR=#DEDEDE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR>
       <H1>myObjectiveTcl</H1><UL>
    }
  }

  #- handlig html for each package
  foreach package $packages {
    #- adding link to main packages index
    puts $fd(left) "<LI><A HREF=$package/index.html>$package</A>"

    #- adding link to others package pages
    foreach p $packages {
      #- not for the current one
      if {"$p" != "$package"} {
        puts $fd($p) "<LI><A HREF=../$package/index.html>$package</A>"
      }
    }

    puts $fd($package) "<LI><A HREF=../packages.html>$package</A><UL>"
  
    if [catch {
      file mkdir [file join doc $package classes]
      file mkdir [file join doc $package classes interfaces]
      file mkdir [file join doc $package classes implementations]
      file mkdir [file join doc $package protocols]
      file mkdir [file join doc $package protocols sources]
      file mkdir [file join doc $package categories]
      file mkdir [file join doc $package categories interfaces]
      file mkdir [file join doc $package categories implementations]
      file mkdir [file join doc $package images]
      file mkdir [file join doc $package samples]
    } err] {
      puts stderr "$err" ; return
    }
    #- classes
    if {[info exists classes($package)] && [string length $classes($package)]} {
      puts $fd($package) "<LI>Classes"
      puts $fd($package) "<UL>"
      foreach class [lsort $classes($package)] {
        if {![string length [set @${class}(superclass)]]} {
          #- root class because of no superclass -> in bold
          puts $fd($package) "<LI>\
            <A HREF=classes/$class.html TARGET=right>\
            <B>$class</B></A>"
        } else {
          puts $fd($package) "<LI>\
            <A HREF=classes/$class.html TARGET=right>\
            $class</A>"
        }
      }
      puts $fd($package) "</UL>"
    }

    #- protocols
    if {[info exists protocols($package)] && \
      [string length $protocols($package)]} {
      puts $fd($package) "<LI>Protocols"
      puts $fd($package) "<UL>"
      foreach protocol [lsort $protocols($package)] {
        puts $fd($package) "<LI>\
          <A HREF=protocols/$protocol.html TARGET=right>\
          $protocol</A>"
      }
      puts $fd($package) "</UL>"
    }
    puts $fd($package) "</UL>"
  }
  puts $fd(left) "</UL>"

  puts $fd(left) {
     </BODY>
    </HTML>
  }
  close $fd(left)

  #- closing all packages
  foreach p $packages {
    puts $fd($p) "</UL></BODY></HTML>"
    close $fd($p)
  }


  #- classes/<class>.html 
  foreach class $RUNTIME(classes) {
    set package [set @${class}(package)]
    set f [open [file join doc $package classes $class.html] w]
    puts $f <HTML>
    puts $f [@docHtmlHead]
    puts $f {
       <BODY BGCOLOR=WHITE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR><BR>
    }
    puts $f "<H1>$class class</H1>\n\
      <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->\n\
      <TABLE BORDER=0>\n"
    if [string length [set inheritedFrom($class)]] {
      puts $f "<TR>\n\
         <TD ALIGN=RIGHT>inherited from: </TD>\n\
         <TD ALIGN=LEFT>[set inheritedFrom($class)]</TD>\n\
       </TR>"
    }

    if [string length [set @${class}(protocols)]] {
      puts $f "<TR><TD ALIGN=RIGHT>protocols:</TD><TD ALIGN=LEFT>"
      set h ""
      foreach p [set @${class}(protocols)] {
        if {![info exists @${p}(package)]} { upvar #0 @${p} @${p} }
        append h \
          "<A HREF=../../[set @${p}(package)]/protocols/${p}.html>$p</A>, "
      }
      puts $f "[string trimright $h {, }]</TD></TR>"
    }

    puts $f "\
       <TR>\n\
         <TD ALIGN=RIGHT>package:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=../../[set @${class}(package)]/index.html TARGET=left>\
             [set @${class}(package)]</A></TD>\n\
       </TR>\n\
    "

    puts $f "<TR><TD></TD><TD></TD></TR>"

    puts $f "<TR>\n\
         <TD ALIGN=RIGHT>declared in:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=interfaces/$class.html>[set @${class}(interface)]</A>\n\
         </TD>\n\
       </TR>\n\
       <TR>\n\
         <TD ALIGN=RIGHT>implemented in:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=implementations/$class.html>\
             [set @${class}(implementation)]</A></TD>\n\
       </TR>\n"

    puts $f "\
      </TABLE><BR><BR>\n\
      <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"

   #- sample code ?
   if [file exists [set sample \
       [file join $RUNTIME(directory) $package samples ${class}.tcl]]] {
      #- sample code ?
        set dest [open [file join doc $package samples ${class}.html] w]
        set src [open $sample r]
        puts $dest {
          <HTML>
        }
        puts $dest [@docHtmlHead]
        puts $dest {
          <BODY BGCOLOR=WHITE><BR><BR>
        }
        puts $dest "<H1>$class sample</H1>"
        puts $dest "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>--><BR>"
        puts $dest "<FONT color=#808080><PRE>\n"
        puts $dest [read $src]
        puts $dest "</PRE></BODY></HTML>"
        close $dest
        close $src
        if {![file exists [file join doc $package images sample.jpg]]} {
           file copy [file join $RUNTIME(directory) images sample.jpg] \
             [file join doc $package images sample.jpg]
        }
        puts $f "<A ALIGN=RIGHT HREF=../samples/${class}.html>\
          <IMG ALIGN=RIGHT BORDER=0 SRC=../images/sample.jpg></A>"
    }

    #- screenshots ?
    if [file exists [set img [file join $RUNTIME(directory) $package screenshots ${class}.jpg]]] {
      #- code ?
      if [file exists [set sample [file join $RUNTIME(directory) $package samples ${class}.tcl]]] {
        set dest [open [file join doc $package samples ${class}.html] w]
        set src [open $sample r]
        puts $dest {
        <HTML>
        }
        puts $dest [@docHtmlHead]
        puts $dest {
       <BODY BGCOLOR=WHITE>
       <BR><BR>
       }
        puts $dest "<H1>$class sample</H1>"
        puts $dest "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>--><BR>"
        puts $dest "<IMG BORDER=0 ALIGN=CENTER SRC=../images/${class}.jpg><BR><BR>"
        puts $dest "<FONT color=#808080><PRE>\n"
        puts $dest [read $src]
        puts $dest "</PRE></BODY></HTML>"
        close $dest 
        close $src

        #- screenshot to display in the main class page to link to the sample
        puts $f "<A ALIGN=RIGHT HREF=../samples/${class}.html>"
        puts $f "<IMG VALIGN=BOTTOM ALIGN=RIGHT BORDER=0 SRC=../images/${class}.jpg></A>"

      } else {
        #- no code
        puts $f "<IMG BORDER=0 ALIGN=RIGHT SRC=../images/${class}.jpg>"
      }

      if [file exists [file join doc $package images ${class}.jpg]] {
        file delete [file join doc $package images ${class}.jpg]
      }
      file copy $img [file join doc $package images ${class}.jpg]
    }

    puts $f "<UL>"
    foreach type2 {class instance} text2 {Class Instance} {
      foreach type3 {variables selectors} {
        foreach type1 {specific overridden inherited} \
          text1 {Specific} {
          if {[info exists ${type1}($class,$type2,$type3)] && \
            [string length [set ${type1}($class,$type2,$type3)]]} {
            set title "$type2 $type3"
            if {"$type1" == "specific"} {
              puts $f "<LI><U>$title</U>"
              puts $f "<UL>"
            }
            foreach value [lsort [set ${type1}($class,$type2,$type3)]] {
              if {"$type1" == "specific"} {
                puts $f "<LI>$value"
              }
              if {[info exists all_${type3}($class)] && \
                ([lsearch [set all_${type3}($class)] $value] == -1) || ![info exists all_${type3}($class)]} {
                lappend all_${type3}($class) $value
              }
            }
            if {"$type1" == "specific"} {
              puts $f "</UL>"
            }
          }
        }
      }
    }
    puts $f "</UL>"

    #- removed all variables/methods
    if {0} {
    puts $f "<HR WIDTH=100%><BR>"
    puts $f "<UL>"
    foreach type {variables selectors} {
      if {[info exists all_${type}($class)] && [llength [set all_${type}($class)]]} {
         puts $f "<LI><U>all $type</U>"
         puts $f "<UL>"
         foreach v [lsort [set all_${type}($class)]] {
           puts $f "<LI>$v"
         }
         puts $f "</UL>"
      }
    }
    puts $f "</UL>"
    }

    puts $f {
       </BODY>
      </HTML>
    }
    close $f

    #- interface & implementation sources
    foreach type {interface implementation} {
      if [catch {
        set f [open [file join doc $package classes \
          ${type}s $class.html] w]
        puts $f {
          <HTML>
        }
        puts $f [@docHtmlHead]
        puts $f {
          <BODY BGCOLOR=WHITE><BR><BR>
        }
        puts $f "<H1>[set @${class}($type)]</H1>"
        puts $f "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
        puts $f "<FONT color=#808080><PRE>\n"
        set ff [open [set @${class}($type)] r]
        set txt [read $ff]
        close $ff
        #- replacing < > with &lt; &gt;
        foreach {src dest} {< {\&lt;} > {\&gt;}} {
          regsub -all -- $src $txt $dest txt
        }
        puts $f $txt
        puts $f "</PRE></FONT></BODY></HTML>"
        close $f
      } err] {
        puts stderr "$err" ; return
      }
    }
  }

  foreach protocol $RUNTIME(protocols) {
    set package [set @${protocol}(package)]

    #- protocols/<protocol>.html
    if [catch {
      set f [open [file join doc $package protocols $protocol.html] w]
      puts $f {
        <HTML>
      }
      puts $f [@docHtmlHead]
      puts $f {
        <BODY BGCOLOR=WHITE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR><BR>
      }
      puts $f "<H1>$protocol protocol</H1>\n\
        <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->\n\
        <TABLE BORDER=0>\n"
      if [string length [set inheritedFromProtocol($protocol)]] {
        puts $f "<TR>\n\
          <TD ALIGN=RIGHT>inherited from:</TD>\n\
          <TD ALIGN=LEFT>[set inheritedFromProtocol($protocol)]</TD>\n\
          </TR>"
      }

    if [string length [set @${protocol}(protocols)]] {
      puts $f "<TR><TD ALIGN=RIGHT>protocols:</TD><TD ALIGN=LEFT>"
      set h ""
      foreach p [set @${protocol}(protocols)] {
        if {![info exists @${p}(package)]} { upvar #0 @${p} @${p} }
        append h \
          "<A HREF=../../[set @${p}(package)]/protocols/${p}.html>$p</A>, "
      }
      puts $f "[string trimright $h {, }]</TD></TR>"
    }

    puts $f "\
       <TR>\n\
         <TD ALIGN=RIGHT>package:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=../../[set @${protocol}(package)]/index.html TARGET=left>\
             [set @${protocol}(package)]</A></TD>\n\
       </TR>\n\
    "

      puts $f "<TR><TD></TD><TD></TD></TR>"
      puts $f "<TR>\n\
          <TD ALIGN=RIGHT>declared in:</TD>\n\
          <TD ALIGN=LEFT>\
            <A HREF=sources/$protocol.html>[set @${protocol}(interface)]</A>\n\
           </TD>\n\
          </TR>\n"


      puts $f "\
         </TABLE><BR><BR>\n\
        <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"

      foreach type1 {specific inherited} text1 { {} Inherited} {
        foreach type3 {selectors} {
          foreach type2 {instance class} text2 {Instance Class} {
            if {[info exists ${type1}(protocol,$protocol,$type2,$type3)] && \
               [string length \
                 [set ${type1}(protocol,$protocol,$type2,$type3)]]} {
              if [string length $text1] {
                set title "$text1 $type2 $type3"
              } else {
                set title "$text2 $type3"
              }
              #puts $f "<U><B>$title</B></U>"
              puts $f "<U>$title</U>"
              puts $f "<UL>"
              foreach value [set ${type1}(protocol,$protocol,$type2,$type3)] {
                puts $f "<LI>$value"
              }
              puts $f "</UL><!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
            }
          }
        }
      }
      puts $f {
         </BODY>
        </HTML>
      }
      close $f
    } err] {
      puts stderr "$err" ; return
    }

    #- protocol source
    if [catch {
      set f [open [file join doc $package \
        protocols sources $protocol.html] w]
        puts $f "<HTML><BODY BGCOLOR=WHITE><BR><BR>"
        puts $f "<H1>[set @${protocol}(interface)]</H1>"
        puts $f "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
        puts $f "<FONT color=#808080><PRE>\n"
        set ff [open [set @${protocol}(interface)] r]
        set txt [read $ff]
        close $ff
        #- replacing < > with &lt; &gt;
        foreach {src dest} {< {\&lt;} > {\&gt;}} {
          regsub -all -- $src $txt $dest txt
        }
        puts $f $txt
        puts $f "</PRE></FONT></BODY></HTML>"
        close $f
    } err] {
      puts stderr "$err" ; return
    }
  }
}

proc @docHtmlHead {} {
  return {
    <HEAD>
     <TITLE>myObjectiveTcl</TITLE>
      <META NAME="Description" CONTENT="Objective-C like framework in pure tcl">
      <META NAME="Keywords"
        CONTENT="myObjectiveTcl Objective-C tcl tk object-oriented OO">
     <STYLE>
      BODY, TABLE, TR, TD, A:link, A:visited {
       font-size: 13 ;
       font-weight: normal ;
       font-family: "arial, helvetica" ;
       text-decoration: none ;
      }
      PRE {
        font-size: 12 ;
        color: #555555 ;
      }
      A:link, A:visited {
        text-decoration: none ;
        color: #8080B0
      }
      A:hover, A:active {
       text-decoration: underline ;
       color: #8080B0
      }
      LI {
       list-style-type: circle ;
      }
     </STYLE>
    </HEAD>
  }
}

