#--------------------------------------
# parser for the class implementation
#--------------------------------------
proc @implementation { {args {}} } {
  upvar #0 @RUNTIME RUNTIME

  set category {}

  #- splitting args into real args and code
  if [regexp {^([^\{]+)[\{](.+)[\}]$} $args -> realArgs code] {
    #- parsing realArgs
    if [regexp {^([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
      $realArgs -> class category] {
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*$} \
        $realArgs -> class superclass] {
    } elseif [regexp {^([^ ]+)[ ]*$} $realArgs -> class] {
    }

    set code [string trimright $code "\r\n "]

    #- removing comments from code
    set newcode ""
    foreach line [split $code \n] {
      if [regexp {^[ ]*[#](.*)$} $line] { continue }
      append newcode $line\n
    }
    set code $newcode

    if [string length $category] {
      #- importing original class if it's a category
      if {![@class $class]} {
        return [error "couldn't import original $class definition"]
      }
    }

    #- package name is specified
    if {[llength [split $class .]] > 1} {
      set package [join [lreplace [split $class .] end end] .]
      set class [lindex [split $class .] end]
    } else {
      set package {}
    }

    upvar #0 @${class} SELF

    if [string length $SELF(superclass)] {
      upvar #0 @$SELF(superclass) SUPER
    }

    foreach {char selector args body} $code {
      switch -- $char {
        - { 
          set type instance
        }
        + {
          set type class
        }
        default { return [error "unknown char:$char"] }
      }
      #- args
      if {![string length $args]} { set args "{args {}}" }
      set args [linsert $args 0 self]

      if {"$type" == "instance"} {
        append newbody [@getAccessToInstanceVariables $class]
      } else {
        append newbody [@getAccessToClassVariables $class]
      }
      append newbody $body
      set body $newbody
      unset newbody
      append body {
        set self
      }

      #- method
      proc @${class},$type,$selector $args $body

      #- adding method in class description if not yet in
      if {![SELF($type,selectors) contains: $selector]} {
        lappend SELF($type,selectors) $selector
      }
      set SELF(${type},methodFor,${selector}) \
        @${class},$type,$selector
    }

    #- check if all methods are implemented
    foreach type {class instance} {
      #- getting only selectors that are not inherited
      set notInheritedSelectors {}
      set inheritedSelectors {}
      if [info exists SUPER] {
        foreach selector $SELF($type,selectors) {
          if {![SUPER($type,selectors) contains: $selector]} {
            lappend notInheritedSelectors $selector
          } else {
            lappend inheritedSelectors $selector
          }
        }
      } else {
        set notInheritedSelectors $SELF($type,selectors)
      }
      foreach selector $notInheritedSelectors {
        if {![string length [info proc @${class},$type,$selector]]} {
          #- can be implemented in an other superclass in a case of multiple inheritence
          if [llength $SELF(superclasses)] {
            set i 0
            foreach _super $SELF(superclasses) {
              incr i
              upvar #0 @$_super SUPER${i}
              if [info exists SUPER${i}] {
                if {![SUPER${i}($type,selectors) contains: $selector]} {
                  lappend notInheritedSelectors2 $selector
                }
              } else {
                lappend notInheritedSelectors2 $selector
              }
            }
            if [info exists notInheritedSelectors2] {
              if {[llength $notInheritedSelectors2] > 1} { set verb "are" } else { set verb "is" }
              return [error "uncompleted implementation for \"$class\" class:\n\
               $notInheritedSelectors2 selectors $verb not implemented !\n"]
            }
          } else {
            return [error "uncompleted implementation for \"$class\" class:\n\
              \"$selector\" selector is not implemented !\n"]
          }
        }
      }

      # checking if inherited method points to the same proc than the parent ones
      # when they are the same 
      # else it must be overridden in the implementation
      foreach selector $inheritedSelectors {
        set field "$type,methodFor,$selector"
        if {"[set SELF($field)]" != "@[set SELF(class)],$type,$selector"} {
          #- must be the same than in parents
          if [llength $SELF(superclasses)] {
            set i 0
            foreach _parent $SELF(superclasses) {
              incr i
              if {![info exists SUPER$i]} {
                upvar #0 @$_parent SUPER$i
              }
            } 
            for {set j 1} {$j <= $i} {incr j} {
              if [info exists SUPER${i}($field)] {
                if {"[set SELF($field)]" != "[set SUPER${i}($field)]"} {
                  @debug "$selector of $type of $SELF(class)"
                }
              }
            }
          }
        }
      }

    }

    #- check if protocol methods are implemented
    if [string length $SELF(protocols)] {
      foreach protocol $SELF(protocols) {
        if {![@class $protocol]} {
          return [error "unknown protocol $protocol for class $class"]
        }
        upvar #0 @${protocol} PROTOCOL
        foreach type {instance class} {
          foreach selector $PROTOCOL(interface,${type},selectors) {
            if {![SELF(${type},selectors) contains: $selector]} {
              return [error \
                "$class class must conform to $protocol protocol and \
                \"$selector\" selector is not implemented!"]
            }
          }
        }
      }
    }

    #- creating class callback
    proc $class { {args {}} } {
      return [@sendMessageToObject [lindex [split [lindex [info level 0] 0] .] end] $args]
    }

    #- adding class name to runtime
    if {![RUNTIME(classes) contains: $class]} {
      lappend RUNTIME(classes) $class
    }
    if {[RUNTIME(protocols) contains: $class]} {
      RUNTIME(protocols) remove: $class
    }

    set SELF(implementation) [@formatLibraryPathName [info script]]

    if [string length $category] {
      if {![SELF(categories) contains: $category]} {
        lappend SELF(categories) $category
      }
    }

  } else {
    return [error "malformed implementation"]
  }

}
