#------------------------------------------------------------------------------
# parser for the class interface
#------------------------------------------------------------------------------
proc @interface { {args {}} } {
  upvar #0 @RUNTIME RUNTIME

  foreach v {
    class superclass superclasses protocols category
    class,variables class,selectors
    instance,variables instance,selectors
    interface,class,selectors interface,instance,selectors
  } {
    set $v {}
  }

  #- splitting args into realArgs and code
  if [regexp {^([^\{]+)[\{](.+)[\}]$} $args -> realArgs code] {
    #- parsing realArgs to get class, superclass, protocols, category if any
    if [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*<([^>]+)>[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class superclass protocols category] {
      #-1- class + superclass + protocols + category
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*<([^>]+)>[ ]*$} \
        $realArgs -> class superclass protocols] {
      #-2- class + superclass + protocols
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class superclass category] {
      #-3- class + superclass + category
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*$} \
        $realArgs -> class superclass] {
      #-4- class + superclass
    } elseif [regexp {^([^ ]+)[ ]*<([^>]+)>[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class protocols category] {
      #-5- class + protocols + category
    } elseif [regexp {^([^ ]+)[ ]*<([^>]+)>[ ]*$} \
        $realArgs -> class protocols] {
      #-6- class + protocols
    } elseif [regexp {^([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class category] {
      #-7- class + category
    } elseif [regexp {^([^ ]+)[ ]*$} $realArgs -> class] {
      #-8- class
    } else {
      return [error "malformed interface"]
    }
  } elseif [regexp {^([^\{]*)[\{](.*)[\}]$} $args] {
    #- empty interface
  } else {
    return [error "malformed interface"]
  }

  set code [string trimright $code "\r\n "]

  #- removing comments from code
  set newcode ""
  foreach line [split $code \n] {
    if [regexp {^[ ]*[#](.*)$} $line] { continue }
    append newcode $line\n
  }
  set code $newcode

  #- multiple inheritence
  if [regexp , $superclass] {
    set superclasses [lreplace [split $superclass ,] 0 0]
    set superclass [lindex [split $superclass ,] 0]
  }

  #- parsing code and setting variables
  if [regexp {[\{][\r\n ]*(.*)[\}](.*)} $code -> variables code] {
    set variables [string trimright $variables "\n\r "]
    set instance,variables $variables
    if {[string length $category] && [string length ${instance,variables}]} {
      return [error "class:$class category:$category \
        -> can't add instance variables with a category !"]
    }
  }

  foreach line [split $code \;] {
    set line [string trim $line "\r\n "]
    if {![string length $line]} { continue }
    if [regexp {^([+-])[ ]*([^$]+)$} $line -> sign message] {
      set selector [@selector $message]
      set args [@args $message]
      switch -- $sign {
        + {
           set type class
        }
        - {
          set type instance
        }
        default { }
      }
      lappend ${type},selectors $selector
      lappend interface,${type},selectors $selector
    } else {
      set classVariables $line
      foreach v $classVariables {
        lappend class,variables $v
      }
    }
  }

  #- package name is specified
  if {[llength [split $class .]] > 1} {
    set package [join [lreplace [split $class .] end end] .]
    set class [lindex [split $class .] end]
  } else {
    set package {}
  }

  upvar #0 @${class} SELF

  foreach v {
    class superclass superclasses protocols categories
    class,variables class,selectors
    instance,variables instance,selectors
    interface implementation 
    interface,class,selectors interface,instance,selectors 
  } {
    if {![info exists SELF($v)]} {
      set SELF($v) {}
    }
  }

  set SELF(package) $package

  set SELF(class) $class
  set SELF(superclass) $superclass
  set SELF(superclasses) $superclasses

  set SELF(interface,class,selectors) ${interface,class,selectors}
  set SELF(interface,instance,selectors) ${interface,instance,selectors}

  if [string length $superclass] { 
    if {![@class $superclass]} {
      return [error \
        "couldn't import \"$superclass\" superclass for \"$class\" class"]
    }
    upvar #0 @${superclass} SUPER
    foreach type {instance class} {
      foreach field {variables selectors} {
        foreach value $SUPER(${type},${field}) {
          if {"$field" == "variables"} {
            if {![SELF(${type},${field}) contains: $value]} {
              lappend SELF(${type},${field}) $value
            }
          } else {
            lappend SELF(${type},${field}) $value
            lappend SELF(interface,${type},${field}) $value
          }
        }
      }
      foreach field [array names SUPER ${type},methodFor,*] {
        set SELF($field) $SUPER($field)
      }
    }
    foreach proto $SUPER(protocols) {
      if {![SELF(protocols) contains: $proto]} {
        lappend SELF(protocols) $proto
        lappend SELF(interface,protocols) $proto
      }
    }

    #- multiple_inheritence
    if {[info exists superclasses] && [llength $superclasses]} {
      set i 0
      foreach _super $superclasses {
        if {![@class $_super]} {
          return [error \
           "couldn't import \"$_super\" class for \"$class\" class"]
        }
        incr i  
        #@debug "importing $_super"
        upvar #0 @${_super} SUPER${i}
        foreach type {instance class} {
          foreach field {variables selectors} {
            foreach value [set SUPER${i}(${type},${field})] {
              if {"$field" == "variables"} {
                if {![SELF(${type},${field}) contains: $value]} {
                  lappend SELF(${type},${field}) $value
                } else {
                  @debug "warning: variable $value already exists in $class class when trying to import it from $_super"
                }
              } else {
                if {![SELF(${type},${field}) contains: $value]} {
                  lappend SELF(${type},${field}) $value
                } else {
                  #@debug "selector $value already exists when importing from $_super -> must be explicitly overridden"
                }
              }
            }
          }
          foreach field [array names SUPER${i} ${type},methodFor,*] {
            #- checking if it's already defined and if it's the same content else, 
            #- if other callback, warning
            if [info exists SELF($field)] {
              if {"[set SELF($field)]" != "[set SUPER${i}($field)]"} {
                @debug "[lindex [split $field ,] end] points to [set SELF($field)] in $class but points to [set SUPER${i}($field)] in [set SUPER${i}(class)]"
              }
            } 
            set SELF($field) [set SUPER${i}($field)]
          }
        }
        foreach proto [set SUPER${i}(protocols)] {
          if {![SELF(protocols) contains: $proto]} {
            lappend SELF(protocols) $proto
           }
        }
      }
    }
  }
  if [string length $category] {
    #- importing original class if it's a category
    if {![@class $class]} {
      return [error "couldn't import original $class definition"]
    }
    if {![SELF(categories) contains: $category]} {
      lappend SELF(categories) $category
    }
  }

  foreach protocol [split $protocols ,] {
    set protocol [string trim $protocol]
    if {![SELF(protocols) contains: $protocol]} {
      lappend SELF(protocols) $protocol
    }
  }
  foreach type1 {class instance} {
    foreach type2 {variables selectors} {
      foreach value [set ${type1},${type2}] {
        if {![SELF($type1,$type2) contains: $value]} {
          lappend SELF($type1,$type2) $value
        }
      }
    }
  }

  #- creating class variables
  foreach var $SELF(class,variables) {
    if [regexp {^@([^$]+)$} $var -> arrayName] {
      global @${class}@${arrayName}
      array set @${class}@${arrayName} {}
    } else {
      set SELF($var) {}
    }
  }

  if {![RUNTIME(protocols) contains: $class]} {
    lappend RUNTIME(protocols) $class
  }

  set SELF(interface) [@formatLibraryPathName [info script]]

}
