#-------------------------------
# add a package into runtime 
#-------------------------------
proc @package {{args {}}} {
  upvar #0 @RUNTIME RUNTIME

  #- default is to return current package
  if {![string length $args]} { 
    set RUNTIME(package) 
  } else {
    #- a package name or a list of packages is specified
    set packages [join $args]
    foreach package $packages {
      #- if directory not found
      if {![file isdirectory $package]} {
        #- if package is not directly in the current library path
        if {![file isdirectory [file join $RUNTIME(directory) $package]]} {
          #- searching if package is under one of the runtime paths
          foreach path $RUNTIME(path) {
            if {"[file tail $path]" == "$package"} {
              if [file isdirectory $path] {
                set dir $path
              }
            } elseif [file isdirectory [file join $path $package]] {
              set dir [file join $path $package]
            }
          }
        } else {
          set dir [file join $RUNTIME(directory) $package]
        }
      } else {
        set dir $package
      }

      if {![RUNTIME(packages) contains: $package]} {
        lappend RUNTIME(packages) $package
      }
      set RUNTIME(package) $package
    }
  }
  set RUNTIME(package)
}

proc @packages {{args {}}} {
  upvar #0 @RUNTIME RUNTIME

  #- default is to return current packages list
  if {![string length $args]} { 
    set RUNTIME(packages) 
  } else {
    #- a package name or a list of packages is specified
    set packages [join $args]
    foreach package $packages {
      @package $package
    }
  }
  set RUNTIME(packages)
}

proc @getPackageFromClass class {
  set package [join [lreplace [split [file tail $class] .] end end] .]
}

proc @getPackageFromFileName fileName {
  if [regexp {(interface|implementation|protocol|\.)} \
    [set package [file tail [file dirname $fileName]]]] {
    set package [file tail [file dirname [file dirname $fileName]]]
  }
  set package
}

proc @getPackageForClass class {
  upvar #0 @RUNTIME RUNTIME
  #- package is specified in the class name
  set package [@getPackageFromClass $class]
  if [string length $package] {
    #- isolate class name from the package one
    set class [@getClassFromClass $class]
  } else {
    #- package is not specified in the class name 
    set packages {}
    #- taking package from runtime if class is already loaded
    foreach class2 $RUNTIME(classes) {
      if [string length [set pkg [@getPackageFromClass $class2]]] {
        set class2 [@getClassFromClass $class2]
      }
      if {"$class2" == "$class"} {
        lappend packages $pkg
      }
    }
    if {[set cnt [llength $packages]] == 1} {
      return $packages
    } elseif {$cnt > 1} {
      @error "must specify package name! (class found under packages: [join $packages {, }])"
    } else {
      #- taking package from current script file name
      set fileName [info script]
      set package [@getPackageFromFileName $fileName]
    }
  } 
  set package
}

if {![info exists @RUNTIME(packages)]} {
  set @RUNTIME(packages) {}
}

if {![info exists @RUNTIME(package)]} {
  set @RUNTIME(package) {}
}
@package core