#--------------------------------------------------------------
# callback used when a message is sent to a class or an object
#--------------------------------------------------------------
proc @sendMessageToObject {self {message {}}} {

  upvar #0 @${self}(class) class
  upvar #0 @${class} SELF

  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }

  set selector [@selector $message]
  set args [@args $message]

  #- empty message or ?
  switch $message {
    {} -
    ?  {
      set txt "$self respondsTo:\n"

      upvar #0 @${self} SELF

      if {"$type" == "instance"} {
        upvar #0 @$SELF(class) CLASS
        lappend sel $CLASS(instance,selectors)
        set sep -
      } else {
        lappend sel $SELF(class,selectors)
        set sep +
      }
      set sel [lsort [split [join $sel]]]
      foreach s $sel {
        append txt "\t${sep} $s\n"
      }
      return [error [string trimright $txt \n]]
    }
  }

  if {[info exists SELF(${type},methodFor,${selector})]} {
    set newCall $SELF(${type},methodFor,${selector}) 
    set level [info level]
    incr level -1
    #@debug "call: $SELF(${type},methodFor,${selector}) < $level) [uplevel $level info level 0]"
    eval $SELF(${type},methodFor,${selector}) $self $args
  } elseif {[info exists SELF(${type},methodFor,forwardInvocation:)]} {
    $SELF(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists SELF(${type},methodFor,forward:)]} {
    $SELF(${type},methodFor,forward:) $self $message
  } elseif {[info exists SELF(${type},methodFor,doesNotRecognize:)]} {
    $SELF(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SELF(${type},methodFor,error:)]} {
    $SELF(${type},methodFor,error:) $self \
      "$self cannot respond to \"$message\" message !"
  } else {
    error "$self cannot respond to \"$message\" message !"
  }

}

#------------------------------------------------------
# callback used when a message has to be sent to super
#------------------------------------------------------
proc @sendMessageToSuper { {args {}} } {

  #- get caller proc
  if [catch {set callerFullProcName [info level -1]} err] {
    return [error "cannot call super outside of a instance/class method ($err)"]
  }
  set callerProcName [lindex $callerFullProcName 0]
  set self [lindex $callerFullProcName 1]

  upvar #0 @${self}(class) class
  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }
  set message $args
  set selector [@selector $message]
  set args [@args $message]

  set _class [string trimleft [lindex [split $callerProcName ,] 0] @]
  upvar #0 @${_class} CLASS
  set superclass $CLASS(superclass)

  if {![string length $superclass]} {
    return [error "cannot call super method when superclass doesn't exist !"]
  }

  upvar #0 @$superclass SUPERCLASS
  if {[info exists SUPERCLASS(${type},methodFor,${selector})]} {
    eval $SUPERCLASS(${type},methodFor,${selector}) $self $args
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forwardInvocation:)]} {
    $SUPERCLASS(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forward:)]} {
    $SUPERCLASS(${type},methodFor,forward:) $self $message
  } elseif {[info exists SUPERCLASS(${type},methodFor,doesNotRecognize:)]} {
    $SUPERCLASS(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SUPERCLASS(${type},methodFor,error:)]} {
    $SUPERCLASS(${type},methodFor,error:) $self \
      "super of $self cannot respond to \"$message\" message !"
  } else {
    error "super of $self cannot respond to \"$message\" message !"
  }

}

proc @sendMessageToSuper@ { superclass {args {}} } {

  #- get caller proc
  if [catch {set callerFullProcName [info level -1]} err] {
    return [error "cannot call super outside of a instance/class method ($err)"]
  }
  set callerProcName [lindex $callerFullProcName 0]
  set self [lindex $callerFullProcName 1]

  upvar #0 @${self}(class) class
  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }
  set message $args
  set selector [@selector $message]
  set args [@args $message]

  if {![string length $superclass]} {
    return [error "cannot call super method when superclass doesn't exist !"]
  }

  upvar #0 @$superclass SUPERCLASS
  if {[info exists SUPERCLASS(${type},methodFor,${selector})]} {
    eval $SUPERCLASS(${type},methodFor,${selector}) $self $args
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forwardInvocation:)]} {
    $SUPERCLASS(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forward:)]} {
    $SUPERCLASS(${type},methodFor,forward:) $self $message
  } elseif {[info exists SUPERCLASS(${type},methodFor,doesNotRecognize:)]} {
    $SUPERCLASS(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SUPERCLASS(${type},methodFor,error:)]} {
    $SUPERCLASS(${type},methodFor,error:) $self \
      "super of $self cannot respond to \"$message\" message !"
  } else {
    error "super of $self cannot respond to \"$message\" message !"
  }

}
