#----------------------
# add a new tcl syntax
#----------------------
proc @syntax {body} {
  append b $body\n[info body unknown]
  proc unknown args $b
}

# -- new syntax --

@syntax {
  #-----------------------
  # dynamic class loading
  #-----------------------
  if {![regexp {^@} [lindex $args 0]] && \
    [string length [info proc @class]] && \
    [string length [lindex $args 0]] && \
    [@class [lindex $args 0]] == 1 && \
    [string length [info proc [lindex $args 0]]]} {
    return [eval $args]
  }
}

@syntax {
  #------------------------
  # myArray dump
  # myFieldValueList dump
  #------------------------
  if {"[lindex $args 1]" == "dump"} {
    upvar [lindex $args 0] list
    set result {}
    set length 0
    if [array exists list] {
      foreach field [lsort [array names list]] {
        if {[set len [string length $field]] > $length} { set length $len }
        lappend result [list $field $list($field)]
      }
    } elseif {[info exists list] && ![expr [llength $list] % 2]} {
      foreach {field value} $list {
        if {[set len [string length $field]] > $length} { set length $len }
        lappend result [list $field $value]
      }
    } else {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set result2 {}
    foreach line $result {
      append result2 [format %${length}s [lindex $line 0]]
      append result2 { = }
      append result2 [lrange $line 1 end]\n
    }
    return [string trimright $result2 \n]
  }
}

@syntax {
  #-----------------------------------------------------------------------
  # set list {34 56 32}
  #
  # list each: { puts stderr "in \"$self\" list, index $index = $each" }
  #   -> in "34 56 32" list, index 0 = 34
  #      in "34 56 32" list, index 1 = 56
  #      in "34 56 32" list, index 2 = 32
  #-----------------------------------------------------------------------
  if {"[lindex $args 1]" == "each:"} {
    upvar [lindex $args 0] _self
    set cmd {
      foreach var {self index each} {
        if [info exists $var] {
          set @@@@@${var} [set $var]
        }
      }
    }
    append cmd "set self "
    append cmd [list $_self]
    append cmd \n
    append cmd "set index 0\n"
    append cmd "foreach each \$self {\n"
    append cmd "  [join [lrange $args 2 end]]\n"
    append cmd "  incr index\n"
    append cmd "}"

    append cmd {
      foreach var {self index each} {
        if [info exists @@@@@${var}] {
          set $var [set @@@@@${var}]
        } else {
          unset $var
        }
      }
    }
    return [uplevel $cmd]
  }
}

@syntax {
  #-----------------
  # mytext encrypt
  #-----------------
  if {"[lindex $args 1]" == "encrypt"} {
    if [regexp {^([+\-0-9\.e]+)$} [lindex $args 0]] {
      return -code error "usage: varName encrypt"
    }
    upvar [lindex $args 0] txt
    if {![info exists txt]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set encrypted {}
    foreach c [split $txt {}] {
      scan $c %c ascii
      if [expr $ascii % 2] {
        incr ascii 15
      } else {
        incr ascii 17
      }
      if {$ascii > 255} { 
        set ascii [expr ($ascii - 255)] 
      }
      append encrypted [format %c $ascii]
    }
    return [set txt $encrypted]
  }
}

@syntax {
  #----------------
  # mytext decrypt
  #----------------
  if {"[lindex $args 1]" == "decrypt"} {
    if [regexp {^([+\-0-9\.e]+)$} [lindex $args 0]] {
      return -code error "usage: varName decrypt"
    }
    upvar [lindex $args 0] txt
    if {![info exists txt]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set decrypted {}
    foreach c [split $txt {}] {
      scan $c %c ascii
      if [expr $ascii % 2] {
        incr ascii -17
      } else {
        incr ascii -15
      }
      if {$ascii < 0} { 
        set ascii [expr (0 - $ascii)] 
      }
      append decrypted [format %c $ascii]
    }
    return [set txt $decrypted]
  }
}

@syntax { 
  #------------------------
  # set myList {1 2 3 3 2}
  # myList sort -> {1 2 3}
  #------------------------
  if {"[lindex $args 1]" == "sort"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set newlist {}
    foreach element [lsort $list] {
      if {[lsearch -exact $newlist $element] == -1} {
        lappend newlist $element
      }
    }
    return [set list $newlist]
  }
}

@syntax {
  #----------------------------------------
  # set myList {12 43 5}
  #
  # myList addIfNotYetIn: 43 -> {12 43 5}
  #----------------------------------------
  if {"[lindex $args 1]" == "addIfNotYetIn:"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set element [lindex $args 2]
    if {[lsearch -exact $list $element] == -1} {
      lappend list $element
    }
    return $list
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList sortByValue
  #   -> {Marc 23 Tom 24 Luc 43}
  #--------------------------------------------
  if {"[lindex $args 1]" == "sortByValue"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    foreach {field value} $list {
      lappend values $value
      set fields($value,$field) $field
    }
    values sort
    set list {}
    foreach value $values {
      foreach f [array names fields ${value},*] {
        lappend list $fields($f)
        lappend list $value
      }
    }
    unset fields values field value
    return $list
  }
}

@syntax {
  #---------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList sortByField
  #   -> {Luc 43 Marc 23 Tom 24}
  #---------------------------------------------
  if {"[lindex $args 1]" == "sortByField"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    foreach {field value} $list {
      lappend fields $field
      set values($field,$value) $value
    }
    fields sort
    set list {}
    foreach field $fields {
      foreach f [array names values ${field},*] {
        lappend list $field
        lappend list $values($f)
      }
    }
    unset fields values field value
    return $list
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList fields
  #   -> {Marc Luc Tom}
  #--------------------------------------------
  if {"[lindex $args 1]" == "fields"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set fields {}
    foreach {field value} $list {
      lappend fields $field
    }
    return $fields
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList values
  #   -> {23 43 24}
  #--------------------------------------------
  if {"[lindex $args 1]" == "values"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set values {}
    foreach {field value} $list {
      lappend values $value
    }
    return $values
  }
}

@syntax {
  #------------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList set: Marc to: 43 
  #   -> $fieldValueList = {Marc 43 Luc 43 Tom 24}
  #------------------------------------------------
  if {"[lindex $args 1]" == "set:" && "[lindex $args 3]" == "to:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    set value [lindex $args 4]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    if {[set index [lsearch -exact $list $field]] != -1} {
      incr index
      set list [lreplace $list $index $index $value]
    } elseif [string length $field] {
      return \
        -code error "couldn't find \"$field\" field in \"[lindex $args 0]\" list"
    }
    return $list
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList get: Marc 
  #   -> 23
  #--------------------------------------------
  if {"[lindex $args 1]" == "get:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    if {[set index [lsearch -exact $list $field]] != -1} {
      if {![expr $index % 2]} {
        #- if event -> it's a field -> returning value
        incr index
      } else {
        #- if odd -> it's a value -> returning value
      }
      return [lindex $list $index]
    } elseif [string length $field] {
      return \
        -code error "couldn't find \"$field\" field in \"[lindex $args 0]\" list"
    }
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList add: {John 2}
  #   -> {Marc 23 Luc 43 Tom 24 John 2}
  #--------------------------------------------
  if {"[lindex $args 1]" == "add:"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set list [join [list $list [lindex $args 2]]]
    return $list
  }
}

@syntax { 
  #-------------------------------------
  # set myFieldValueList {marc 3 luc 4}
  #
  # myFieldValueList removeField: marc 
  #   -> $myFieldValueList = {luc 4}
  #-------------------------------------
  if {"[lindex $args 1]" == "removeField:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    while {[set index [lsearch -exact $list $field]] != -1} {
      set list [lreplace $list $index [expr $index + 1]]
    }
    return $list
  }
}

@syntax {
  #--------------------------
  # set myList {12 45 2 34}
  #
  # myList remove: 45
  #   -> $myList = {12 2 34}
  #--------------------------
  if {"[lindex $args 1]" == "remove:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    while {[set index [lsearch -exact $list $field]] != -1} {
      set list [lreplace $list $index $index]
    }
    return $list
  }
}

@syntax {
  #------------------------
  # set myList {1 2 34 54}
  #
  # myList contains: 34
  #  -> 1
  # myList contains: 35
  #  -> 0
  #------------------------
  if {"[lindex $args 1]" == "contains:"} {
    upvar [lindex $args 0] listname
    set field [lindex $args 2]
    if {![info exists listname]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    } elseif {[set index [lsearch -exact $listname $field]] != -1} {
      return 1
    } else {
      return 0
    }
  }
}

@syntax {
  #------------------------
  # set myList {1 2 34 54}
  #
  # myList indexOf: 34
  #  -> 2
  # myList indexOf: 23
  #  ->
  #------------------------
  if {"[lindex $args 1]" == "indexOf:"} {
    upvar [lindex $args 0] listname
    set field [lindex $args 2]
    if {![info exists listname]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    } elseif {[set index [lsearch -exact $listname $field]] != -1} {
      return $index
    } else { 
      return -1
    }
  }
}

@syntax {
  #-----------------------------------
  # set a [expr $a +  2] -> a += 2
  #
  # nb: will work for +,-,*,/,%,>>,<<
  #-----------------------------------
  if {[regexp {^([+\-\\\*%/><]+)=$} [lindex $args 1] -> sign]} {
    return [uplevel "set [lindex $args 0] \
      \[expr \$[lindex $args 0] $sign [lrange $args 2 end]\]"]
  }
}

@syntax {
  #-------------------
  # incr a -1 -> a --
  #-------------------
  if {"[lindex $args 1]" == "--"} { 
    if {[regexp {^([+\-0-9\.e]+)$} [lindex $args 0] -> nb]} {
      return [expr $nb - 1]
    } else {
      return [uplevel "incr [lindex $args 0] -1"]
    }
  }
}

@syntax {
  #-------------------
  # incr a 1  -> a ++
  #-------------------
  if {"[lindex $args 1]" == "++"} { 
    if {[regexp {^([+\-0-9\.e]+)$} [lindex $args 0] -> nb]} {
      return [expr $nb + 1]
    } else {
      return [uplevel "incr [lindex $args 0]"]
    }
  }
}

@syntax {
  #------------------
  # set a 1 -> a = 1
  #------------------
  if {"[lindex $args 1]" == "="} {
    set cmd "set [lindex $args 0] [lrange $args 2 end]"
    return [uplevel "set [lindex $args 0] [lrange $args 2 end]"]
  }
}

@syntax {
  #---------------------------------
  # print string to standard output
  #
  # "Hello world" stdout
  #---------------------------------
  if {"[lindex $args 1]" == "stdout"} {
    return [puts stdout [lindex $args 0]]
  }
}

@syntax {
  #--------------------------------
  # print string to standard error
  #
  # "Hello world" stderr
  #--------------------------------
  if {"[lindex $args 1]" == "stderr"} {
    return [puts stderr [lindex $args 0]]
  }
}

@syntax {
  #-----------------------
  # returns string length
  #
  # "Hello world" length
  #-----------------------
  if {"[lindex $args 1]" == "length"} {
    return [string length [lindex $args 0]]
  }
}

@syntax {
  #-----------------------
  # returns list length
  #
  # "Hello world" llength
  #-----------------------
  if {"[lindex $args 1]" == "llength"} {
    return [llength [lindex $args 0]]
  }
}

#-------------------------
# with <object> {
#   <message1> <message2>
#   <message3> ...
# }
#-------------------------
proc with {object commands} {
  #set nb [expr [info level] -1]
  foreach command $commands {
    uplevel 1 $object $command
  }
  return
}
