source mOT.tcl

@class Object

proc createObjects {nb class containerName} {
  upvar $containerName objects
  puts stderr "creating $nb objects of class: $class ... " nonewline
  set t1 [clock clicks -milliseconds]
  for {set i 1} {$i <= $nb} {incr i} { 
    lappend objects [$class new] 
  }
  set t2 [clock clicks -milliseconds]
  set ms [expr $t2 - $t1]
  puts stderr "done in $ms ms (1 in [expr (1.0 * $ms)/($nb * 1.0)])"
  return $ms
}

proc sendMessageToObjects {message containerName} {
  upvar $containerName objects
  set nb [llength $objects]
  puts stderr "sending $message message to $nb objects ... " nonewline
  set t1 [clock clicks -milliseconds]
  foreach obj $objects {
    $obj $message
  }
  set t2 [clock clicks -milliseconds]
  set ms [expr $t2 - $t1]
  puts stderr "done in $ms ms (1 in [expr (1.0 * $ms)/($nb * 1.0)])"
  return $ms
}

foreach t {tries tc1 tm1 tm2 tr1} { set $t 0 }

foreach factor {1 10 100 1000} {
  foreach count {1 5} {
    set nb [expr $count * $factor]
    set objects {}
    incr tries
    set tc1 [expr ( $tc1 + (1.0 * [createObjects $nb Object objects]) / (1.0 * $nb) ) ]
    set tm1 [expr ( $tm1 + (1.0 * [sendMessageToObjects refCount objects]) / (1.0 * $nb) ) ]
    set tm2 [expr ( $tm2 + (1.0 * [sendMessageToObjects self objects]) / (1.0 * $nb) ) ]
    set tr1 [expr ( $tr1 + (1.0 * [sendMessageToObjects release objects]) / (1.0 * $nb) ) ]
    set objects {}
  }
}

puts stderr ""
puts stderr "average"
puts stderr "-------"
puts stderr ""
puts stderr " 1 object creation = [expr (1.0 * $tc1) / (1.0 * $tries)] ms"
puts stderr " 1 object message (refCount) = [expr (1.0 * $tm1) / (1.0 * $tries)] ms"
puts stderr " 1 object message (self) = [expr (1.0 * $tm2) / (1.0 * $tries)] ms"
puts stderr " 1 object release = [expr (1.0 * $tr1) / (1.0 * $tries)] ms"
puts stderr ""
