
#------------------------------------------------------------------------------
# tk specific
#------------------------------------------------------------------------------

proc @tk {} {
  upvar #0 @RUNTIME RUNTIME 
  global tcl_platform

  if [catch {winfo exists .} err] {
    return 0
  } 

  @package gui

  if {![info exists RUNTIME(images)]} {
    set RUNTIME(images) {}
  } elseif {[llength $RUNTIME(images)] > 0} {
    return 1
  }

  #----------------------------------------------------------------------------
  # ressources
  #----------------------------------------------------------------------------
  option add *highlightThickness 0
  option add *selectBorderWidth 0
  foreach widget {Button Scrollbar Listbox Entry Canvas Text Menu MenuButton} {
    option add *${widget}.borderWidth 1 
  }
  option add *Text.wrap none

  #- colors
  if {$tcl_platform(platform) == "windows"} {
    set bg [[frame ._s] cget -bg]
    destroy ._s
  } else {
    set bg #d0d0d0
    option add *.background $bg
    foreach w {Frame Label Button Canvas Scrollbar} {
      option add *${w}.background $bg
    }
  }
      
  foreach w {Listbox Text Entry} {
    option add *$w.background #f0f0f0
    option add *$w.foreground black 
    option add *$w.selectBackground #c0c0d0
    option add *$w.selectForeground black
  }
 
  #- removing bg change on focus for scrollbar
  option add *Scrollbar.activeBackground $bg
  option add *Button.activeBackground $bg
  option add *Button.highlightBackground $bg

  #- normalize font use 
  set w [label ._test]
  set defaultFont [$w cget -font]
  if {[set index [lsearch -exact $defaultFont bold]] != -1} {
    set defaultFont [lreplace $defaultFont $index $index]
  }
  destroy $w
 
  foreach class {Label Text Listbox Entry Button} { 
    option add *${class}*Font $defaultFont
  }

  @image down_arrow [image create bitmap \
    -data {
      #define down_arrow_width 11
      #define down_arrow_height 11
      static unsigned char down_arrow_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x03, 0xfc, 0x01,
        0xf8, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]
   
  @image up_arrow [image create bitmap \
    -data {
      #define up_arrow_width 11
      #define up_arrow_height 11
      static unsigned char up_arrow_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x70, 0x00, 0xf8, 0x00, 0xfc, 0x01,
        0xfe, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image left_arrow [image create bitmap \
    -data {
      #define left_arrow_width 11
      #define left_arrow_height 11
      static unsigned char left_arrow_bits[] = {
        0x00, 0x00, 0x40, 0x00, 0x60, 0x00, 0x70, 0x00, 0x78, 0x00, 0x7c, 0x00,
        0x78, 0x00, 0x70, 0x00, 0x60, 0x00, 0x40, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image right_arrow [image create bitmap \
    -data {
      #define right_arrow_width 11
      #define right_arrow_height 11
      static unsigned char right_arrow_bits[] = {
        0x00, 0x00, 0x10, 0x00, 0x30, 0x00, 0x70, 0x00, 0xf0, 0x00, 0xf0, 0x01,
        0xf0, 0x00, 0x70, 0x00, 0x30, 0x00, 0x10, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image return [image create bitmap \
    -data {
      #define return_width 15
      #define return_height 10
      static unsigned char return_bits[] = {
        0x10, 0x3c, 0x18, 0x04, 0x2c, 0x04, 0xe6, 0x07, 0x03, 0x00, 0x01, 0x00,
        0x02, 0x00, 0x04, 0x00, 0x08, 0x00, 0x10, 0x00, 
      };
    } -maskdata {
        #define returnmask_width 15
        #define returnmask_height 10
        static unsigned char returnmask_bits[] = {
          0x10, 0x7c, 0x18, 0x44, 0x2c, 0x44, 0xe6, 0x47, 0x03, 0x40, 0x01, 0x40,
          0xe2, 0x7f, 0x24, 0x00, 0x28, 0x00, 0x30, 0x00, 
        };
    } -background white -foreground black 
  ]
  
  @image radioOn [image create bitmap \
    -data {
      #define radioOn_width 16
      #define radioOn_height 16
      static unsigned char radioOn_bits[] = {
        0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
        0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x04, 0x00,
        0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
    } -maskdata {
        #define radioOnmask_width 16
        #define radioOnmask_height 16
        static unsigned char radioOnmask_bits[] = {
          0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0xc4, 0x07, 0xe4, 0x0f,
          0xf2, 0x1f, 0xf2, 0x9f, 0xf2, 0x9f, 0xf2, 0x9f, 0xf2, 0x9f, 0xe4, 0x4f,
          0xc4, 0x47, 0x00, 0x20, 0x20, 0x18, 0xc0, 0x07, };
    } -background white -foreground black
  ]
  
  @image radioOff [image create bitmap \
    -data {
      #define radioOff_width 16
      #define radioOff_height 16
      static unsigned char radioOff_bits[] = {
        0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
        0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x04, 0x00,
        0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    } -maskdata {
        #define radioOffmask_width 16
        #define radioOffmask_height 16
        static unsigned char radioOffmask_bits[] = {
          0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
          0x02, 0x00, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x04, 0x40,
          0x04, 0x40, 0x00, 0x20, 0x10, 0x18, 0xe0, 0x07, 
        };
    } -background white -foreground black
  ]
  
  @image empty16x16 [image create bitmap \
    -data {
      #define switchOff_width 16
      #define switchOff_height 16
      static unsigned char switchOff_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image switchOff [@imageCopy empty16x16]
  @image switchOn [image create bitmap \
    -data {
      #define switchOn_width 16
      #define switchOn_height 16
      static unsigned char switchOn_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x0c,
        0x20, 0x06, 0x20, 0x03, 0xa0, 0x01, 0xc0, 0x00, 0x60, 0x00, 0x30, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    } -maskdata {
        #define switchOnmask_width 16
        #define switchOnmask_height 16
        static unsigned char switchOnmask_bits[] = {
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x00, 0x0e,
          0x30, 0x07, 0xb0, 0x03, 0xf0, 0x01, 0xf0, 0x00, 0x70, 0x00, 0x30, 0x00,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        };
    } -background white -foreground black
  ]
   
  @image empty9x9 [image create bitmap \
    -data {
      #define empty_width 9
      #define empty_height 9
      static unsigned char empty_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00
      };
    }
  ]
  
  @image empty [@imageCopy empty9x9]
  
  @image plus [image create bitmap \
    -data {
      #define plus_width 9
      #define plus_height 9
      static unsigned char plus_bits[] = {
        0xff, 0x01, 0x01, 0x01, 0x11, 0x01, 0x11, 0x01, 0x7d, 0x01, 0x11, 0x01,
        0x11, 0x01, 0x01, 0x01, 0xff, 0x01
      };
    }
  ]
  
  @image minus [image create bitmap \
    -data {
      #define minus_width 9
      #define minus_height 9
      static unsigned char minus_bits[] = {
        0xff, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x7d, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0xff, 0x01
      };
    }
  ]

set red_sphere {R0lGODlhDAAMAMYAALGRkYY+PnwREYwQEKs6OtKPj+Tf325AQGAKCnULC4kJ
CZwICKsHB7UGBsY5OfXe3mJBQVIODmkMDIAKCpcICKwHB78FBcwEBNADA9I3
N5yTkz8QEFcODm8MDIgKCqAICLcGBs4EBOECAugBAdoDA+WNjVNCQkAQEFkO
DnEMDIoJCaIHB7sFBdQDA+wBAfsAAOICAtU3NysYGD4QEIcKCp8ICLYGBs0E
BOACAuYCAtkDA8UKCicYGDkQEFEODpYICL0FBcoEBMUEBLgMDElDQzIREUkP
D18KCnQLC5oICKkHB7IGBq8GBrY6OpSUlCgSEj0QEGUKCncLC5MJCZgICMyP
jy8REUMPD1QODmQKCnILC3wLC4IKCoQKCpo8PODg4E1CQkIQEFAODlwKCmoM
DIk+Puvf35mUlFlCQkAWFkkVFXFAQLCRkf//////////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiIB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+mT8MQEE3QkNERUZHSB5JSksKTE1OT1ARUVI0
U0kLVFV/RFZXWFlaW1xdXoJfYEVhYmNRZGVmin9naNKoWcNGUSAAOw==}

set orange_sphere {R0lGODlhDAAMAMYAALGqlYZ2RnxlG4xwG6uSRtLDleTj4G5lRWBOFXVeFolt
F5x7F6uHF7WOGManR/Xw4GJbRFJDFWlVFoBmFpd3F6yIF7+WGMygGNCjGNKx
R5yalD80FVdHFW9ZFohsFqB+F7eQGM6hGOGwGei1GdqrGeXSllNQREA1FVlI
FXFbFopuF6KAF7uTGNSmGOy5GfvEGeKxGdWyRysmGj40FYdsFp9+F7aQGM2h
GOCvGea0GdmqGcWcHScjGjkwFFFCFWlUFpZ3F72VGMqeGMWbGLiSHUlIRDIr
FEk8FV9NFXRdFpp6F6mFF7KNGK+KF7abRpSUlCgjFD0zFWVSFndgFodrFpN1
F5x8F5h4F8y+lS8pFEM3FVRFFWRRFnJbFnxjFoJoFoRpFpqFRuDg4E1LRDIq
FEI3FVBCFVxLFWVRFmpVFol4Ruvo4JmYlFlUREA2G0k+G3FmRbCplf//////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+PxNADEFCN0NERUZHSEkeSktMCk1OT1BREVJT
VFVKVldYf0VZWltcXV5fYGGCYmNkZWZnaGlqa4p/bNq4eQMnjqJAADs=}

set yellow_sphere {R0lGODlhDAAMAMYAALGxkoaFPnx7EoyLEauqPNLSkOTk325uQWBfDnV0DImI
C5ybCqupCbWzCMbFOvX03mJiQlJRD2lpCoB/DJeWCqyrCL+9B8zKBtDOBtLS
OZyckz8+EFdWD29uCoiHC6CeCbe2CM7MBuHgBejmBNrYBeXkjlNTQkBAEFlY
D3FwCoqJC6KhCbu6B9TSBuzqBPv5A+LgBdXTOSsrGD4+EIeGC5+eCba1CM3L
BuDeBebkBNnXBcXDCicmGDk5EVFRD2loCpaVCr27B8rIBs3MBsXEB7i2DklJ
QzIyEUlIEF9eDnRzCpqZCqmoCbKxCK+uCLa1O5SUlCgnEj09EGVlDnd2DJOS
CpiXCszLkC8vEUNCEFRUD2RkDnJxCnx7DIKBC4SDC5qZPuDg4E1NQzIxEUJB
EFBQD1xbDmVkDmppComIPuvr35mZlFlZQkA/F0lJFnFxQLCwkv//////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+PxNADEFCQ0RFRkdISUoeS0xNCk5PUFFSEVNU
NFVLC1ZXf0ZYWVpbXF1eX2CCYWJjZGVmZ2hpaop/a9i0cfMGjqJAADs=}

set green_sphere {R0lGODlhDAAMAMYAAJyxklWGQDR8EziME2CrPqXSkOHk309uQShgDy51DjSJ
DDmcCz2rCj+1CmfGPOX13kxiQiRSECtpDjGACjeXDD2sCkK/CUbMCUfQCGrS
O5ackx8/ESZXDyxvDjOICjqgC0C3CkbOCEzhB03oB0raCKrlj0hTQiBAESZZ
Dy1xDjSKDDqiC0G7CkjUCE/sB1L7BkziB2vVOx0rGB8+ESZXEDOHCjmfC0C2
CkbNCEvgB03mB0nZCEjFDxwnGR45ESRREDeWDEK9CUXKCUTFCUS4EEVJQxwy
EiJJEChfDy50DjiaCzypCz+yCj6vCmO2PZSUlBkoEh89ESplDy93DjaTDDiY
DKPMkEZJQxsvEiBDESVUECpkDy1yDjB8CjKCCjKECluaPuDg4EZNQyBCESRQ
ECdcDytqDlaJP+Pr35WZlElZQiNAFyZJF1BxQZuwkv//////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzNB01Njc4OTo7PD0+P5lADEFCOENERUZHSEkeSktMCk1OT1BREVJT
NVRKC1VWf1dYWVpbXF1eX2CCYWJGY2RlUmZnaIp/adSsYdPGjaJAADs=}

set blue_sphere {R0lGODlhDAAMAMYAAJOisUFlhhVKfBVRjD92q5Gz0t/i5EJZbhA5YA9EdQ5O
iQ5XnApeqwpjtT6Ext/q9UJTYhEyUhA+aQ9JgA5UlwpfrAxovwxvzAtw0D6K
0pSYnBIpPxE1VxBBbw9NiA5ZoApktwxvzgt54Qp86At22pC85UNLUxIpQBE2
WRBCcQ5Oigpaogxmuwty1Ap/7AqG+wt64j6L1RkiKxIpPg9Nhw5Ynwpktgxv
zQt44At75gt12RJuxRkfJxImOREyURA9aQ5UlgxnvQxuygxrxRNouENGSRIi
MhEuSRA5Xw9DdA5WmgpdqQpisgpgrz58tpSUlBMdKBIoPRA8ZQ9Fdw9Mhw5S
kw5VmJGvzBIhLxErQxE0VBA7ZA9Ccg9HfA9Kgg9LhEBumuDg4ENJTRIqQhEx
UBA3XBA+akFmid/l65SWmUJOWRgsQBcxSUJacZOisP//////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+PxNADEFCN0NERUZHSEkeSktMCk1OT1BREVJT
VFVKC1ZXf0VYWVpbXF1eX2CCYWJGY2RlUmZnaIp/adSsYdPGjaJAADs=}

set white_sphere {R0lGODlhDAAMAMYAALGxsYaGhnx8fIyMjKurq9LS0uTk5G5ubmBgYHV1dYmJ
iZycnLW1tcbGxvX19WJiYlJSUmlpaYCAgJeXl6ysrL+/v8zMzNDQ0D8/P1dX
V29vb4iIiKCgoLe3t87OzuHh4ejo6Nra2uXl5VNTU0BAQFlZWXFxcYqKiqKi
oru7u9TU1Ozs7Pv7++Li4tXV1SsrKz4+PoeHh5+fn7a2ts3NzeDg4Obm5tnZ
2cXFxScnJzk5OVFRUZaWlr29vcrKyri4uElJSTIyMl9fX3R0dJqamqmpqbKy
sq+vr5SUlCgoKD09PWVlZXd3d5OTk5iYmC8vL0NDQ1RUVGRkZHJycoKCgoSE
hE1NTUJCQlBQUFxcXGpqauvr65mZmbCwsP//////////////////////////
////////////////////////////////////////////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHh4B/ggABAgME
BYKCBgcICQoLBAwKDoIPEBESExQVFheJCxgZGhscHR4fICEiIyQlJicoKSor
LC0uLzCjMTIzNDU2Nzg5OjuZPAQ9PjQ4P0BBQEJDG0RFRgxHM0hJShBLTDFN
RAtOFn9AT1BRUlMCVFVEgjVWQVdYWUtaCluKf1wlJICY6KIoEAA7}

set black_sphere {R0lGODlhDAAMAMYAAKOio2VlZUtKS1NSU3h4eLOzs+Li4llZWTo6OkVFRU9P
T1hYWGBfYGVkZYWFherq6lNTUzMzMz8/P1ZVVmFgYWppanFwcXNyc4yMjJiY
mCkpKTY1NkJCQk5OTlpaWmZmZnJxcnx7fH9+f3h3eL29vUxLTCoqKjY2NkND
Q1xbXGhoaHV0dYGAgYmIiY2MjSIiIjU1NUJBQmZlZnFxcXt6e359fnd2d3Bv
cCAgICcnJzMyMz8+P0pKSlVVVWlpaW5tbkZGRiMjIy4uLjo5OkRERE9OT1hX
WF9eX2RjZGJiYn19fZSUlB4eHikoKT09PU5NTlRTVFlYWVdWV7CwsCsrKzQ0
NDw8PEhISExMTODg4ElJSTIyMjg4OD08PWdmZ+Xl5ZaWliwsLDIxMltaW6Ki
ov//////////////////////////////////////////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERICExQVFhcYfxkaGxwdHh8gISIjJCUmJygKKSor
LC0hLi8aMDGmMjM0NTY3ODk6Ozw9DD43ID8VQEFCQ0RFRkdICklKS0xNEU5A
T1BGUVJTf0AvVFVWKFclWDeCWVpBVFtcXRJeX4p/YDqEETOGjKJAADs=}

set turquoise_sphere {R0lGODlhDAAMAMYAAJKxsECGhBR8eROMiD6rp5DS0N/k5EJubQ9gXQ51cgqJ
hgycmAurpgu1sDzGwt719EJiYRBSUA9pZwqAfQyXkwuspwq/uQnMxgnQyjzS
zpOcnBE/PRBXVQ5vbQqIhAygmwq3sgnOyAjh2wjo4gna1JDl4kJTUxFAPxBZ
Vw5xbgqKhgyingq7tgnUzgjs5gf79Aji3DvV0BgrKhE+PQ5vbAqHhAyfmwq2
sgnNxwjg2Qjm3wnZ0xDFwRknJhE5OBBRTw9pZg6AfAyWkgq9uAnKxAnNyArF
wBG4s0NJSRIyMRFJRw9fXA50cQqIhQyalguppAuyrguvqz62s5SUlBIoJxE9
PA9lYw53dAqHgwyTjwyYlJHMyhIvLxFDQRBUUw9kYg5ybw58eQqCfwqEgD6a
mODg4ENNTRFCQBBQTg9cWg9lYg9qZ0CJh9/r65SZmUJZWBdAPxdJSEFxcJKw
r////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHi4B/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzHDQ1Njc4OTo7PD0+P0BBQgxDREVGR0hJSktMTU5PUApRUlNUVRFW
V1hZTgtaW39IXF1eX2BhYmNkgmVmSWdoaWprbG2Kf9y8gRNHzhxFgQAAOw==}

set mauve_sphere {R0lGODlhDAAMAMYAALCTsYRChnoWfIkWjKhAq9GS0uTf5G1Cbl4RYHMQdYcP
iZkPnKgOq7EOtcM/xvTf9WFCYlARUmcRaX4QgJQPl6kOrLsOv8gKzMwK0NA+
0pyUnD4SP1URV20Qb4UQiJ0PoLQOt8oKzt0M4eMM6NYK2uOR5VNDUz8SQFcR
WW8QcYcPip8PorcOu9AK1OgM7PYM+94M4tI+1SoZKz0SPoUQh5wPn7MOtskK
zdsK4OEM5tQK2cETxSYZJzkSOVARUX0QgJMPlqcOq7kOvcYKysIKxbQUuElD
STESMkcSSV0RX3IQdIYQiJcPmqYOqa8OsqwOr7NAtpSUlCcTKDwSPVERUmQR
ZXUQd4QQh5APk5gPmpUPmMuSzC8TL0ISQ1MRVGMRZHAQcnoQfIAQgoEQhJhB
muDg4E1DTUESQk8RUFoRXGMRZWgRaohCievf65mUmVhDWT8YQEgYSXFCca+T
sP///////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+Ej9AQUJDN0RFRkdISUpLTE1OCk9QUVJTVFVW
V1hZC1pbf0ZcXV5fYGFiY2SCZWZHZ2hpamtsbYp/bt7AiSNnjqJAADs=}

set brown_sphere {R0lGODlhDAAMAMYAAKihl3FhSl1FI2hMJItxTr+wmuLh4GFXSEg2G1c/HWVI
IHFQIXxXI4NbJJ1+Ue7p4llRRj4vGk86HF9EH25OIX1YI4lgJZNmJpVoJ6eE
UpqYlTEnGEIyGlM9HWNHH3RSIoRdJJRnJ6FvKaZyKZxsKMy4nE5KRjInGEMy
G1Q9HXZTIodeJZhpJ6l0KrN7K6JwKSUhG0IxGlI8HYRcJKBvKKRxKZtsKJBm
KyIfGy4kF046HF5EH21OIYhfJZFlJpRnJo5jJoZgKUdGRCghFzgrGUc1G1Y/
HWRIH3BQIXpWI4FaJINcJH9ZI5N3T5SUlCEcFjAmGEw4HFhAHmtMIXJRIW9P
IbqtmkhGRCcgFjQpGEAxGks4HFU+HVxCHmBFH2FGH39qTeDg4EpIRTMoGD0v
GkU0G3RjSufl4ZiWlVJNRjQqHTsvHmNYSKegl///////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkKKiss
LS4vGTAbMTKmMxc0NTY3ODkROjs8DD0+P0BBQkNERUZHSElKS0xNTk9QEVFS
HlNIVFVWf1dYWVpbXF1eX2CCYWJDY2RlWxJmZ4p/aNKoWcOmjaJAADs=}

  foreach color {
    white black green yellow orange red blue brown mauve turquoise
  } {
    @image ${color}_sphere [image create photo -data [set ${color}_sphere]]
    unset ${color}_sphere
  }
  unset color

  @image sphere [@imageCopy white_sphere]
  
  @image bullet [image create bitmap \
    -data {
      #define bullet_width 6
      #define bullet_height 6
      static unsigned char bullet_bits[] = {
        0x1e, 0x3f, 0x07, 0x03, 0x03, 0x02, 
      };
    } -maskdata {
      #define bulletmask_width 6
      #define bulletmask_height 6
      static unsigned char bulletmask_bits[] = {
        0x1e, 0x3f, 0x3f, 0x3f, 0x3f, 0x1e, 
      };
    } -background white -foreground black
  ]
}
 

proc @image {{name {}} {image {}}} {
  upvar #0 @RUNTIME RUNTIME
  if [string length $image] {
    RUNTIME(images) add: [list $name $image]
    return $image
  } elseif {![string length $name]} {
    return $RUNTIME(images)
  } elseif [RUNTIME(images) contains: $name] {
    return [RUNTIME(images) get: $name]
  } else {
    set file [file join $RUNTIME(directory) images ${name}.*]
    if [catch {set files [glob $file]} err] {
      return
    } else {
      set file [lindex $files 0]
      RUNTIME(images) add: \
        [list $name [set img [image create photo -file $file]]]
      return $img
    }
  }
}

proc @imageCopy {image} {
  if {![string length [set img [@image $image]]]} {
    set img $image 
  }
  set newimg [image create [image type $img]]
  if [catch {
    foreach option [$img configure] {
      foreach {option_name val1 val2 val3 option_value} $option {
        $newimg configure ${option_name} ${option_value}
      }
    }
  } err] {
    return [error "couldn't get bitmap configuration ($err)"]
  }
  return $newimg
}

#- returns a more suitable list of available event data (from a bind command)
proc @eventData {} {
  return [list [list \
    path %W focus %f width %w height %h keyCode %k x %x y %y X %X Y %Y \
    button %b char %A event %E keySym %K keySymNumber %N \
    type %T delta %D
  ]]
}

#- return the code to execute for a Destroy event to remove bindings
proc @removeBindingsOnDestroyEvent {} {
  return { <Destroy> { foreach event [bind %W] { catch {bind %W $event {}} } } }
}

proc @showBitmaps {} {
  global @RUNTIME

  toplevel .toplevel 
  set col 0
  set row 0
  foreach {name img} [set @RUNTIME(images)] {
    grid [set f [frame .toplevel.f${col}${row} -relief sunken -bd 1]] \
      -sticky snew -padx 1 -pady 1 -column $col -row $row
    label $f.text -text "${name} : " -anchor w
    label $f.img -image $img 
    pack $f.text -side left -fill both -expand 1
    pack $f.img -side right -anchor e
    incr row
    if {![expr $row % 10]} {
      incr col 2 
      set row 0
    }
  }
}

@tk
