#-------------
# date getter
#-------------
proc @date {} {
  return [clock format [clock seconds] -format {%d/%m/%Y}]
}

#-------------
# time getter
#-------------
proc @time {} {
  return [clock format [clock seconds] -format {%H:%M:%S}]
}

#----------------------
# date and time getter
#----------------------
proc @now {} {
  return [clock format [clock seconds] -format {%d/%m/%Y %H:%M:%S}]
}

#-------------------
# log ouput to file
#-------------------
proc @log {file message} {
  set f [open $file a+]
  puts $f "\[[@date] [@time]\] " nonewline
  puts $f $message
  close $f
  return
}

#---------------------
# common debug output
#---------------------
proc @debug {message} {
  puts stderr "\[[@date] [@time]\] \[DEBUG\] $message"
}

#---------------------
# common error output
#---------------------
proc @error {message} {
  puts stderr "\[[@date] [@time]\] \[ERROR\] $message"
}

#-----------------------------------------
# add a trace process for all future proc
#-----------------------------------------
if [info exists TRACE] {
  rename proc _proc
  _proc proc {args} {
    foreach {proc arg body} $args {}
    if { ("$proc" != "unknown") && ("$proc" != "::unknown") } {
      set bod {
        set _s +-
        set _l [uplevel info level]
        if {$_l > 0} {
          for {set _i 0} {$_i < $_l} {incr _i} {
            append _s "-"
          }
        }
        append _s "> [info level]) [info level 0] "
        @debug $_s
        catch {unset _s _l}
      }
    }
    append bod $body
    _proc $proc $arg $bod
  }
}
