#
#   procedure : @args 
#   arguments : message
# description : returns arguments list extracted from a message
#
# ex: 
#     set message "do: theAction with: myObject"
#     @args $message 
#     {theAction myObject}
#
proc @args { {message {}} } {
    set args {}
    if {![regexp : $message]} {
        set args [lrange $message 1 end]
    } elseif [string length [string trim $message { }]] {
        foreach {field value} $message {
            if {![regexp : $field]} { 
                lappend args $field 
            }
            lappend args $value
        }
    }
    set args
}

#
#   procedure : @selector 
#   arguments : message
# description : returns selector extracted from a message
#
# ex: 
#     set message "do: theAction with: myObject"
#     @selector $message 
#     do:with:
#
proc @selector { {message {}} } {
    set selector {}
    if [string length [string trim $message { }]] {
        if {![regexp : $message]} {
            set selector [lindex $message 0]
        } else {
            foreach {field value} $message {
                if [regexp : $field] { 
                    append selector $field 
                }
            }
        }
    }
    set selector
}

