
#-
#- beginning of main.tcl
#-


proc @ {} {
  uplevel #0 {
    #- setting library main directory
    set @RUNTIME(directory) [file dirname [info script]]
    if {[set @RUNTIME(directory)] == "."} {
      set @RUNTIME(directory) [pwd]
    } elseif {[set @RUNTIME(directory)] == ".."} {
      set @RUNTIME(directory) [file dirname [pwd]]
    }

    #- listing runtime files to load
    set @RUNTIME(files) {
      util.tcl syntax.tcl path.tcl package.tcl 
      selector.tcl args.tcl class.tcl 
      interface.tcl variablesAccess.tcl implementation.tcl 
      nextId.tcl newInstance.tcl callerLevel.tcl sendMessage.tcl
      import.tcl dumpRuntime.tcl doc.tcl tk.tcl
    }

    #- loading runtime files if we are coming from main.tcl file
    if {[file tail [info script]] == "main.tcl"} {
      foreach file [set @RUNTIME(files)] {
        source [file join [set @RUNTIME(directory)] $file]
      } 
    }
  }
}
@

#-
#- end of main.tcl
#-


#-
#- beginning of util.tcl
#-

#-------------
# date getter
#-------------
proc @date {} {
  return [clock format [clock seconds] -format {%d/%m/%Y}]
}

#-------------
# time getter
#-------------
proc @time {} {
  return [clock format [clock seconds] -format {%H:%M:%S}]
}

#----------------------
# date and time getter
#----------------------
proc @now {} {
  return [clock format [clock seconds] -format {%d/%m/%Y %H:%M:%S}]
}

#-------------------
# log ouput to file
#-------------------
proc @log {file message} {
  set f [open $file a+]
  puts $f "\[[@date] [@time]\] " nonewline
  puts $f $message
  close $f
  return
}

#---------------------
# common debug output
#---------------------
proc @debug {message} {
  puts stderr "\[[@date] [@time]\] \[DEBUG\] $message"
}

#---------------------
# common error output
#---------------------
proc @error {message} {
  puts stderr "\[[@date] [@time]\] \[ERROR\] $message"
}

#-----------------------------------------
# add a trace process for all future proc
#-----------------------------------------
if [info exists TRACE] {
  rename proc _proc
  _proc proc {args} {
    foreach {proc arg body} $args {}
    if { ("$proc" != "unknown") && ("$proc" != "::unknown") } {
      set bod {
        set _s +-
        set _l [uplevel info level]
        if {$_l > 0} {
          for {set _i 0} {$_i < $_l} {incr _i} {
            append _s "-"
          }
        }
        append _s "> [info level]) [info level 0] "
        @debug $_s
        catch {unset _s _l}
      }
    }
    append bod $body
    _proc $proc $arg $bod
  }
}


#-
#- end of util.tcl
#-


#-
#- beginning of syntax.tcl
#-

#----------------------
# add a new tcl syntax
#----------------------
proc @syntax {body} {
  append b $body\n[info body unknown]
  proc unknown args $b
}

# -- new syntax --

@syntax {
  #-----------------------
  # dynamic class loading
  #-----------------------
  if {![regexp {^@} [lindex $args 0]] && \
    [string length [info proc @class]] && \
    [string length [lindex $args 0]] && \
    [@class [lindex $args 0]] == 1 && \
    [string length [info proc [lindex $args 0]]]} {
    return [eval $args]
  }
}

@syntax {
  #------------------------
  # myArray dump
  # myFieldValueList dump
  #------------------------
  if {"[lindex $args 1]" == "dump"} {
    upvar [lindex $args 0] list
    set result {}
    set length 0
    if [array exists list] {
      foreach field [lsort [array names list]] {
        if {[set len [string length $field]] > $length} { set length $len }
        lappend result [list $field $list($field)]
      }
    } elseif {[info exists list] && ![expr [llength $list] % 2]} {
      foreach {field value} $list {
        if {[set len [string length $field]] > $length} { set length $len }
        lappend result [list $field $value]
      }
    } else {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set result2 {}
    foreach line $result {
      append result2 [format %${length}s [lindex $line 0]]
      append result2 { = }
      append result2 [lrange $line 1 end]\n
    }
    return [string trimright $result2 \n]
  }
}

@syntax {
  #-----------------------------------------------------------------------
  # set list {34 56 32}
  #
  # list each: { puts stderr "in \"$self\" list, index $index = $each" }
  #   -> in "34 56 32" list, index 0 = 34
  #      in "34 56 32" list, index 1 = 56
  #      in "34 56 32" list, index 2 = 32
  #-----------------------------------------------------------------------
  if {"[lindex $args 1]" == "each:"} {
    upvar [lindex $args 0] _self
    set cmd {
      foreach var {self index each} {
        if [info exists $var] {
          set @@@@@${var} [set $var]
        }
      }
    }
    append cmd "set self "
    append cmd [list $_self]
    append cmd \n
    append cmd "set index 0\n"
    append cmd "foreach each \$self {\n"
    append cmd "  [join [lrange $args 2 end]]\n"
    append cmd "  incr index\n"
    append cmd "}"

    append cmd {
      foreach var {self index each} {
        if [info exists @@@@@${var}] {
          set $var [set @@@@@${var}]
        } else {
          unset $var
        }
      }
    }
    return [uplevel $cmd]
  }
}

@syntax {
  #-----------------
  # mytext encrypt
  #-----------------
  if {"[lindex $args 1]" == "encrypt"} {
    if [regexp {^([+\-0-9\.e]+)$} [lindex $args 0]] {
      return -code error "usage: varName encrypt"
    }
    upvar [lindex $args 0] txt
    if {![info exists txt]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set encrypted {}
    foreach c [split $txt {}] {
      scan $c %c ascii
      if [expr $ascii % 2] {
        incr ascii 15
      } else {
        incr ascii 17
      }
      if {$ascii > 255} { 
        set ascii [expr ($ascii - 255)] 
      }
      append encrypted [format %c $ascii]
    }
    return [set txt $encrypted]
  }
}

@syntax {
  #----------------
  # mytext decrypt
  #----------------
  if {"[lindex $args 1]" == "decrypt"} {
    if [regexp {^([+\-0-9\.e]+)$} [lindex $args 0]] {
      return -code error "usage: varName decrypt"
    }
    upvar [lindex $args 0] txt
    if {![info exists txt]} {
      return [error "\"[lindex $args 0]\" doesn't exist"]
    }
    set decrypted {}
    foreach c [split $txt {}] {
      scan $c %c ascii
      if [expr $ascii % 2] {
        incr ascii -17
      } else {
        incr ascii -15
      }
      if {$ascii < 0} { 
        set ascii [expr (0 - $ascii)] 
      }
      append decrypted [format %c $ascii]
    }
    return [set txt $decrypted]
  }
}

@syntax { 
  #------------------------
  # set myList {1 2 3 3 2}
  # myList sort -> {1 2 3}
  #------------------------
  if {"[lindex $args 1]" == "sort"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set newlist {}
    foreach element [lsort $list] {
      if {[lsearch -exact $newlist $element] == -1} {
        lappend newlist $element
      }
    }
    return [set list $newlist]
  }
}

@syntax {
  #----------------------------------------
  # set myList {12 43 5}
  #
  # myList addIfNotYetIn: 43 -> {12 43 5}
  #----------------------------------------
  if {"[lindex $args 1]" == "addIfNotYetIn:"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set element [lindex $args 2]
    if {[lsearch -exact $list $element] == -1} {
      lappend list $element
    }
    return $list
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList sortByValue
  #   -> {Marc 23 Tom 24 Luc 43}
  #--------------------------------------------
  if {"[lindex $args 1]" == "sortByValue"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    foreach {field value} $list {
      lappend values $value
      set fields($value,$field) $field
    }
    values sort
    set list {}
    foreach value $values {
      foreach f [array names fields ${value},*] {
        lappend list $fields($f)
        lappend list $value
      }
    }
    unset fields values field value
    return $list
  }
}

@syntax {
  #---------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList sortByField
  #   -> {Luc 43 Marc 23 Tom 24}
  #---------------------------------------------
  if {"[lindex $args 1]" == "sortByField"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    foreach {field value} $list {
      lappend fields $field
      set values($field,$value) $value
    }
    fields sort
    set list {}
    foreach field $fields {
      foreach f [array names values ${field},*] {
        lappend list $field
        lappend list $values($f)
      }
    }
    unset fields values field value
    return $list
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList fields
  #   -> {Marc Luc Tom}
  #--------------------------------------------
  if {"[lindex $args 1]" == "fields"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set fields {}
    foreach {field value} $list {
      lappend fields $field
    }
    return $fields
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList values
  #   -> {23 43 24}
  #--------------------------------------------
  if {"[lindex $args 1]" == "values"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set values {}
    foreach {field value} $list {
      lappend values $value
    }
    return $values
  }
}

@syntax {
  #------------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList set: Marc to: 43 
  #   -> $fieldValueList = {Marc 43 Luc 43 Tom 24}
  #------------------------------------------------
  if {"[lindex $args 1]" == "set:" && "[lindex $args 3]" == "to:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    set value [lindex $args 4]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    if {[set index [lsearch -exact $list $field]] != -1} {
      incr index
      set list [lreplace $list $index $index $value]
    } elseif [string length $field] {
      return \
        -code error "couldn't find \"$field\" field in \"[lindex $args 0]\" list"
    }
    return $list
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList get: Marc 
  #   -> 23
  #--------------------------------------------
  if {"[lindex $args 1]" == "get:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    if {[set index [lsearch -exact $list $field]] != -1} {
      if {![expr $index % 2]} {
        #- if event -> it's a field -> returning value
        incr index
      } else {
        #- if odd -> it's a value -> returning value
      }
      return [lindex $list $index]
    } elseif [string length $field] {
      return \
        -code error "couldn't find \"$field\" field in \"[lindex $args 0]\" list"
    }
  }
}

@syntax {
  #--------------------------------------------
  # set fieldValueList {Marc 23 Luc 43 Tom 24}
  #
  # fieldValueList add: {John 2}
  #   -> {Marc 23 Luc 43 Tom 24 John 2}
  #--------------------------------------------
  if {"[lindex $args 1]" == "add:"} {
    upvar [lindex $args 0] list
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    set list [join [list $list [lindex $args 2]]]
    return $list
  }
}

@syntax { 
  #-------------------------------------
  # set myFieldValueList {marc 3 luc 4}
  #
  # myFieldValueList removeField: marc 
  #   -> $myFieldValueList = {luc 4}
  #-------------------------------------
  if {"[lindex $args 1]" == "removeField:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    while {[set index [lsearch -exact $list $field]] != -1} {
      set list [lreplace $list $index [expr $index + 1]]
    }
    return $list
  }
}

@syntax {
  #--------------------------
  # set myList {12 45 2 34}
  #
  # myList remove: 45
  #   -> $myList = {12 2 34}
  #--------------------------
  if {"[lindex $args 1]" == "remove:"} {
    upvar [lindex $args 0] list
    set field [lindex $args 2]
    if {![info exists list]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    }
    while {[set index [lsearch -exact $list $field]] != -1} {
      set list [lreplace $list $index $index]
    }
    return $list
  }
}

@syntax {
  #------------------------
  # set myList {1 2 34 54}
  #
  # myList contains: 34
  #  -> 1
  # myList contains: 35
  #  -> 0
  #------------------------
  if {"[lindex $args 1]" == "contains:"} {
    upvar [lindex $args 0] listname
    set field [lindex $args 2]
    if {![info exists listname]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    } elseif {[set index [lsearch -exact $listname $field]] != -1} {
      return 1
    } else {
      return 0
    }
  }
}

@syntax {
  #------------------------
  # set myList {1 2 34 54}
  #
  # myList indexOf: 34
  #  -> 2
  # myList indexOf: 23
  #  ->
  #------------------------
  if {"[lindex $args 1]" == "indexOf:"} {
    upvar [lindex $args 0] listname
    set field [lindex $args 2]
    if {![info exists listname]} {
      return -code error "\"[lindex $args 0]\" doesn't exist"
    } elseif {[set index [lsearch -exact $listname $field]] != -1} {
      return $index
    } else { 
      return -1
    }
  }
}

@syntax {
  #-----------------------------------
  # set a [expr $a +  2] -> a += 2
  #
  # nb: will work for +,-,*,/,%,>>,<<
  #-----------------------------------
  if {[regexp {^([+\-\\\*%/><]+)=$} [lindex $args 1] -> sign]} {
    return [uplevel "set [lindex $args 0] \
      \[expr \$[lindex $args 0] $sign [lrange $args 2 end]\]"]
  }
}

@syntax {
  #-------------------
  # incr a -1 -> a --
  #-------------------
  if {"[lindex $args 1]" == "--"} { 
    if {[regexp {^([+\-0-9\.e]+)$} [lindex $args 0] -> nb]} {
      return [expr $nb - 1]
    } else {
      return [uplevel "incr [lindex $args 0] -1"]
    }
  }
}

@syntax {
  #-------------------
  # incr a 1  -> a ++
  #-------------------
  if {"[lindex $args 1]" == "++"} { 
    if {[regexp {^([+\-0-9\.e]+)$} [lindex $args 0] -> nb]} {
      return [expr $nb + 1]
    } else {
      return [uplevel "incr [lindex $args 0]"]
    }
  }
}

@syntax {
  #------------------
  # set a 1 -> a = 1
  #------------------
  if {"[lindex $args 1]" == "="} {
    set cmd "set [lindex $args 0] [lrange $args 2 end]"
    return [uplevel "set [lindex $args 0] [lrange $args 2 end]"]
  }
}

@syntax {
  #---------------------------------
  # print string to standard output
  #
  # "Hello world" stdout
  #---------------------------------
  if {"[lindex $args 1]" == "stdout"} {
    return [puts stdout [lindex $args 0]]
  }
}

@syntax {
  #--------------------------------
  # print string to standard error
  #
  # "Hello world" stderr
  #--------------------------------
  if {"[lindex $args 1]" == "stderr"} {
    return [puts stderr [lindex $args 0]]
  }
}

@syntax {
  #-----------------------
  # returns string length
  #
  # "Hello world" length
  #-----------------------
  if {"[lindex $args 1]" == "length"} {
    return [string length [lindex $args 0]]
  }
}

@syntax {
  #-----------------------
  # returns list length
  #
  # "Hello world" llength
  #-----------------------
  if {"[lindex $args 1]" == "llength"} {
    return [llength [lindex $args 0]]
  }
}

#-------------------------
# with <object> {
#   <message1> <message2>
#   <message3> ...
# }
#-------------------------
proc with {object commands} {
  #set nb [expr [info level] -1]
  foreach command $commands {
    uplevel 1 $object $command
  }
  return
}


#-
#- end of syntax.tcl
#-


#-
#- beginning of path.tcl
#-

#-------------------------------
# extends or return search path 
#-------------------------------
if {![info exists @RUNTIME(path)]} {
  set @RUNTIME(path) .
}
if {![info exists @RUNTIME(packages)]} {
  set @RUNTIME(packages) {}
}

proc @path {{args {}}} {
  upvar #0 @RUNTIME RUNTIME

  if {![string length $args]} { return $RUNTIME(path) }
  set directories [join $args]

  foreach directory $directories {
    set dir1 $directory
    set dir2 [file join $RUNTIME(directory) $directory]

    foreach i {1 2} {
      if [file isdirectory [set dir$i]] {
        if {[file pathtype [set dir$i]] == "relative"} {
          set dir$i [file join [pwd] [set dir$i]]
        }
        if {![RUNTIME(path) contains: [set dir$i]]} {
          if {![RUNTIME(packages) contains: [file tail [set dir$i]]]} {
            lappend RUNTIME(path) [set dir$i]
          }
        }
        break
      }
    }
  }
  set RUNTIME(path)
}


#------------------------------------------
# remote library directory from path name
#------------------------------------------
proc @formatLibraryPathName {path} {
  upvar #0 @RUNTIME RUNTIME
  if [regexp (.*)([string trimright [file join $RUNTIME(directory) " "] " "])(.*) $path -> a b c] {
    set path $c
  }
  set path
}


#-
#- end of path.tcl
#-


#-
#- beginning of package.tcl
#-

#-------------------------------
# add a package into runtime 
#-------------------------------
proc @package {{args {}}} {
  upvar #0 @RUNTIME RUNTIME

  #- default is to return current package
  if {![string length $args]} { 
    set RUNTIME(package) 
  } else {
    #- a package name or a list of packages is specified
    set packages [join $args]
    foreach package $packages {
      #- if directory not found
      if {![file isdirectory $package]} {
        #- if package is not directly in the current library path
        if {![file isdirectory [file join $RUNTIME(directory) $package]]} {
          #- searching if package is under one of the runtime paths
          foreach path $RUNTIME(path) {
            if {"[file tail $path]" == "$package"} {
              if [file isdirectory $path] {
                set dir $path
              }
            } elseif [file isdirectory [file join $path $package]] {
              set dir [file join $path $package]
            }
          }
        } else {
          set dir [file join $RUNTIME(directory) $package]
        }
      } else {
        set dir $package
      }

      if {![RUNTIME(packages) contains: $package]} {
        lappend RUNTIME(packages) $package
      }
      set RUNTIME(package) $package
    }
  }
  set RUNTIME(package)
}

proc @packages {{args {}}} {
  upvar #0 @RUNTIME RUNTIME

  #- default is to return current packages list
  if {![string length $args]} { 
    set RUNTIME(packages) 
  } else {
    #- a package name or a list of packages is specified
    set packages [join $args]
    foreach package $packages {
      @package $package
    }
  }
  set RUNTIME(packages)
}

proc @getPackageFromClass class {
  set package [join [lreplace [split [file tail $class] .] end end] .]
}

proc @getPackageFromFileName fileName {
  if [regexp {(interface|implementation|protocol|\.)} \
    [set package [file tail [file dirname $fileName]]]] {
    set package [file tail [file dirname [file dirname $fileName]]]
  }
  set package
}

proc @getPackageForClass class {
  upvar #0 @RUNTIME RUNTIME
  #- package is specified in the class name
  set package [@getPackageFromClass $class]
  if [string length $package] {
    #- isolate class name from the package one
    set class [@getClassFromClass $class]
  } else {
    #- package is not specified in the class name 
    set packages {}
    #- taking package from runtime if class is already loaded
    foreach class2 $RUNTIME(classes) {
      if [string length [set pkg [@getPackageFromClass $class2]]] {
        set class2 [@getClassFromClass $class2]
      }
      if {"$class2" == "$class"} {
        lappend packages $pkg
      }
    }
    if {[set cnt [llength $packages]] == 1} {
      return $packages
    } elseif {$cnt > 1} {
      @error "must specify package name! (class found under packages: [join $packages {, }])"
    } else {
      #- taking package from current script file name
      set fileName [info script]
      set package [@getPackageFromFileName $fileName]
    }
  } 
  set package
}

if {![info exists @RUNTIME(packages)]} {
  set @RUNTIME(packages) {}
}

if {![info exists @RUNTIME(package)]} {
  set @RUNTIME(package) {}
}
@package core

#-
#- end of package.tcl
#-


#-
#- beginning of selector.tcl
#-

#-------------------------------
# extract selector from message
#-------------------------------
proc @selector { {message {}} } {
  set selector {}
  if [string length [string trim $message { }]] {
    if {![regexp : $message]} {
      set selector [lindex $message 0]
    } else {
      foreach {field value} $message {
        if [regexp : $field] { append selector $field }
      }
    }
  }
  set selector
}


#-
#- end of selector.tcl
#-


#-
#- beginning of args.tcl
#-

#--------------------------------
# extract arguments from message
#--------------------------------
proc @args { {message {}} } {
  set args {}
  if {![regexp : $message]} {
    set args [lrange $message 1 end]
  } elseif [string length [string trim $message { }]] {
    foreach {field value} $message {
      if {![regexp : $field]} { lappend args $field }
      lappend args $value
    }
  }
  set args
}

#-
#- end of args.tcl
#-


#-
#- beginning of class.tcl
#-


if {![info exists @RUNTIME(classes)]} {
  set @RUNTIME(classes) {}
  set @RUNTIME(categories) {}
  set @RUNTIME(protocols) {}
}

proc @getPathFromClass class {
  if [llength [file split $class]] {
    set exp [file tail $class]
    if [regexp [file join (.*) [file tail $class]] $class -> path] {
      set path
    } 
  }
}

proc @getClassFromClass class {
  set class [lindex [split [file tail $class] .] end]
}

#----------------------------
# loads a class into runtime
#----------------------------
proc @class {class} {
  upvar #0 @RUNTIME RUNTIME

  set code 0

  #- don't allow class name beginning with a @ character
  if [regexp {^@} $class] { return $code }

  set path [@getPathFromClass $class]
  set package [@getPackageFromClass $class]
  set class [@getClassFromClass $class]

  if [RUNTIME(classes) contains: ${class}] {
    return 1
  } elseif [RUNTIME(protocols) contains: ${class}] {
    return 1
  }

  #- if directory is specified, adding it to paths list if not yet in
  if {[string length $path] && ![RUNTIME(path) contains: $path]} {
    @path $path
  }

  if {[string length $package]} {
    @package $package 
  } else {
    if {[llength $RUNTIME(packages)] == 0} {
      lappend packages {}
    }
  }
  set paths {}
 
  foreach path $RUNTIME(path) {
    foreach p $RUNTIME(packages) {
      lappend paths [file join $RUNTIME(directory) $p]
      lappend paths [file join $RUNTIME(directory) $path $p]
      lappend paths [file join $path $p]
      set pkg([file join $RUNTIME(directory) $p]) $p
      set pkg([file join $RUNTIME(directory) $path $p]) $p
      set pkg([file join $path $p]) $p
    }
  }

  set files {}

  foreach path $paths {
    set _pkg [set pkg($path)]
    set i 0
    foreach v {interface implementation} {
      set $v [file join $path $v ${class}.tcl]
      if [file exists [set $v]] {
        if [catch {source [set $v]} err] {
          return [error $err]
        } else {
          incr i
        }
      }
    }
    if {$i == 0} { 
      #- no source found in interface/implementation folders 
      #- -> can be in only one file and outside
      #- the interface/implementation folders
      if [file exists [set f [file join $path ${class}.tcl]]] {
        if [catch {source $f} err] {
          return [error $err]
        } else {
          incr i
        }
      }
    }
    if {$i} {
      upvar #0 @$class CLASS
      if {![string length $CLASS(package)]} {
        set CLASS(package) $_pkg
      } else {
        set _pkg $CLASS(package)
      }
      if [string length $CLASS(implementation)] {
        interp alias "" "${_pkg}.$class" "" $class
      } else {
        #RUNTIME(protocols) addIfNotYetIn: $class
      } 
      return 1
    }
  }

  return $code
}



#-
#- end of class.tcl
#-


#-
#- beginning of interface.tcl
#-

#------------------------------------------------------------------------------
# parser for the class interface
#------------------------------------------------------------------------------
proc @interface { {args {}} } {
  upvar #0 @RUNTIME RUNTIME

  foreach v {
    class superclass superclasses protocols category
    class,variables class,selectors
    instance,variables instance,selectors
    interface,class,selectors interface,instance,selectors
  } {
    set $v {}
  }

  #- splitting args into realArgs and code
  if [regexp {^([^\{]+)[\{](.+)[\}]$} $args -> realArgs code] {
    #- parsing realArgs to get class, superclass, protocols, category if any
    if [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*<([^>]+)>[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class superclass protocols category] {
      #-1- class + superclass + protocols + category
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*<([^>]+)>[ ]*$} \
        $realArgs -> class superclass protocols] {
      #-2- class + superclass + protocols
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class superclass category] {
      #-3- class + superclass + category
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*$} \
        $realArgs -> class superclass] {
      #-4- class + superclass
    } elseif [regexp {^([^ ]+)[ ]*<([^>]+)>[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class protocols category] {
      #-5- class + protocols + category
    } elseif [regexp {^([^ ]+)[ ]*<([^>]+)>[ ]*$} \
        $realArgs -> class protocols] {
      #-6- class + protocols
    } elseif [regexp {^([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
        $realArgs -> class category] {
      #-7- class + category
    } elseif [regexp {^([^ ]+)[ ]*$} $realArgs -> class] {
      #-8- class
    } else {
      return [error "malformed interface"]
    }
  } elseif [regexp {^([^\{]*)[\{](.*)[\}]$} $args] {
    #- empty interface
  } else {
    return [error "malformed interface"]
  }

  set code [string trimright $code "\r\n "]

  #- removing comments from code
  set newcode ""
  foreach line [split $code \n] {
    if [regexp {^[ ]*[#](.*)$} $line] { continue }
    append newcode $line\n
  }
  set code $newcode

  #- multiple inheritence
  if [regexp , $superclass] {
    set superclasses [lreplace [split $superclass ,] 0 0]
    set superclass [lindex [split $superclass ,] 0]
  }

  #- parsing code and setting variables
  if [regexp {[\{][\r\n ]*(.*)[\}](.*)} $code -> variables code] {
    set variables [string trimright $variables "\n\r "]
    set instance,variables $variables
    if {[string length $category] && [string length ${instance,variables}]} {
      return [error "class:$class category:$category \
        -> can't add instance variables with a category !"]
    }
  }

  foreach line [split $code \;] {
    set line [string trim $line "\r\n "]
    if {![string length $line]} { continue }
    if [regexp {^([+-])[ ]*([^$]+)$} $line -> sign message] {
      set selector [@selector $message]
      set args [@args $message]
      switch -- $sign {
        + {
           set type class
        }
        - {
          set type instance
        }
        default { }
      }
      lappend ${type},selectors $selector
      lappend interface,${type},selectors $selector
    } else {
      set classVariables $line
      foreach v $classVariables {
        lappend class,variables $v
      }
    }
  }

  #- package name is specified
  if {[llength [split $class .]] > 1} {
    set package [join [lreplace [split $class .] end end] .]
    set class [lindex [split $class .] end]
  } else {
    set package {}
  }

  upvar #0 @${class} SELF

  foreach v {
    class superclass superclasses protocols categories
    class,variables class,selectors
    instance,variables instance,selectors
    interface implementation 
    interface,class,selectors interface,instance,selectors 
  } {
    if {![info exists SELF($v)]} {
      set SELF($v) {}
    }
  }

  set SELF(package) $package

  set SELF(class) $class
  set SELF(superclass) $superclass
  set SELF(superclasses) $superclasses

  set SELF(interface,class,selectors) ${interface,class,selectors}
  set SELF(interface,instance,selectors) ${interface,instance,selectors}

  if [string length $superclass] { 
    if {![@class $superclass]} {
      return [error \
        "couldn't import \"$superclass\" superclass for \"$class\" class"]
    }
    upvar #0 @${superclass} SUPER
    foreach type {instance class} {
      foreach field {variables selectors} {
        foreach value $SUPER(${type},${field}) {
          if {"$field" == "variables"} {
            if {![SELF(${type},${field}) contains: $value]} {
              lappend SELF(${type},${field}) $value
            }
          } else {
            lappend SELF(${type},${field}) $value
            lappend SELF(interface,${type},${field}) $value
          }
        }
      }
      foreach field [array names SUPER ${type},methodFor,*] {
        set SELF($field) $SUPER($field)
      }
    }
    foreach proto $SUPER(protocols) {
      if {![SELF(protocols) contains: $proto]} {
        lappend SELF(protocols) $proto
        lappend SELF(interface,protocols) $proto
      }
    }

    #- multiple_inheritence
    if {[info exists superclasses] && [llength $superclasses]} {
      set i 0
      foreach _super $superclasses {
        if {![@class $_super]} {
          return [error \
           "couldn't import \"$_super\" class for \"$class\" class"]
        }
        incr i  
        #@debug "importing $_super"
        upvar #0 @${_super} SUPER${i}
        foreach type {instance class} {
          foreach field {variables selectors} {
            foreach value [set SUPER${i}(${type},${field})] {
              if {"$field" == "variables"} {
                if {![SELF(${type},${field}) contains: $value]} {
                  lappend SELF(${type},${field}) $value
                } else {
                  @debug "warning: variable $value already exists in $class class when trying to import it from $_super"
                }
              } else {
                if {![SELF(${type},${field}) contains: $value]} {
                  lappend SELF(${type},${field}) $value
                } else {
                  #@debug "selector $value already exists when importing from $_super -> must be explicitly overridden"
                }
              }
            }
          }
          foreach field [array names SUPER${i} ${type},methodFor,*] {
            #- checking if it's already defined and if it's the same content else, 
            #- if other callback, warning
            if [info exists SELF($field)] {
              if {"[set SELF($field)]" != "[set SUPER${i}($field)]"} {
                @debug "[lindex [split $field ,] end] points to [set SELF($field)] in $class but points to [set SUPER${i}($field)] in [set SUPER${i}(class)]"
              }
            } 
            set SELF($field) [set SUPER${i}($field)]
          }
        }
        foreach proto [set SUPER${i}(protocols)] {
          if {![SELF(protocols) contains: $proto]} {
            lappend SELF(protocols) $proto
           }
        }
      }
    }
  }
  if [string length $category] {
    #- importing original class if it's a category
    if {![@class $class]} {
      return [error "couldn't import original $class definition"]
    }
    if {![SELF(categories) contains: $category]} {
      lappend SELF(categories) $category
    }
  }

  foreach protocol [split $protocols ,] {
    set protocol [string trim $protocol]
    if {![SELF(protocols) contains: $protocol]} {
      lappend SELF(protocols) $protocol
    }
  }
  foreach type1 {class instance} {
    foreach type2 {variables selectors} {
      foreach value [set ${type1},${type2}] {
        if {![SELF($type1,$type2) contains: $value]} {
          lappend SELF($type1,$type2) $value
        }
      }
    }
  }

  #- creating class variables
  foreach var $SELF(class,variables) {
    if [regexp {^@([^$]+)$} $var -> arrayName] {
      global @${class}@${arrayName}
      array set @${class}@${arrayName} {}
    } else {
      set SELF($var) {}
    }
  }

  if {![RUNTIME(protocols) contains: $class]} {
    lappend RUNTIME(protocols) $class
  }

  set SELF(interface) [@formatLibraryPathName [info script]]

}


#-
#- end of interface.tcl
#-


#-
#- beginning of variablesAccess.tcl
#-


#----------------------------------------------
# return code to get access to class variables 
#   (in a class method)
#----------------------------------------------

proc @getAccessToClassVariables {self} {
  upvar #0 @${self} SELF
  upvar #0 @$SELF(class) CLASS

  set txt "\nset super @sendMessageToSuper\n"

  set type class

  foreach field $CLASS(${type},variables) {
    if {![regexp {^@([^$]+)$} $field -> arrayName]} {
      append txt "upvar #0 @\${self}($field) $field\n"
    } else {
      append txt "upvar #0 @\${self}@${arrayName} $arrayName\n"
    }
  }

  set txt
}


#----------------------------------------------
# return code to get access to instance variables 
#   (in an instance method)
#----------------------------------------------

proc @getAccessToInstanceVariables {self} {
  upvar #0 @${self} SELF
  upvar #0 @$SELF(class) CLASS

  set txt "\nset super @sendMessageToSuper\n"

  set type instance

  foreach field $CLASS(${type},variables) {
    if {![regexp {^@([^$]+)$} $field -> arrayName]} {
      append txt "upvar #0 @\${self}($field) $field\n"
    } else {
      append txt "upvar #0 @\${self}@${arrayName} $arrayName\n"
    }
  }

  set txt
}


#-
#- end of variablesAccess.tcl
#-


#-
#- beginning of implementation.tcl
#-

#--------------------------------------
# parser for the class implementation
#--------------------------------------
proc @implementation { {args {}} } {
  upvar #0 @RUNTIME RUNTIME

  set category {}

  #- splitting args into real args and code
  if [regexp {^([^\{]+)[\{](.+)[\}]$} $args -> realArgs code] {
    #- parsing realArgs
    if [regexp {^([^ ]+)[ ]*\(([^\)]+)\)[ ]*$} \
      $realArgs -> class category] {
    } elseif [regexp {^([^ ]+)[ ]*:[ ]*([^ ]+)[ ]*$} \
        $realArgs -> class superclass] {
    } elseif [regexp {^([^ ]+)[ ]*$} $realArgs -> class] {
    }

    set code [string trimright $code "\r\n "]

    #- removing comments from code
    set newcode ""
    foreach line [split $code \n] {
      if [regexp {^[ ]*[#](.*)$} $line] { continue }
      append newcode $line\n
    }
    set code $newcode

    if [string length $category] {
      #- importing original class if it's a category
      if {![@class $class]} {
        return [error "couldn't import original $class definition"]
      }
    }

    #- package name is specified
    if {[llength [split $class .]] > 1} {
      set package [join [lreplace [split $class .] end end] .]
      set class [lindex [split $class .] end]
    } else {
      set package {}
    }

    upvar #0 @${class} SELF

    if [string length $SELF(superclass)] {
      upvar #0 @$SELF(superclass) SUPER
    }

    foreach {char selector args body} $code {
      switch -- $char {
        - { 
          set type instance
        }
        + {
          set type class
        }
        default { return [error "unknown char:$char"] }
      }
      #- args
      if {![string length $args]} { set args "{args {}}" }
      set args [linsert $args 0 self]

      if {"$type" == "instance"} {
        append newbody [@getAccessToInstanceVariables $class]
      } else {
        append newbody [@getAccessToClassVariables $class]
      }
      append newbody $body
      set body $newbody
      unset newbody
      append body {
        set self
      }

      #- method
      proc @${class},$type,$selector $args $body

      #- adding method in class description if not yet in
      if {![SELF($type,selectors) contains: $selector]} {
        lappend SELF($type,selectors) $selector
      }
      set SELF(${type},methodFor,${selector}) \
        @${class},$type,$selector
    }

    #- check if all methods are implemented
    foreach type {class instance} {
      #- getting only selectors that are not inherited
      set notInheritedSelectors {}
      if [info exists SUPER] {
        foreach selector $SELF($type,selectors) {
          if {![SUPER($type,selectors) contains: $selector]} {
            lappend notInheritedSelectors $selector
          }
        }
      } else {
        set notInheritedSelectors $SELF($type,selectors)
      }
      foreach selector $notInheritedSelectors {
        if {![string length [info proc @${class},$type,$selector]]} {
          #- can be implemented in an other superclass in a case of multiple inheritence
          if [llength $SELF(superclasses)] {
            set i 0
            foreach _super $SELF(superclasses) {
              incr i
              upvar #0 @$_super SUPER${i}
              if [info exists SUPER${i}] {
                if {![SUPER${i}($type,selectors) contains: $selector]} {
                  lappend notInheritedSelectors2 $selector
                }
              } else {
                lappend notInheritedSelectors2 $selector
              }
            }
            if [info exists notInheritedSelectors2] {
              if {[llength $notInheritedSelectors2] > 1} { set verb "are" } else { set verb "is" }
              return [error "uncompleted implementation for \"$class\" class:\n\
               $notInheritedSelectors2 selectors $verb not implemented !\n"]
            }
          } else {
            return [error "uncompleted implementation for \"$class\" class:\n\
              \"$selector\" selector is not implemented !\n"]
          }
        }
      }
    }

    #- check if protocol methods are implemented
    if [string length $SELF(protocols)] {
      foreach protocol $SELF(protocols) {
        if {![@class $protocol]} {
          return [error "unknown protocol $protocol for class $class"]
        }
        upvar #0 @${protocol} PROTOCOL
        foreach type {instance class} {
          foreach selector $PROTOCOL(interface,${type},selectors) {
            if {![SELF(${type},selectors) contains: $selector]} {
              return [error \
                "$class class must conform to $protocol protocol and \
                \"$selector\" selector is not implemented!"]
            }
          }
        }
      }
    }

    #- creating class callback
    proc $class { {args {}} } {
      return [@sendMessageToObject [lindex [split [lindex [info level 0] 0] .] end] $args]
    }

    #- adding class name to runtime
    if {![RUNTIME(classes) contains: $class]} {
      lappend RUNTIME(classes) $class
    }
    if {[RUNTIME(protocols) contains: $class]} {
      RUNTIME(protocols) remove: $class
    }

    set SELF(implementation) [@formatLibraryPathName [info script]]

    if [string length $category] {
      if {![SELF(categories) contains: $category]} {
        lappend SELF(categories) $category
      }
    }

  } else {
    return [error "malformed implementation"]
  }

}


#-
#- end of implementation.tcl
#-


#-
#- beginning of nextId.tcl
#-

#----------------------
# gets integer maximum 
#----------------------
proc @maxInteger {} {
  set i 1
  set index 1

  while {$i > 0} {
    set i [expr 1 << [incr index]]
  }
  expr 1 << [expr $index - 1]
}

uplevel #0 { 
  set @RUNTIME(maxInteger) [@maxInteger] 
  set @RUNTIME(lastId) 0
  set @RUNTIME(counter) 0
}

#--------------------------------
# gets next available object Id
#--------------------------------
proc @nextId {} {
  upvar #0 @RUNTIME(counter) counter
  upvar #0 @RUNTIME(lastId) lastId
  upvar #0 @RUNTIME(maxInteger) maxInteger

  set maxIntegerLength [string length $maxInteger]
  set len [expr $maxIntegerLength -1]
  set incrementing 1
  set index [expr [llength $counter] -1]

  while {$incrementing} {
    #- incrementing the current counter
    set count [expr [string trimleft [lindex $counter $index] 0] + 1]

    #- have we reached the limit ?
    if {[string length $count] == $maxIntegerLength} {
      #- reset it + report to next counter
      set counter [lreplace $counter $index $index [format %0${len}i 0]]
      incr index -1
      if {$index == -1} {
        #- next counter doesn't exist yet
        set counter [linsert $counter 0 0]
        set index 0
      }
    } else {
      #- no limit reached so, updating counter
      set counter [lreplace $counter $index $index [format %0${len}i $count]]
      set incrementing 0
    }
  }
  set lastId [string trimleft [join $counter {}] 0]
}

#-
#- end of nextId.tcl
#-


#-
#- beginning of newInstance.tcl
#-


#---------------------------------------------
# creates new instance from a class template
#---------------------------------------------
proc @newInstanceFromClass { class } {
  upvar #0 @$class CLASS

  set id id[@nextId]
  upvar #0 @$id ID

  foreach var $CLASS(instance,variables) {
    if [regexp {^@([^$]+)$} $var -> arrayName] {
      upvar #0 @${id}@${arrayName} myArray

      #- workaround to set the array under Windows
      set myArray(foo) bar
      unset myArray(foo)
    } else {
      set ID($var) {}
    }
  }
  set ID(class) $class

  proc $id { {args {}} } {
    @sendMessageToObject [lindex [info level 0] 0] $args
  }

  set id
}


#-
#- end of newInstance.tcl
#-


#-
#- beginning of callerLevel.tcl
#-

#------------------------------------------------------------------------------
# gets level of caller procedure (NB: works only when message without arguments)
#------------------------------------------------------------------------------
proc @getCallerLevel {} {
  set cmd "uplevel info level"
  set currentLevel [eval $cmd]
  set currentMessage [eval $cmd 0]
  if [regexp {^@} $currentMessage] {
    set selector [lindex [split [lindex $currentMessage 0] ,] end]
    set object [lindex $currentMessage end]
    set msg "$object $selector"
    set level $currentLevel
    while {$level > 0} {
      if [catch {
        set level [eval $cmd]
        set message [eval $cmd 0]
      } err] {
        return 0
      } elseif {"$msg" == "$message"} {
        #- maybe called by unknown
        set cmd "uplevel $cmd"
        if [catch {
          set level [eval $cmd]
          set message [eval $cmd 0]
        } err] {
          return $level
        } elseif {"unknown $msg" == "$message" \
           || "::unknown $msg" == "$message"} {
          return [expr $level -1]
        } else {
          return $level
        }
      }
      set cmd "uplevel $cmd"
      incr level -1
    }
  }
  return
}



#-
#- end of callerLevel.tcl
#-


#-
#- beginning of sendMessage.tcl
#-

#--------------------------------------------------------------
# callback used when a message is sent to a class or an object
#--------------------------------------------------------------
proc @sendMessageToObject {self {message {}}} {

  upvar #0 @${self}(class) class
  upvar #0 @${class} SELF

  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }

  set selector [@selector $message]
  set args [@args $message]

  #- empty message or ?
  switch $message {
    {} -
    ?  {
      set txt "$self respondsTo:\n"

      upvar #0 @${self} SELF

      if {"$type" == "instance"} {
        upvar #0 @$SELF(class) CLASS
        lappend sel $CLASS(instance,selectors)
        set sep -
      } else {
        lappend sel $SELF(class,selectors)
        set sep +
      }
      set sel [lsort [split [join $sel]]]
      foreach s $sel {
        append txt "\t${sep} $s\n"
      }
      return [error [string trimright $txt \n]]
    }
  }

  if {[info exists SELF(${type},methodFor,${selector})]} {
    set newCall $SELF(${type},methodFor,${selector}) 
    set level [info level]
    incr level -1
    #@debug "call: $SELF(${type},methodFor,${selector}) < $level) [uplevel $level info level 0]"
    eval $SELF(${type},methodFor,${selector}) $self $args
  } elseif {[info exists SELF(${type},methodFor,forwardInvocation:)]} {
    $SELF(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists SELF(${type},methodFor,forward:)]} {
    $SELF(${type},methodFor,forward:) $self $message
  } elseif {[info exists SELF(${type},methodFor,doesNotRecognize:)]} {
    $SELF(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SELF(${type},methodFor,error:)]} {
    $SELF(${type},methodFor,error:) $self \
      "$self cannot respond to \"$message\" message !"
  } else {
    error "$self cannot respond to \"$message\" message !"
  }

}

#------------------------------------------------------
# callback used when a message has to be sent to super
#------------------------------------------------------
proc @sendMessageToSuper { {args {}} } {

  #- get caller proc
  if [catch {set callerFullProcName [info level -1]} err] {
    return [error "cannot call super outside of a instance/class method ($err)"]
  }
  set callerProcName [lindex $callerFullProcName 0]
  set self [lindex $callerFullProcName 1]

  upvar #0 @${self}(class) class
  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }
  set message $args
  set selector [@selector $message]
  set args [@args $message]

  set _class [string trimleft [lindex [split $callerProcName ,] 0] @]
  upvar #0 @${_class} CLASS
  set superclass $CLASS(superclass)

  if {![string length $superclass]} {
    return [error "cannot call super method when superclass doesn't exist !"]
  }

  upvar #0 @$superclass SUPERCLASS
  if {[info exists SUPERCLASS(${type},methodFor,${selector})]} {
    eval $SUPERCLASS(${type},methodFor,${selector}) $self $args
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forwardInvocation:)]} {
    $SUPERCLASS(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forward:)]} {
    $SUPERCLASS(${type},methodFor,forward:) $self $message
  } elseif {[info exists SUPERCLASS(${type},methodFor,doesNotRecognize:)]} {
    $SUPERCLASS(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SUPERCLASS(${type},methodFor,error:)]} {
    $SUPERCLASS(${type},methodFor,error:) $self \
      "super of $self cannot respond to \"$message\" message !"
  } else {
    error "super of $self cannot respond to \"$message\" message !"
  }

}

proc @sendMessageToSuper@ { superclass {args {}} } {

  #- get caller proc
  if [catch {set callerFullProcName [info level -1]} err] {
    return [error "cannot call super outside of a instance/class method ($err)"]
  }
  set callerProcName [lindex $callerFullProcName 0]
  set self [lindex $callerFullProcName 1]

  upvar #0 @${self}(class) class
  if {"$self" == "$class"} {
    set type class
  } else {
    set type instance
  }
  set message $args
  set selector [@selector $message]
  set args [@args $message]

  if {![string length $superclass]} {
    return [error "cannot call super method when superclass doesn't exist !"]
  }

  upvar #0 @$superclass SUPERCLASS
  if {[info exists SUPERCLASS(${type},methodFor,${selector})]} {
    eval $SUPERCLASS(${type},methodFor,${selector}) $self $args
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forwardInvocation:)]} {
    $SUPERCLASS(${type},methodFor,forwardInvocation:) $self $message
  } elseif {[info exists \
    SUPERCLASS(${type},methodFor,forward:)]} {
    $SUPERCLASS(${type},methodFor,forward:) $self $message
  } elseif {[info exists SUPERCLASS(${type},methodFor,doesNotRecognize:)]} {
    $SUPERCLASS(${type},methodFor,doesNotRecognize:) $self [@selector $message]
  } elseif {[info exists SUPERCLASS(${type},methodFor,error:)]} {
    $SUPERCLASS(${type},methodFor,error:) $self \
      "super of $self cannot respond to \"$message\" message !"
  } else {
    error "super of $self cannot respond to \"$message\" message !"
  }

}


#-
#- end of sendMessage.tcl
#-


#-
#- beginning of import.tcl
#-

proc @import {string} {
  upvar #0 @RUNTIME RUNTIME

  set package [@getPackageFromClass $string]
  set class [@getClassFromClass $string]

  @package $package

  if [regexp {\*} $class] {
    #- listing all classes from package
    foreach path $RUNTIME(path) {
      set f {}
      if [file isdirectory [set f [file join $RUNTIME(directory) $path $package]]] {
      } elseif [file isdirectory [set f [file join $RUNTIME(directory) $package]]] {
      } elseif [file isdirectory [set f [file join $path $package]]] {
      }
    }
    if [string length $f] {
      set files {}
      set classes {}
      if [llength [set _files [glob -nocomplain [file join $f interface *.tcl]]]] {
        foreach f $_files { lappend files $f }
      }
      if [llength [set _files [glob -nocomplain [file join $f *.tcl]]]] {
        foreach f $_files { lappend files $f }
      }
      foreach file $files {
        set _class [join [lreplace [split [file tail $file] .] end end] .]
        lappend classes $_class
      }
      foreach _class $classes {  
        @class $_class
      }
    }
  } else {
    @class $class
  }
}


#-
#- end of import.tcl
#-


#-
#- beginning of dumpRuntime.tcl
#-


#-------------------------------
# make one bigfile from runtime
#-------------------------------
proc @dumpRuntime {} {
  upvar #0 @RUNTIME RUNTIME
  set files $RUNTIME(files)

  #- file containing the runtime load procedure called "@" must be in the first place
  set files [linsert $files 0 main.tcl]

  #- protocols
  foreach protocol $RUNTIME(protocols) {
    upvar #0 @$protocol PROTO
    files addIfNotYetIn: $PROTO(interface)
  }

  #- classes
  foreach class $RUNTIME(classes) {
    upvar #0 @$class CLASS
    foreach type {interface implementation} {
      foreach file [set CLASS($type)] {
        files addIfNotYetIn: $file
      }
    }
  }

  #- dumping runtime into 1 big file
  set out [open mOT.tcl w+]
  set i 1
  foreach file $files {
    puts stderr "$i) $file" ; incr i
    set f [open $file r]
    puts $out "\n#-\n#- beginning of $file\n#-\n"
    puts $out [read $f]
    puts $out "\n#-\n#- end of $file\n#-\n"
    close $f
  }

  close $out
}


#-
#- end of dumpRuntime.tcl
#-


#-
#- beginning of doc.tcl
#-



#------------------------------------------------------------------------------
# build the doc from the current runtime
#------------------------------------------------------------------------------
proc @makedoc {} {
  upvar #0 @RUNTIME RUNTIME

  #- creating directories
  if [catch {
    file mkdir doc
  } err] {
    puts stderr "$err" ; return
  }

  #- index.html - frameset
  if [catch {
    set f [open [file join doc index.html] w]
    puts $f {
      <HTML>
      <HEAD>
      <TITLE>myObjectiveTcl</TITLE>
      <META NAME="Description" CONTENT="Objective-C like framework in pure tcl">
      <META NAME="Keywords" 
        CONTENT="myObjectiveTcl Objective-C tcl tk object-oriented OO">
      </HEAD>
       <FRAMESET COLS="300,*" FRAMEBORDER=1 FRAMESPACING=1 BORDER=1 
         BORDERCOLOR=BLACK>
        <FRAME FRAMEBORDER=1 AUTOSCROLLING=1 SRC=packages.html NAME=left
          MARGINWIDTH=10 MARGINHEIGHT=10>
        <FRAME FRAMEBORDER=0 AUTOSCROLLING=1 SRC=main.html NAME=right
          MARGINWIDTH=40 MARGINHEIGHT=10>
       </FRAMESET>
      </HTML>
    }
    close $f
  } err] {
    puts stderr "$err" ; return
  }

  #- main.html - empty file for first
  set f [open [file join doc main.html] w]
  puts $f "<HTML><BODY BGCOLOR=WHITE></BODY></HTML>"
  close $f

  #- classes.html - empty file for first
  set f [open [file join doc classes.html] w]
  puts $f "<HTML><BODY BGCOLOR=WHITE></BODY></HTML>"
  close $f

  #- collecting informations about classes and sorting them
  foreach class $RUNTIME(classes) {
    upvar #0 @$class @$class

    #- sort classes into packages
    lappend classes([set @${class}(package)]) $class

    #- sort selectors/variables into class specific, overridden, inherited
    foreach type1 {instance class} {
      foreach type2 {selectors variables} {
        set field ${type1},${type2}
        set value [set @${class}($field)]
        if {![string length [set superclass [set @${class}(superclass)]]]} {
          #- no superclass -> all is specific
          set specific($class,$field) [set @${class}($field)]
        } else {
          foreach subvalue $value { 
            #- doesn't exists in superclass -> class specific
            if {![@${superclass}($field) contains: $subvalue]} {
              lappend specific($class,$field) $subvalue
            } elseif [regexp variables $field] {
              #- exist in superclass -> inherited for variables
              lappend inherited($class,$field) $subvalue
            } elseif [string length [info proc @$class,$type1,$subvalue]] {
              #- the selector is overridden by the current class
              lappend overridden($class,$field) $subvalue
            } else {
              #- just simple inherited
              lappend inherited($class,$field) $subvalue
            }
          }
        }
      }
    }
    
    #- going deeply in the class hierarchy    
    if {0} {
    set inheritedFrom($class) ""
    set superExists 1
    set super $class
    while {$superExists} {
      if [string length [set super [set @${super}(superclass)]]] {
        lappend inheritedFrom($class) $super
        if {![info exists @${super}]} {
          upvar #0 @${super} @${super}
        }
      } else {
        set superExists 0
        break
      }
    }
    }

    set inheritedFrom($class) [set @${class}(superclass)]
    #- multiple inheritence
    if [string length @${class}(superclasses)] {
      foreach c [set @${class}(superclasses)] {
        lappend inheritedFrom($class) $c
      }
    }

    if [string length [set inheritedFrom($class)]] {
      set i 0
      foreach c [set inheritedFrom($class)] {
        set pkg [set @${c}(package)] 
        if {"[set @${class}(package)]" == "."} { set pkg $pkg }

        if {$i} {
          append inheritedFromInHtml($class) \
            "\n,&nbsp;<A HREF=../../$pkg/classes/$c.html>$c</A>"
        } else {
          append inheritedFromInHtml($class) \
            "\n<A HREF=../../$pkg/classes/$c.html>$c</A>" 
        }
        incr i
      }
      set inheritedFrom($class) [set inheritedFromInHtml($class)]
      unset inheritedFromInHtml($class)
    }
  }

  #- collecting informations about protocols
  foreach proto $RUNTIME(protocols) {
    upvar #0 @$proto @$proto

    #- sort classes into packages
    lappend protocols([set @${proto}(package)]) $proto

    #- sort selectors into protocol specific and inherited
    foreach type1 {instance class} {
     foreach type2 {selectors} {
        set field ${type1},${type2}
        set value [set @${proto}(interface,$field)]

        if {![string length [set parents [set @${proto}(protocols)]]]} {
          #- no parent protocols -> all is specific
          set specific(protocol,$proto,$field) [set @${proto}(interface,$field)]
        } else {
          #- parent protocols exists, checking if there is inheritence
          foreach subvalue $value {
            set count 0
          
            foreach parent $parents {
              #- doesn't exists in superclass -> class specific
              if [@${parent}($field) contains: $subvalue] {
                incr count
              }
            }
            if { $count > 0 } {
              lappend inherited(protocol,$proto,$field) $subvalue
            } else {
              lappend specific(protocol,$proto,$field) $subvalue
            } 
          }
        }
      }
    }
    set inheritedFromProtocol($proto) [set @${proto}(protocols)]
    if [string length $inheritedFromProtocol($proto)] {
      set i 0
      foreach p $inheritedFromProtocol($proto) {
        set pkg [set @${p}(package)]
        if {"[set @${proto}(package)]" == "."} { set pkg $pkg }
        if {$i} {
          append inheritedFromInHtml($proto) \
            ",&nbsp;<A HREF=../../$pkg/protocols/$p.html>$p</A>"
        } else {
          append inheritedFromInHtml($proto) \
            "<A HREF=../../$pkg/protocols/$p.html>$p</A>"
        }
        incr i
      }
      set inheritedFromProtocol($proto) [set inheritedFromInHtml($proto)]
      unset inheritedFromInHtml($proto)
    }
   
  }
 
  #- packages.html - navigation throught packages
  set fd(left) [open [file join doc packages.html] w]
  puts $fd(left) {<HTML>}
  puts $fd(left) [@docHtmlHead]
  puts $fd(left) {
     <BODY BGCOLOR=#DEDEDE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR>
       <H1>myObjectiveTcl</H1><UL>
  }

  #- opening all packages file descriptor
  foreach package [set packages [lsort [array names classes]]] {
    #- creating package file 
    if [catch {
      file mkdir [file join doc $package]
      set fd($package) [open [file join doc $package index.html] w]
    } err] {
      puts stderr "$err" ; return
    }
    puts $fd($package) {<HTML>}
    puts $fd($package) [@docHtmlHead]
    puts $fd($package) {
     <BODY BGCOLOR=#DEDEDE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR>
       <H1>myObjectiveTcl</H1><UL>
    }
  }

  #- handlig html for each package
  foreach package $packages {
    #- adding link to main packages index
    puts $fd(left) "<LI><A HREF=$package/index.html>$package</A>"

    #- adding link to others package pages
    foreach p $packages {
      #- not for the current one
      if {"$p" != "$package"} {
        puts $fd($p) "<LI><A HREF=../$package/index.html>$package</A>"
      }
    }

    puts $fd($package) "<LI><A HREF=../packages.html>$package</A><UL>"
  
    if [catch {
      file mkdir [file join doc $package classes]
      file mkdir [file join doc $package classes interfaces]
      file mkdir [file join doc $package classes implementations]
      file mkdir [file join doc $package protocols]
      file mkdir [file join doc $package protocols sources]
      file mkdir [file join doc $package categories]
      file mkdir [file join doc $package categories interfaces]
      file mkdir [file join doc $package categories implementations]
      file mkdir [file join doc $package images]
      file mkdir [file join doc $package samples]
    } err] {
      puts stderr "$err" ; return
    }
    #- classes
    if {[info exists classes($package)] && [string length $classes($package)]} {
      puts $fd($package) "<LI>Classes"
      puts $fd($package) "<UL>"
      foreach class [lsort $classes($package)] {
        if {![string length [set @${class}(superclass)]]} {
          #- root class because of no superclass -> in bold
          puts $fd($package) "<LI>\
            <A HREF=classes/$class.html TARGET=right>\
            <B>$class</B></A>"
        } else {
          puts $fd($package) "<LI>\
            <A HREF=classes/$class.html TARGET=right>\
            $class</A>"
        }
      }
      puts $fd($package) "</UL>"
    }

    #- protocols
    if {[info exists protocols($package)] && \
      [string length $protocols($package)]} {
      puts $fd($package) "<LI>Protocols"
      puts $fd($package) "<UL>"
      foreach protocol [lsort $protocols($package)] {
        puts $fd($package) "<LI>\
          <A HREF=protocols/$protocol.html TARGET=right>\
          $protocol</A>"
      }
      puts $fd($package) "</UL>"
    }
    puts $fd($package) "</UL>"
  }
  puts $fd(left) "</UL>"

  puts $fd(left) {
     </BODY>
    </HTML>
  }
  close $fd(left)

  #- closing all packages
  foreach p $packages {
    puts $fd($p) "</UL></BODY></HTML>"
    close $fd($p)
  }


  #- classes/<class>.html 
  foreach class $RUNTIME(classes) {
    set package [set @${class}(package)]
    set f [open [file join doc $package classes $class.html] w]
    puts $f <HTML>
    puts $f [@docHtmlHead]
    puts $f {
       <BODY BGCOLOR=WHITE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR><BR>
    }
    puts $f "<H1>$class class</H1>\n\
      <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->\n\
      <TABLE BORDER=0>\n"
    if [string length [set inheritedFrom($class)]] {
      puts $f "<TR>\n\
         <TD ALIGN=RIGHT>inherited from: </TD>\n\
         <TD ALIGN=LEFT>[set inheritedFrom($class)]</TD>\n\
       </TR>"
    }

    if [string length [set @${class}(protocols)]] {
      puts $f "<TR><TD ALIGN=RIGHT>protocols:</TD><TD ALIGN=LEFT>"
      set h ""
      foreach p [set @${class}(protocols)] {
        if {![info exists @${p}(package)]} { upvar #0 @${p} @${p} }
        append h \
          "<A HREF=../../[set @${p}(package)]/protocols/${p}.html>$p</A>, "
      }
      puts $f "[string trimright $h {, }]</TD></TR>"
    }

    puts $f "\
       <TR>\n\
         <TD ALIGN=RIGHT>package:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=../../[set @${class}(package)]/index.html TARGET=left>\
             [set @${class}(package)]</A></TD>\n\
       </TR>\n\
    "

    puts $f "<TR><TD></TD><TD></TD></TR>"

    puts $f "<TR>\n\
         <TD ALIGN=RIGHT>declared in:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=interfaces/$class.html>[set @${class}(interface)]</A>\n\
         </TD>\n\
       </TR>\n\
       <TR>\n\
         <TD ALIGN=RIGHT>implemented in:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=implementations/$class.html>\
             [set @${class}(implementation)]</A></TD>\n\
       </TR>\n"

    puts $f "\
      </TABLE><BR><BR>\n\
      <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"

   #- sample code ?
   if [file exists [set sample \
       [file join $RUNTIME(directory) $package samples ${class}.tcl]]] {
      #- sample code ?
        set dest [open [file join doc $package samples ${class}.html] w]
        set src [open $sample r]
        puts $dest {
          <HTML>
        }
        puts $dest [@docHtmlHead]
        puts $dest {
          <BODY BGCOLOR=WHITE><BR><BR>
        }
        puts $dest "<H1>$class sample</H1>"
        puts $dest "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>--><BR>"
        puts $dest "<FONT color=#808080><PRE>\n"
        puts $dest [read $src]
        puts $dest "</PRE></BODY></HTML>"
        close $dest
        close $src
        if {![file exists [file join doc $package images sample.jpg]]} {
           file copy [file join $RUNTIME(directory) images sample.jpg] \
             [file join doc $package images sample.jpg]
        }
        puts $f "<A ALIGN=RIGHT HREF=../samples/${class}.html>\
          <IMG ALIGN=RIGHT BORDER=0 SRC=../images/sample.jpg></A>"
    }

    #- screenshots ?
    if [file exists [set img [file join $RUNTIME(directory) $package screenshots ${class}.jpg]]] {
      #- code ?
      if [file exists [set sample [file join $RUNTIME(directory) $package samples ${class}.tcl]]] {
        set dest [open [file join doc $package samples ${class}.html] w]
        set src [open $sample r]
        puts $dest {
        <HTML>
        }
        puts $dest [@docHtmlHead]
        puts $dest {
       <BODY BGCOLOR=WHITE>
       <BR><BR>
       }
        puts $dest "<H1>$class sample</H1>"
        puts $dest "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>--><BR>"
        puts $dest "<IMG BORDER=0 ALIGN=CENTER SRC=../images/${class}.jpg><BR><BR>"
        puts $dest "<FONT color=#808080><PRE>\n"
        puts $dest [read $src]
        puts $dest "</PRE></BODY></HTML>"
        close $dest 
        close $src

        #- screenshot to display in the main class page to link to the sample
        puts $f "<A ALIGN=RIGHT HREF=../samples/${class}.html>"
        puts $f "<IMG VALIGN=BOTTOM ALIGN=RIGHT BORDER=0 SRC=../images/${class}.jpg></A>"

      } else {
        #- no code
        puts $f "<IMG BORDER=0 ALIGN=RIGHT SRC=../images/${class}.jpg>"
      }

      if [file exists [file join doc $package images ${class}.jpg]] {
        file delete [file join doc $package images ${class}.jpg]
      }
      file copy $img [file join doc $package images ${class}.jpg]
    }

    puts $f "<UL>"
    foreach type2 {class instance} text2 {Class Instance} {
      foreach type3 {variables selectors} {
        foreach type1 {specific overridden inherited} \
          text1 {Specific} {
          if {[info exists ${type1}($class,$type2,$type3)] && \
            [string length [set ${type1}($class,$type2,$type3)]]} {
            set title "$type2 $type3"
            if {"$type1" == "specific"} {
              puts $f "<LI><U>$title</U>"
              puts $f "<UL>"
            }
            foreach value [lsort [set ${type1}($class,$type2,$type3)]] {
              if {"$type1" == "specific"} {
                puts $f "<LI>$value"
              }
              if {[info exists all_${type3}($class)] && \
                ([lsearch [set all_${type3}($class)] $value] == -1) || ![info exists all_${type3}($class)]} {
                lappend all_${type3}($class) $value
              }
            }
            if {"$type1" == "specific"} {
              puts $f "</UL>"
            }
          }
        }
      }
    }
    puts $f "</UL>"

    #- removed all variables/methods
    if {0} {
    puts $f "<HR WIDTH=100%><BR>"
    puts $f "<UL>"
    foreach type {variables selectors} {
      if {[info exists all_${type}($class)] && [llength [set all_${type}($class)]]} {
         puts $f "<LI><U>all $type</U>"
         puts $f "<UL>"
         foreach v [lsort [set all_${type}($class)]] {
           puts $f "<LI>$v"
         }
         puts $f "</UL>"
      }
    }
    puts $f "</UL>"
    }

    puts $f {
       </BODY>
      </HTML>
    }
    close $f

    #- interface & implementation sources
    foreach type {interface implementation} {
      if [catch {
        set f [open [file join doc $package classes \
          ${type}s $class.html] w]
        puts $f {
          <HTML>
        }
        puts $f [@docHtmlHead]
        puts $f {
          <BODY BGCOLOR=WHITE><BR><BR>
        }
        puts $f "<H1>[set @${class}($type)]</H1>"
        puts $f "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
        puts $f "<FONT color=#808080><PRE>\n"
        set ff [open [set @${class}($type)] r]
        set txt [read $ff]
        close $ff
        #- replacing < > with &lt; &gt;
        foreach {src dest} {< {\&lt;} > {\&gt;}} {
          regsub -all -- $src $txt $dest txt
        }
        puts $f $txt
        puts $f "</PRE></FONT></BODY></HTML>"
        close $f
      } err] {
        puts stderr "$err" ; return
      }
    }
  }

  foreach protocol $RUNTIME(protocols) {
    set package [set @${protocol}(package)]

    #- protocols/<protocol>.html
    if [catch {
      set f [open [file join doc $package protocols $protocol.html] w]
      puts $f {
        <HTML>
      }
      puts $f [@docHtmlHead]
      puts $f {
        <BODY BGCOLOR=WHITE LINK=#8080A0 ALINK=#8080A0 VLINK=#8080A0><BR><BR>
      }
      puts $f "<H1>$protocol protocol</H1>\n\
        <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->\n\
        <TABLE BORDER=0>\n"
      if [string length [set inheritedFromProtocol($protocol)]] {
        puts $f "<TR>\n\
          <TD ALIGN=RIGHT>inherited from:</TD>\n\
          <TD ALIGN=LEFT>[set inheritedFromProtocol($protocol)]</TD>\n\
          </TR>"
      }

    if [string length [set @${protocol}(protocols)]] {
      puts $f "<TR><TD ALIGN=RIGHT>protocols:</TD><TD ALIGN=LEFT>"
      set h ""
      foreach p [set @${protocol}(protocols)] {
        if {![info exists @${p}(package)]} { upvar #0 @${p} @${p} }
        append h \
          "<A HREF=../../[set @${p}(package)]/protocols/${p}.html>$p</A>, "
      }
      puts $f "[string trimright $h {, }]</TD></TR>"
    }

    puts $f "\
       <TR>\n\
         <TD ALIGN=RIGHT>package:</TD>\n\
         <TD ALIGN=LEFT>\
           <A HREF=../../[set @${protocol}(package)]/index.html TARGET=left>\
             [set @${protocol}(package)]</A></TD>\n\
       </TR>\n\
    "

      puts $f "<TR><TD></TD><TD></TD></TR>"
      puts $f "<TR>\n\
          <TD ALIGN=RIGHT>declared in:</TD>\n\
          <TD ALIGN=LEFT>\
            <A HREF=sources/$protocol.html>[set @${protocol}(interface)]</A>\n\
           </TD>\n\
          </TR>\n"


      puts $f "\
         </TABLE><BR><BR>\n\
        <!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"

      foreach type1 {specific inherited} text1 { {} Inherited} {
        foreach type3 {selectors} {
          foreach type2 {instance class} text2 {Instance Class} {
            if {[info exists ${type1}(protocol,$protocol,$type2,$type3)] && \
               [string length \
                 [set ${type1}(protocol,$protocol,$type2,$type3)]]} {
              if [string length $text1] {
                set title "$text1 $type2 $type3"
              } else {
                set title "$text2 $type3"
              }
              #puts $f "<U><B>$title</B></U>"
              puts $f "<U>$title</U>"
              puts $f "<UL>"
              foreach value [set ${type1}(protocol,$protocol,$type2,$type3)] {
                puts $f "<LI>$value"
              }
              puts $f "</UL><!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
            }
          }
        }
      }
      puts $f {
         </BODY>
        </HTML>
      }
      close $f
    } err] {
      puts stderr "$err" ; return
    }

    #- protocol source
    if [catch {
      set f [open [file join doc $package \
        protocols sources $protocol.html] w]
        puts $f "<HTML><BODY BGCOLOR=WHITE><BR><BR>"
        puts $f "<H1>[set @${protocol}(interface)]</H1>"
        puts $f "<!--<HR NOSHADE SIZE=1 BGCOLOR=BLACK WIDTH=100%>-->"
        puts $f "<FONT color=#808080><PRE>\n"
        set ff [open [set @${protocol}(interface)] r]
        set txt [read $ff]
        close $ff
        #- replacing < > with &lt; &gt;
        foreach {src dest} {< {\&lt;} > {\&gt;}} {
          regsub -all -- $src $txt $dest txt
        }
        puts $f $txt
        puts $f "</PRE></FONT></BODY></HTML>"
        close $f
    } err] {
      puts stderr "$err" ; return
    }
  }
}

proc @docHtmlHead {} {
  return {
    <HEAD>
     <TITLE>myObjectiveTcl</TITLE>
      <META NAME="Description" CONTENT="Objective-C like framework in pure tcl">
      <META NAME="Keywords"
        CONTENT="myObjectiveTcl Objective-C tcl tk object-oriented OO">
     <STYLE>
      BODY, TABLE, TR, TD, A:link, A:visited {
       font-size: 13 ;
       font-weight: normal ;
       font-family: "arial, helvetica" ;
       text-decoration: none ;
      }
      PRE {
        font-size: 12 ;
        color: #555555 ;
      }
      A:link, A:visited {
        text-decoration: none ;
        color: #8080B0
      }
      A:hover, A:active {
       text-decoration: underline ;
       color: #8080B0
      }
      LI {
       list-style-type: circle ;
      }
     </STYLE>
    </HEAD>
  }
}



#-
#- end of doc.tcl
#-


#-
#- beginning of tk.tcl
#-


#------------------------------------------------------------------------------
# tk specific
#------------------------------------------------------------------------------

proc @tk {} {
  upvar #0 @RUNTIME RUNTIME 
  global tcl_platform

  if [catch {winfo exists .} err] {
    return 0
  } 

  @package gui

  if {![info exists RUNTIME(images)]} {
    set RUNTIME(images) {}
  } elseif {[llength $RUNTIME(images)] > 0} {
    return 1
  }

  #----------------------------------------------------------------------------
  # ressources
  #----------------------------------------------------------------------------
  option add *highlightThickness 0
  option add *selectBorderWidth 0
  foreach widget {Button Scrollbar Listbox Entry Canvas Text Menu MenuButton} {
    option add *${widget}.borderWidth 1 
  }
  option add *Text.wrap none

  #- colors
  if {$tcl_platform(platform) == "windows"} {
    set bg [[frame ._s] cget -bg]
    destroy ._s
  } else {
    set bg #d0d0d0
    option add *.background $bg
    foreach w {Frame Label Button Canvas Scrollbar} {
      option add *${w}.background $bg
    }
  }
      
  foreach w {Listbox Text Entry} {
    option add *$w.background #f0f0f0
    option add *$w.foreground black 
    option add *$w.selectBackground #c0c0d0
    option add *$w.selectForeground black
  }
 
  #- removing bg change on focus for scrollbar
  option add *Scrollbar.activeBackground $bg
  option add *Button.activeBackground $bg
  option add *Button.highlightBackground $bg

  #- normalize font use 
  set w [label ._test]
  set defaultFont [$w cget -font]
  if {[set index [lsearch -exact $defaultFont bold]] != -1} {
    set defaultFont [lreplace $defaultFont $index $index]
  }
  destroy $w
 
  foreach class {Label Text Listbox Entry Button} { 
    option add *${class}*Font $defaultFont
  }

  @image down_arrow [image create bitmap \
    -data {
      #define down_arrow_width 11
      #define down_arrow_height 11
      static unsigned char down_arrow_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x03, 0xfc, 0x01,
        0xf8, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]
   
  @image up_arrow [image create bitmap \
    -data {
      #define up_arrow_width 11
      #define up_arrow_height 11
      static unsigned char up_arrow_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x70, 0x00, 0xf8, 0x00, 0xfc, 0x01,
        0xfe, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image left_arrow [image create bitmap \
    -data {
      #define left_arrow_width 11
      #define left_arrow_height 11
      static unsigned char left_arrow_bits[] = {
        0x00, 0x00, 0x40, 0x00, 0x60, 0x00, 0x70, 0x00, 0x78, 0x00, 0x7c, 0x00,
        0x78, 0x00, 0x70, 0x00, 0x60, 0x00, 0x40, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image right_arrow [image create bitmap \
    -data {
      #define right_arrow_width 11
      #define right_arrow_height 11
      static unsigned char right_arrow_bits[] = {
        0x00, 0x00, 0x10, 0x00, 0x30, 0x00, 0x70, 0x00, 0xf0, 0x00, 0xf0, 0x01,
        0xf0, 0x00, 0x70, 0x00, 0x30, 0x00, 0x10, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image return [image create bitmap \
    -data {
      #define return_width 15
      #define return_height 10
      static unsigned char return_bits[] = {
        0x10, 0x3c, 0x18, 0x04, 0x2c, 0x04, 0xe6, 0x07, 0x03, 0x00, 0x01, 0x00,
        0x02, 0x00, 0x04, 0x00, 0x08, 0x00, 0x10, 0x00, 
      };
    } -maskdata {
        #define returnmask_width 15
        #define returnmask_height 10
        static unsigned char returnmask_bits[] = {
          0x10, 0x7c, 0x18, 0x44, 0x2c, 0x44, 0xe6, 0x47, 0x03, 0x40, 0x01, 0x40,
          0xe2, 0x7f, 0x24, 0x00, 0x28, 0x00, 0x30, 0x00, 
        };
    } -background white -foreground black 
  ]
  
  @image radioOn [image create bitmap \
    -data {
      #define radioOn_width 16
      #define radioOn_height 16
      static unsigned char radioOn_bits[] = {
        0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
        0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x04, 0x00,
        0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
    } -maskdata {
        #define radioOnmask_width 16
        #define radioOnmask_height 16
        static unsigned char radioOnmask_bits[] = {
          0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0xc4, 0x07, 0xe4, 0x0f,
          0xf2, 0x1f, 0xf2, 0x9f, 0xf2, 0x9f, 0xf2, 0x9f, 0xf2, 0x9f, 0xe4, 0x4f,
          0xc4, 0x47, 0x00, 0x20, 0x20, 0x18, 0xc0, 0x07, };
    } -background white -foreground black
  ]
  
  @image radioOff [image create bitmap \
    -data {
      #define radioOff_width 16
      #define radioOff_height 16
      static unsigned char radioOff_bits[] = {
        0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
        0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x04, 0x00,
        0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    } -maskdata {
        #define radioOffmask_width 16
        #define radioOffmask_height 16
        static unsigned char radioOffmask_bits[] = {
          0x00, 0x00, 0xc0, 0x07, 0x30, 0x18, 0x08, 0x20, 0x04, 0x00, 0x04, 0x00,
          0x02, 0x00, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x04, 0x40,
          0x04, 0x40, 0x00, 0x20, 0x10, 0x18, 0xe0, 0x07, 
        };
    } -background white -foreground black
  ]
  
  @image empty16x16 [image create bitmap \
    -data {
      #define switchOff_width 16
      #define switchOff_height 16
      static unsigned char switchOff_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    }
  ]
  
  @image switchOff [@imageCopy empty16x16]
  @image switchOn [image create bitmap \
    -data {
      #define switchOn_width 16
      #define switchOn_height 16
      static unsigned char switchOn_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x0c,
        0x20, 0x06, 0x20, 0x03, 0xa0, 0x01, 0xc0, 0x00, 0x60, 0x00, 0x30, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      };
    } -maskdata {
        #define switchOnmask_width 16
        #define switchOnmask_height 16
        static unsigned char switchOnmask_bits[] = {
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x00, 0x0e,
          0x30, 0x07, 0xb0, 0x03, 0xf0, 0x01, 0xf0, 0x00, 0x70, 0x00, 0x30, 0x00,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        };
    } -background white -foreground black
  ]
   
  @image empty9x9 [image create bitmap \
    -data {
      #define empty_width 9
      #define empty_height 9
      static unsigned char empty_bits[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00
      };
    }
  ]
  
  @image empty [@imageCopy empty9x9]
  
  @image plus [image create bitmap \
    -data {
      #define plus_width 9
      #define plus_height 9
      static unsigned char plus_bits[] = {
        0xff, 0x01, 0x01, 0x01, 0x11, 0x01, 0x11, 0x01, 0x7d, 0x01, 0x11, 0x01,
        0x11, 0x01, 0x01, 0x01, 0xff, 0x01
      };
    }
  ]
  
  @image minus [image create bitmap \
    -data {
      #define minus_width 9
      #define minus_height 9
      static unsigned char minus_bits[] = {
        0xff, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x7d, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0xff, 0x01
      };
    }
  ]

set red_sphere {R0lGODlhDAAMAMYAALGRkYY+PnwREYwQEKs6OtKPj+Tf325AQGAKCnULC4kJ
CZwICKsHB7UGBsY5OfXe3mJBQVIODmkMDIAKCpcICKwHB78FBcwEBNADA9I3
N5yTkz8QEFcODm8MDIgKCqAICLcGBs4EBOECAugBAdoDA+WNjVNCQkAQEFkO
DnEMDIoJCaIHB7sFBdQDA+wBAfsAAOICAtU3NysYGD4QEIcKCp8ICLYGBs0E
BOACAuYCAtkDA8UKCicYGDkQEFEODpYICL0FBcoEBMUEBLgMDElDQzIREUkP
D18KCnQLC5oICKkHB7IGBq8GBrY6OpSUlCgSEj0QEGUKCncLC5MJCZgICMyP
jy8REUMPD1QODmQKCnILC3wLC4IKCoQKCpo8PODg4E1CQkIQEFAODlwKCmoM
DIk+Puvf35mUlFlCQkAWFkkVFXFAQLCRkf//////////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiIB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+mT8MQEE3QkNERUZHSB5JSksKTE1OT1ARUVI0
U0kLVFV/RFZXWFlaW1xdXoJfYEVhYmNRZGVmin9naNKoWcNGUSAAOw==}

set orange_sphere {R0lGODlhDAAMAMYAALGqlYZ2RnxlG4xwG6uSRtLDleTj4G5lRWBOFXVeFolt
F5x7F6uHF7WOGManR/Xw4GJbRFJDFWlVFoBmFpd3F6yIF7+WGMygGNCjGNKx
R5yalD80FVdHFW9ZFohsFqB+F7eQGM6hGOGwGei1GdqrGeXSllNQREA1FVlI
FXFbFopuF6KAF7uTGNSmGOy5GfvEGeKxGdWyRysmGj40FYdsFp9+F7aQGM2h
GOCvGea0GdmqGcWcHScjGjkwFFFCFWlUFpZ3F72VGMqeGMWbGLiSHUlIRDIr
FEk8FV9NFXRdFpp6F6mFF7KNGK+KF7abRpSUlCgjFD0zFWVSFndgFodrFpN1
F5x8F5h4F8y+lS8pFEM3FVRFFWRRFnJbFnxjFoJoFoRpFpqFRuDg4E1LRDIq
FEI3FVBCFVxLFWVRFmpVFol4Ruvo4JmYlFlUREA2G0k+G3FmRbCplf//////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+PxNADEFCN0NERUZHSEkeSktMCk1OT1BREVJT
VFVKVldYf0VZWltcXV5fYGGCYmNkZWZnaGlqa4p/bNq4eQMnjqJAADs=}

set yellow_sphere {R0lGODlhDAAMAMYAALGxkoaFPnx7EoyLEauqPNLSkOTk325uQWBfDnV0DImI
C5ybCqupCbWzCMbFOvX03mJiQlJRD2lpCoB/DJeWCqyrCL+9B8zKBtDOBtLS
OZyckz8+EFdWD29uCoiHC6CeCbe2CM7MBuHgBejmBNrYBeXkjlNTQkBAEFlY
D3FwCoqJC6KhCbu6B9TSBuzqBPv5A+LgBdXTOSsrGD4+EIeGC5+eCba1CM3L
BuDeBebkBNnXBcXDCicmGDk5EVFRD2loCpaVCr27B8rIBs3MBsXEB7i2DklJ
QzIyEUlIEF9eDnRzCpqZCqmoCbKxCK+uCLa1O5SUlCgnEj09EGVlDnd2DJOS
CpiXCszLkC8vEUNCEFRUD2RkDnJxCnx7DIKBC4SDC5qZPuDg4E1NQzIxEUJB
EFBQD1xbDmVkDmppComIPuvr35mZlFlZQkA/F0lJFnFxQLCwkv//////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+PxNADEFCQ0RFRkdISUoeS0xNCk5PUFFSEVNU
NFVLC1ZXf0ZYWVpbXF1eX2CCYWJjZGVmZ2hpaop/a9i0cfMGjqJAADs=}

set green_sphere {R0lGODlhDAAMAMYAAJyxklWGQDR8EziME2CrPqXSkOHk309uQShgDy51DjSJ
DDmcCz2rCj+1CmfGPOX13kxiQiRSECtpDjGACjeXDD2sCkK/CUbMCUfQCGrS
O5ackx8/ESZXDyxvDjOICjqgC0C3CkbOCEzhB03oB0raCKrlj0hTQiBAESZZ
Dy1xDjSKDDqiC0G7CkjUCE/sB1L7BkziB2vVOx0rGB8+ESZXEDOHCjmfC0C2
CkbNCEvgB03mB0nZCEjFDxwnGR45ESRREDeWDEK9CUXKCUTFCUS4EEVJQxwy
EiJJEChfDy50DjiaCzypCz+yCj6vCmO2PZSUlBkoEh89ESplDy93DjaTDDiY
DKPMkEZJQxsvEiBDESVUECpkDy1yDjB8CjKCCjKECluaPuDg4EZNQyBCESRQ
ECdcDytqDlaJP+Pr35WZlElZQiNAFyZJF1BxQZuwkv//////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzNB01Njc4OTo7PD0+P5lADEFCOENERUZHSEkeSktMCk1OT1BREVJT
NVRKC1VWf1dYWVpbXF1eX2CCYWJGY2RlUmZnaIp/adSsYdPGjaJAADs=}

set blue_sphere {R0lGODlhDAAMAMYAAJOisUFlhhVKfBVRjD92q5Gz0t/i5EJZbhA5YA9EdQ5O
iQ5XnApeqwpjtT6Ext/q9UJTYhEyUhA+aQ9JgA5UlwpfrAxovwxvzAtw0D6K
0pSYnBIpPxE1VxBBbw9NiA5ZoApktwxvzgt54Qp86At22pC85UNLUxIpQBE2
WRBCcQ5Oigpaogxmuwty1Ap/7AqG+wt64j6L1RkiKxIpPg9Nhw5Ynwpktgxv
zQt44At75gt12RJuxRkfJxImOREyURA9aQ5UlgxnvQxuygxrxRNouENGSRIi
MhEuSRA5Xw9DdA5WmgpdqQpisgpgrz58tpSUlBMdKBIoPRA8ZQ9Fdw9Mhw5S
kw5VmJGvzBIhLxErQxE0VBA7ZA9Ccg9HfA9Kgg9LhEBumuDg4ENJTRIqQhEx
UBA3XBA+akFmid/l65SWmUJOWRgsQBcxSUJacZOisP//////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+PxNADEFCN0NERUZHSEkeSktMCk1OT1BREVJT
VFVKC1ZXf0VYWVpbXF1eX2CCYWJGY2RlUmZnaIp/adSsYdPGjaJAADs=}

set white_sphere {R0lGODlhDAAMAMYAALGxsYaGhnx8fIyMjKurq9LS0uTk5G5ubmBgYHV1dYmJ
iZycnLW1tcbGxvX19WJiYlJSUmlpaYCAgJeXl6ysrL+/v8zMzNDQ0D8/P1dX
V29vb4iIiKCgoLe3t87OzuHh4ejo6Nra2uXl5VNTU0BAQFlZWXFxcYqKiqKi
oru7u9TU1Ozs7Pv7++Li4tXV1SsrKz4+PoeHh5+fn7a2ts3NzeDg4Obm5tnZ
2cXFxScnJzk5OVFRUZaWlr29vcrKyri4uElJSTIyMl9fX3R0dJqamqmpqbKy
sq+vr5SUlCgoKD09PWVlZXd3d5OTk5iYmC8vL0NDQ1RUVGRkZHJycoKCgoSE
hE1NTUJCQlBQUFxcXGpqauvr65mZmbCwsP//////////////////////////
////////////////////////////////////////////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHh4B/ggABAgME
BYKCBgcICQoLBAwKDoIPEBESExQVFheJCxgZGhscHR4fICEiIyQlJicoKSor
LC0uLzCjMTIzNDU2Nzg5OjuZPAQ9PjQ4P0BBQEJDG0RFRgxHM0hJShBLTDFN
RAtOFn9AT1BRUlMCVFVEgjVWQVdYWUtaCluKf1wlJICY6KIoEAA7}

set black_sphere {R0lGODlhDAAMAMYAAKOio2VlZUtKS1NSU3h4eLOzs+Li4llZWTo6OkVFRU9P
T1hYWGBfYGVkZYWFherq6lNTUzMzMz8/P1ZVVmFgYWppanFwcXNyc4yMjJiY
mCkpKTY1NkJCQk5OTlpaWmZmZnJxcnx7fH9+f3h3eL29vUxLTCoqKjY2NkND
Q1xbXGhoaHV0dYGAgYmIiY2MjSIiIjU1NUJBQmZlZnFxcXt6e359fnd2d3Bv
cCAgICcnJzMyMz8+P0pKSlVVVWlpaW5tbkZGRiMjIy4uLjo5OkRERE9OT1hX
WF9eX2RjZGJiYn19fZSUlB4eHikoKT09PU5NTlRTVFlYWVdWV7CwsCsrKzQ0
NDw8PEhISExMTODg4ElJSTIyMjg4OD08PWdmZ+Xl5ZaWliwsLDIxMltaW6Ki
ov//////////////////////////////////////////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERICExQVFhcYfxkaGxwdHh8gISIjJCUmJygKKSor
LC0hLi8aMDGmMjM0NTY3ODk6Ozw9DD43ID8VQEFCQ0RFRkdICklKS0xNEU5A
T1BGUVJTf0AvVFVWKFclWDeCWVpBVFtcXRJeX4p/YDqEETOGjKJAADs=}

set turquoise_sphere {R0lGODlhDAAMAMYAAJKxsECGhBR8eROMiD6rp5DS0N/k5EJubQ9gXQ51cgqJ
hgycmAurpgu1sDzGwt719EJiYRBSUA9pZwqAfQyXkwuspwq/uQnMxgnQyjzS
zpOcnBE/PRBXVQ5vbQqIhAygmwq3sgnOyAjh2wjo4gna1JDl4kJTUxFAPxBZ
Vw5xbgqKhgyingq7tgnUzgjs5gf79Aji3DvV0BgrKhE+PQ5vbAqHhAyfmwq2
sgnNxwjg2Qjm3wnZ0xDFwRknJhE5OBBRTw9pZg6AfAyWkgq9uAnKxAnNyArF
wBG4s0NJSRIyMRFJRw9fXA50cQqIhQyalguppAuyrguvqz62s5SUlBIoJxE9
PA9lYw53dAqHgwyTjwyYlJHMyhIvLxFDQRBUUw9kYg5ybw58eQqCfwqEgD6a
mODg4ENNTRFCQBBQTg9cWg9lYg9qZ0CJh9/r65SZmUJZWBdAPxdJSEFxcJKw
r////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHi4B/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzHDQ1Njc4OTo7PD0+P0BBQgxDREVGR0hJSktMTU5PUApRUlNUVRFW
V1hZTgtaW39IXF1eX2BhYmNkgmVmSWdoaWprbG2Kf9y8gRNHzhxFgQAAOw==}

set mauve_sphere {R0lGODlhDAAMAMYAALCTsYRChnoWfIkWjKhAq9GS0uTf5G1Cbl4RYHMQdYcP
iZkPnKgOq7EOtcM/xvTf9WFCYlARUmcRaX4QgJQPl6kOrLsOv8gKzMwK0NA+
0pyUnD4SP1URV20Qb4UQiJ0PoLQOt8oKzt0M4eMM6NYK2uOR5VNDUz8SQFcR
WW8QcYcPip8PorcOu9AK1OgM7PYM+94M4tI+1SoZKz0SPoUQh5wPn7MOtskK
zdsK4OEM5tQK2cETxSYZJzkSOVARUX0QgJMPlqcOq7kOvcYKysIKxbQUuElD
STESMkcSSV0RX3IQdIYQiJcPmqYOqa8OsqwOr7NAtpSUlCcTKDwSPVERUmQR
ZXUQd4QQh5APk5gPmpUPmMuSzC8TL0ISQ1MRVGMRZHAQcnoQfIAQgoEQhJhB
muDg4E1DTUESQk8RUFoRXGMRZWgRaohCievf65mUmVhDWT8YQEgYSXFCca+T
sP///////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkqKywt
Li8wMTIzpDQ1Njc4OTo7PD0+Ej9AQUJDN0RFRkdISUpLTE1OCk9QUVJTVFVW
V1hZC1pbf0ZcXV5fYGFiY2SCZWZHZ2hpamtsbYp/bt7AiSNnjqJAADs=}

set brown_sphere {R0lGODlhDAAMAMYAAKihl3FhSl1FI2hMJItxTr+wmuLh4GFXSEg2G1c/HWVI
IHFQIXxXI4NbJJ1+Ue7p4llRRj4vGk86HF9EH25OIX1YI4lgJZNmJpVoJ6eE
UpqYlTEnGEIyGlM9HWNHH3RSIoRdJJRnJ6FvKaZyKZxsKMy4nE5KRjInGEMy
G1Q9HXZTIodeJZhpJ6l0KrN7K6JwKSUhG0IxGlI8HYRcJKBvKKRxKZtsKJBm
KyIfGy4kF046HF5EH21OIYhfJZFlJpRnJo5jJoZgKUdGRCghFzgrGUc1G1Y/
HWRIH3BQIXpWI4FaJINcJH9ZI5N3T5SUlCEcFjAmGEw4HFhAHmtMIXJRIW9P
IbqtmkhGRCcgFjQpGEAxGks4HFU+HVxCHmBFH2FGH39qTeDg4EpIRTMoGD0v
GkU0G3RjSufl4ZiWlVJNRjQqHTsvHmNYSKegl///////////////////////
/////////////////////////////////////////////////yH+FUNyZWF0
ZWQgd2l0aCBUaGUgR0lNUAAh+QQBCgB/ACwAAAAADAAMAAAHiYB/ggABAgME
BYKCBgcICQoLDAoOD4IQERITFBUWFxgZfxobHB0eHyAhIiMkJSYnKCkKKiss
LS4vGTAbMTKmMxc0NTY3ODkROjs8DD0+P0BBQkNERUZHSElKS0xNTk9QEVFS
HlNIVFVWf1dYWVpbXF1eX2CCYWJDY2RlWxJmZ4p/aNKoWcOmjaJAADs=}

  foreach color {
    white black green yellow orange red blue brown mauve turquoise
  } {
    @image ${color}_sphere [image create photo -data [set ${color}_sphere]]
    unset ${color}_sphere
  }
  unset color

  @image sphere [@imageCopy white_sphere]
  
  @image bullet [image create bitmap \
    -data {
      #define bullet_width 6
      #define bullet_height 6
      static unsigned char bullet_bits[] = {
        0x1e, 0x3f, 0x07, 0x03, 0x03, 0x02, 
      };
    } -maskdata {
      #define bulletmask_width 6
      #define bulletmask_height 6
      static unsigned char bulletmask_bits[] = {
        0x1e, 0x3f, 0x3f, 0x3f, 0x3f, 0x1e, 
      };
    } -background white -foreground black
  ]
}
 

proc @image {{name {}} {image {}}} {
  upvar #0 @RUNTIME RUNTIME
  if [string length $image] {
    RUNTIME(images) add: [list $name $image]
    return $image
  } elseif {![string length $name]} {
    return $RUNTIME(images)
  } elseif [RUNTIME(images) contains: $name] {
    return [RUNTIME(images) get: $name]
  } else {
    set file [file join $RUNTIME(directory) images ${name}.*]
    if [catch {set files [glob $file]} err] {
      return
    } else {
      set file [lindex $files 0]
      RUNTIME(images) add: \
        [list $name [set img [image create photo -file $file]]]
      return $img
    }
  }
}

proc @imageCopy {image} {
  if {![string length [set img [@image $image]]]} {
    set img $image 
  }
  set newimg [image create [image type $img]]
  if [catch {
    foreach option [$img configure] {
      foreach {option_name val1 val2 val3 option_value} $option {
        $newimg configure ${option_name} ${option_value}
      }
    }
  } err] {
    return [error "couldn't get bitmap configuration ($err)"]
  }
  return $newimg
}

#- returns a more suitable list of available event data (from a bind command)
proc @eventData {} {
  return [list [list \
    path %W focus %f width %w height %h keyCode %k x %x y %y X %X Y %Y \
    button %b char %A event %E keySym %K keySymNumber %N \
    type %T delta %D
  ]]
}

#- return the code to execute for a Destroy event to remove bindings
proc @removeBindingsOnDestroyEvent {} {
  return { <Destroy> { foreach event [bind %W] { catch {bind %W $event {}} } } }
}

proc @showBitmaps {} {
  global @RUNTIME

  toplevel .toplevel 
  set col 0
  set row 0
  foreach {name img} [set @RUNTIME(images)] {
    grid [set f [frame .toplevel.f${col}${row} -relief sunken -bd 1]] \
      -sticky snew -padx 1 -pady 1 -column $col -row $row
    label $f.text -text "${name} : " -anchor w
    label $f.img -image $img 
    pack $f.text -side left -fill both -expand 1
    pack $f.img -side right -anchor e
    incr row
    if {![expr $row % 10]} {
      incr col 2 
      set row 0
    }
  }
}

@tk


#-
#- end of tk.tcl
#-


#-
#- beginning of core/interface/AutoRelease.tcl
#-

@interface AutoRelease {
  - refCount ;
  - retain ;
  - release ;
  - autorelease ;
}


#-
#- end of core/interface/AutoRelease.tcl
#-


#-
#- beginning of core/interface/Object.tcl
#-

@interface Object <AutoRelease> {
  {
     refCount 
  }

  + self ;
  + class ;
  + superclass ;

  + isKindOfClass: aClass ;
  + isMemberOfClass: aClass ;
  + forwardInvocation: aMessage ;
  + doesNotRecognize: aSelector ;
  + respondsTo: aSelector ;
  + conformsTo: aProtocol ;
  + error: aString ;

  + new ;
  + alloc ;

  - self ;
  - class ;
  - copy ;

  - forwardInvocation: aMessage ;
  - doesNotRecognize: aSelector ;
  - isKindOfClass: aClass ;
  - isMemberOfClass: aClass ;
  - conformsTo: aProtocol ;
  - respondsTo: aSelector ;
  - error: aString ;

  - init ;
  - dealloc ;

}


#-
#- end of core/interface/Object.tcl
#-


#-
#- beginning of core/implementation/Object.tcl
#-

@implementation Object {

  + new {} {
      return [[$self alloc] init]
    }

  + alloc {} {
      return [@newInstanceFromClass $self]
    }

  - init {} {
      set refCount 1
    }

  - copy {} {
      set _newInstance [[$self class] new]
      upvar #0 @[$self class] _Class
      upvar #0 @${_newInstance} _NewInstance
      foreach _var $_Class(instance,variables) {
        if [regexp {^@} $_var] {
          global @${_newInstance}@{_$var}
          array set @${_newInstance}@{_$var} [array get $_var]
        } else {
          set _NewInstance($_var) [set $_var]
        }
      }
      return $_newInstance
    }

  - refCount {} {
      return $refCount
    }

  - retain {} {
      incr refCount
    }

  - release {} {
      incr refCount -1
      if {$refCount <= 0} {
        return [$self dealloc]
      }
    }

  - autorelease {} {
      if [string length [set _level [@getCallerLevel]]] {
        if {![uplevel #${_level} info exists AutoReleasePool]} {
          return -code error "must first create an AutoReleasePool object \
            before to invoke autorelease msg!"
        }
        set _cmd "\$AutoReleasePool add: $self"
        uplevel #${_level} $_cmd
      } else {
        return -code error "couldn't find caller level ([info level 0])"
      }
    }

  + self {} {
      return $self
    }

  - self {} {
      return $self
    }

  + class {} {
      return $self
    }

  - class {} {
      upvar #0 @$self _SELF
      return $_SELF(class)
    }

  + superclass {} {
      upvar #0 @$self _CLASS
      return $_CLASS(superclass)
    }

  - superclass {} {
      return [[$self class] superclass]
    }

  + isMemberOfClass: aClass {
      if {![string compare $aClass [$self class]]} {
        return 1
      } else {
        return 0
      }
    }

  - isMemberOfClass: aClass {
      return [[$self class] isMemberOfClass: $aClass]
    }

  + isKindOfClass: aClass {
      set _superclass [$self superclass]
      while [string length $_superclass] {
        if {![string compare $_superclass $aClass]} {
          return 1
        } else {
          set _superclass [$_superclass superclass]
          if {![string length $_superclass]} { return 0 }
        }
      }
      return 0
    }

  - isKindOfClass: aClass {
      return [[$self class] isKindOfClass: $aClass]
    }

  + respondsTo: aSelector {
      upvar #0 @$self _SELF
      if [_SELF(class,selectors) contains: $aSelector] {
        return 1
      } else {
        return 0
      }
    }

  - respondsTo: aSelector {
      upvar #0 @[$self class] _CLASS
      if [_CLASS(instance,selectors) contains: $aSelector] {
        return 1
      } else {
        return 0
      }
    }

  + conformsTo: aProtocol {
      upvar #0 @$self _SELF
      if [_SELF(protocols) contains: $aProtocol] {
        return 1
      } else {
        return 0
      }
    }

  - conformsTo: aProtocol {
      return [[$self class] conformsTo: $aProtocol]
    }

  + forwardInvocation: aMessage {
      return [$self doesNotRecognize: [@selector $aMessage]]
    }

  - forwardInvocation: aMessage {
      return [$self doesNotRecognize: [@selector $aMessage]]
    }
 
  + doesNotRecognize: aSelector {
      return [$self error: \
        "$self cannot respond to \"$aSelector\" selector !"]
    }

  - doesNotRecognize: aSelector {
      return [$self error: \
        "$self cannot respond to \"$aSelector\" selector !"]
    }

  + error: aString {
      return -code error $aString
    }

  - error: aString {
      return -code error $aString
    }

  + notImplemented: aSelector {
      return [$self error: \
        "\"$aSelector\" selector is not yet implemented !"]
    }

  - notImplemented: aSelector {
      return [$self error: \
        "\"$aSelector\" selector is not yet implemented !"]
    }

  - dealloc {} {
      upvar #0 @${self} _SELF
      unset _SELF
      rename $self ""
      return 
    }

}


#-
#- end of core/implementation/Object.tcl
#-


#-
#- beginning of core/interface/AutoReleasePool.tcl
#-

@interface AutoReleasePool : Object {
  {
    # the pool instance variable contains all autoreleased objects
    pool
  }

  # to add an object to the pool
  - add: anObject ;

  # to remove an object from the pool
  - remove: anObject ;

  # to get the pool content
  - pool ;
}


#-
#- end of core/interface/AutoReleasePool.tcl
#-


#-
#- beginning of core/implementation/AutoReleasePool.tcl
#-

@implementation AutoReleasePool {

  + new {} {
      set _id [$self alloc]
      if [string length [set _level [@getCallerLevel]]] {
        set _cmd "set AutoReleasePool $_id"
        uplevel #${_level} $_cmd
        return [$_id init]
      } else {
        puts stderr "couldn't found caller level ([info level 0])"
        return [$_id dealloc]
      }
    }

  - add: anObject {
      if {![pool contains: $anObject]} {
        set pool [linsert $pool 0 $anObject]
      }
    }

  - remove: anObject {
      if [pool contains: $anObject] {
        pool remove: $anObject
      }
    }

  - dealloc {} {
      foreach _object $pool {
        catch {$_object release}
      }
      if [string length [set _level [@getCallerLevel]]] {
        set _cmd "catch {unset AutoReleasePool}"
        uplevel #${_level} $_cmd
      } else {
        return -code error "couldn't found caller level ([info level 0])"
      }
      return [$super dealloc]
    }

  - pool {} {
      return $pool
    }

}


#-
#- end of core/implementation/AutoReleasePool.tcl
#-


#-
#- beginning of core/interface/Connection.tcl
#-

@interface Connection : Object {
  {
     host port socket encrypted
  }
  - host: anHost ;
  - host ;

  - port: aPort ;
  - port ;

  - socket ;

  - isAlive ;
  - isEncrypted ;

  - enableEncryption ;
  - disableEncryption ;

  - send: message ;
  - receive ;
  - remoteInvocation: anInvocation ;

  - open ;
  - close ;
}



#-
#- end of core/interface/Connection.tcl
#-


#-
#- beginning of core/implementation/Connection.tcl
#-

@implementation Connection {
  - init {} {
      $super init
      set port 4000
      set host localhost
      set encrypted 1
    }

  - host: anHost {
      set host $anHost
    }

  - host {} {
      return $host
    }

  - port: aPort {
      set port $aPort
    }

  - port {} {
      return $port
    }

  - socket {} {
      return $socket
    }

  - isEncrypted {} {
      return $encrypted
    }

  - enableEncryption {} {
      set encrypted 1
    }

  - disableEncryption {} {
      set encrypted 0
    }

  - send: msg {
      if {$encrypted} {
        msg encrypt
      }
      puts $socket $msg
      flush $socket
    }
  - receive {} {
      set msg [gets $socket]
      if {$encrypted} {
        msg decrypt
      }
      return $msg
    }

  - remoteInvocation: { anInvocation } {
      set cmd {@RMI }
      append cmd $anInvocation
      if [$self isAlive] {
        $self send: $cmd
        return [$self receive]
      } else {
        return
      }
    }

  - open {} {
      if [catch {set socket [socket $host $port]} err] {
        return -code error "couldn't connect to $host at $port ($err)"
      } else {
        fconfigure $socket -buffering line -blocking yes
      }
    }

  - close {} {
      catch {close $socket}
      set socket {}
    }

  - isAlive {} {
      if {![string length $socket]} { return 0 }
      if [eof $socket] {
        $self close
        return 0
      } else {
        return 1
      }
    }

  - dealloc {} {
      $self close 
      return [$super dealloc]
    }
}


#-
#- end of core/implementation/Connection.tcl
#-


#-
#- beginning of core/interface/HtmlPage.tcl
#-

@interface HtmlPage : Object {
  {
    html head title style body frameset
    fileName
  }
 
  - head ;
  - title ;
  - style ;
  - body ;

  - head: aHead ;
  - title: aTitle ;
  - style: aStyle ;
  - body: aBody ;

  - headTag ;
  - titleTag ;
  - styleTag ;
  - bodyTag ;
  - htmlTag ;

  - add: anHtmlTag ;
  - tags ;
 
  - fileName: aFileName ;
  - fileName ;

  - writeToFile: aFileName ;
  - writeToFile ;

  + newWithFileName: aFileName ;

  - clear ;
}


#-
#- end of core/interface/HtmlPage.tcl
#-


#-
#- beginning of core/implementation/HtmlPage.tcl
#-

@implementation HtmlPage {
  - init {} {
      $super init
      set html [HtmlTag new: HTML]
      set head [HtmlTag new: HEAD]
      set title [HtmlTag new: TITLE]
      set style [HtmlTag new: STYLE]
      set body [HtmlTag new: BODY]
      set frameset [HtmlTag new: FRAMESET]
      $html add: $head
      $head add: $title
      $head add: $style
      $html add: $body
      $html add: $frameset
    }

  - head {} {
      return [$head value]
    }
  - title {} {
      return [$title value]
    }
  - style {} {
      return [$style value]
    }
  - body {} {
      return [$body value]
    }
  - frameset {} {
      return [$frameset value]
    }
  - html {} {
      return [$html value]
    }

  - head: aValue {
      $head value: $aValue
    }
  - title: aValue {
      $title value: $aValue
    }
  - style: aValue {
      $style value: $aValue
    }
  - body: aValue {
      $body value: $aValue
    }
  - frameset: aValue {
      $frameset value: $aValue
    }
  - html: aValue {
      $html value: $aValue
    }

  - headTag {} {
      return $head
    }

  - titleTag {} {
      return $title
    }

  - styleTag {} {
      return $style
    }

  - bodyTag {} {
      return $body
    }

  - framesetTag {} {
      return $frameset
    }

  - htmlTag {} {
      return $html
    }

  - add: anHtmlTag {
      $body add: $anHtmlTag
    }

  - tags {} {
      return [$body tags]
    }

  - fileName: aFileName {
      set fileName $aFileName
    }

  - fileName {} {
      return $fileName
    }

  - html {} {
      #- check if body of frameset is used
      if {![string length [$frameset attributes]] \
        && ![string length [$frameset value]]} {
        $html remove: $frameset
      } else {
        $html remove: $body
      }
      return [$html html]
    }

  - writeToFile: aFileName {
      set f [open $aFileName w]
      puts $f [$self html]
      close $f
    }

  - writeToFile {} {
      return [$self writeToFile: $fileName]
    }

  + newWithFileName: aFileName {
      return [[$self new] fileName: $aFileName]
    }

  - clear {} {

    }
}


#-
#- end of core/implementation/HtmlPage.tcl
#-


#-
#- beginning of core/interface/HtmlTag.tcl
#-

@interface HtmlTag : Object {
  {
    type attributes value closed tags
  }

  + new: tagName ;

  + new: tagName withAttributes: anAttributesString ;
  + new: tagName withValue: aValue ;
  + new: tagName withAttributes: anAttributesString andValue: aValue ;
  + new: tagName withValue: aValue andAttributes: anAttributesString ;

  - type: tagName ;
  - type ;

  - value: aValue ;
  - value ;

  - attributes: anAttributesString ;
  - attributes ;

  - dontClose ;

  - add: aTagObject ;
  - remove: aTagObject ;

  - html ;

  - tags ;
  - tagsOfType: aTagName ;
  - allTagsOfType: aTagName ;
}


#-
#- end of core/interface/HtmlTag.tcl
#-


#-
#- beginning of core/implementation/HtmlTag.tcl
#-


@implementation HtmlTag {

  + new: { tagName } {
      return [[$self new] type: $tagName]
    }

  + new:withAttributes: { tagName anAttributesString } {
      return [[[$self new] type: $tagName] \
                    attributes: $anAttributesString]
    }

  + new:withValue: { tagName aValue } {
      return [[[$self new] type: $tagName] \
                         value: $aValue]
    }

  + new:withAttributes:andValue: { tagName anAttributesString aValue } {
      return [[[[$self new] type: $tagName] \
                     attributes: $anAttributesString] \
                          value: $aValue]
    }

  + new:withValue:andAttributes: { tagName aValue anAttributesString } {
      return [[[[$self new] type: $tagName] \
                     attributes: $anAttributesString] \
                          value: $aValue]
    }

  - init {} {
      $super init
      set closed 1
      set value ""
      set type HTML
    }

  - type: { tagName } {
      set type $tagName
      switch -exact -- [string toupper $tagName] {
        BR { $self dontClose }
        BODY { 
          $self attributes: \
           "BGCOLOR=WHITE TEXT=BLACK LINK=8080C0 VLINK=8080C0 ALINK=808C0" 
        }
        STYLE { 
          $self value: {
            BODY, TABLE, TR, TD, A:link, A:visited {
              font-size: 13 ;
              font-weight: normal ;
              font-family: "arial, helvetica" ;
              text-decoration: none ;
              background-color: white ;
            }
            A:link, A:visited, A:active {
              color: #8080C0
            }
            A:hover, A:active {
              text-decoration: underline ;
            }
            LI {
              list-style-type: circle;
            }
          }
        }
        UL { $self attributes: "TYPE=CIRCLE" }
        TABLE { 
          $self attributes: \
            "BORDER=0 CELLSPACING=0 CELLPADDING=0 WIDTH=100% ALIGN=CENTER"
        }
        TD { $self attributes: "ALIGN=CENTER VALIGN=CENTER" }
        TR { $self attributes: "ALIGN=CENTER VALIGN=CENTER" }
      }
    }
  - type {} {
      return $type
    }

  - value: { aValue } {
      set value $aValue
    }
  - value {} {
      return $value
    }

  - attributes: { anAttributesString } {
      set attributes $anAttributesString
    }
  - attributes {} {
      return $attributes
    }

  - dontClose {} {
      set closed 0
    }

  - add: aTagObject {
      lappend tags $aTagObject
    }

  - remove: aTagObject {
      if [tags contains: $aTagObject] {
        tags remove: $aTagObject
      }
    }

  - html {} {
      set txt "<$type"
      if [string length $attributes] { 
        append txt " "
        append txt $attributes
      }
      append txt ">\n"
      if [string length $value] {
        append txt $value
        append txt "\n"
      }
      if [llength $tags] {
        foreach subtag $tags {
          append txt [$subtag html]
        }
      }
      if {$closed} { 
        append txt "</$type>"
        append txt "\n"
      }
      return $txt
    }
  - dealloc {} { 
      foreach subtag $tags {
        $subtag release
      }
      return [$super dealloc]
    }

  - tags {} {
      return $tags
    }

  - tagsOfType: aTagName {
      set subtags ""
      foreach tag $tags {
        if {"[string toupper [$tag type]]" == "[string toupper $aTagName]"} {
          lappend subtags $tag
        }
      }
      return $subtags
    }

  - allTagsOfType: aTagName {
      set subtags [$self tagsOfType: $aTagName]
      foreach tag $tags {
        if [string length [set t [$tag allTagsOfType: $aTagName]]] {
          lappend subtags $t
          set subtags [join $subtags]
        }
      }
      return $subtags
    }
}


#-
#- end of core/implementation/HtmlTag.tcl
#-


#-
#- beginning of core/interface/Server.tcl
#-

@interface Server : Connection {
  { 
    logfile
  }
  - start ;
  - stop ;
  - status ;

  - logfile: aFileName ;
  - logfile ;
}


#-
#- end of core/interface/Server.tcl
#-


#-
#- beginning of core/implementation/Server.tcl
#-

@implementation Server {
 - init {} {
     $super init
     set logfile server.log
   }

 - start {} {
      if [catch {set socket \
        [socket -server [list $self handleNewConnection] $port]} err] {
        puts stderr "couldn't bind server socket at port $port ($err)"
        return
      }
      if [catch {winfo geometry .} err] {
        set forever 1
        catch {vwait forever}
      }
    }

  - stop {} {
      return [$self close]
    }

  - status {} {
      if [$self isAlive] { return up } else { return down }
    }

  - handleNewConnection {aSocket anAddress aPort} {
      fconfigure $aSocket -buffering line
      fileevent $aSocket r "$self getMessageFromSocket: $aSocket \
                                               address: $anAddress \
                                               andPort: $aPort"
    }

  - getMessageFromSocket:address:andPort: {aSocket anAddress aPort} {
      if [catch {gets $aSocket line} err] {
        catch {close $aSocket}
      }
      if [eof $aSocket] {
        catch {close $aSocket}
        return
      }
      if {$encrypted} { line decrypt }
      return [$self handleRequest: $line \
                       fromSocket: $aSocket \
                          address: $anAddress \
                          andPort: $aPort]
    }

  - handleRequest:fromSocket:address:andPort: \
      {aRequest aSocket anAddress aPort} {
      puts stderr "request:\"$aRequest\" from: $anAddress port: $aPort"
      switch -exact -- [string tolower $aRequest] {
        quit -
        exit -
        default { catch {close $aSocket} }
      }
    }

  - logfile: aFileName {
      set logfile $aFileName
    }

  - logfile {} {
      return $logfile
    }

  - send:to: {msg sock} {
      if {$encrypted} {
        msg encrypt
      }
      puts $sock $msg
      flush $sock
    }
}


#-
#- end of core/implementation/Server.tcl
#-


#-
#- beginning of core/interface/HttpServer.tcl
#-

@interface HttpServer : Server {
  {
    root logfile
  }
  - root: aRootDirectory ;
  - root ;

  - sendPage: anHtmlPage \
    toSocket: aSocket \
  forAddress: anAddress \
     andPort: aPort ;

}



#-
#- end of core/interface/HttpServer.tcl
#-


#-
#- beginning of core/implementation/HttpServer.tcl
#-

@implementation HttpServer {
  - init {} {
      $super init
      set port 80
      set encrypted 0
      set root web
      set logfile HttpServer.log
    }

  - root: aRootDirectory {
      set root $aRootDirectory
    }

  - root {} {
      return $root
    }

  - handleRequest:fromSocket:address:andPort: {
      aRequest aSocket anAddress aPort
    } {
      set msg "$anAddress:$aPort -> "
      append msg $aRequest
      @log $logfile $msg

      set request [lindex [split $aRequest] 0]

      if {![string compare [string toupper $request] GET]} {
        set page [lreplace $aRequest 0 0]
        set page [lreplace $page end end]
        set mode [lindex [split $aRequest] end]

        #- removing heading /
        if [regexp {^[/]+([^$]+)$} $page -> page] {}

        return [$self sendPage: $page \
                      toSocket: $aSocket \
                    forAddress: $anAddress \
                       andPort: $aPort]
      }
      catch {
        puts $aSocket \
          "<HTML><BODY BGCOLOR=WHITE><FONT COLOR=RED>Error</FONT></BODY></HTML>"
        close $aSocket
      }
      return
    }

  - sendPage:toSocket:forAddress:andPort: {
      aPage aSocket anAddress aPort
    } {
      if {"$aPage" == "/"} {
        set aPage [file join $root index.html]
      } elseif [file exists [file join $root $aPage]] {
        set aPage [file join $root $aPage]
      }
      if [file owned $aPage] {
        if [catch {
          set f [open $aPage r]
          switch [file extension $aPage] {
            default { 
              fconfigure $f -translation binary 
              fconfigure $aSocket -translation binary 
            }
          }
          puts $aSocket [read $f]
          close $f
          close $aSocket
        } err] {
          @log $logfile "$err"
        } else {
          return
        }
      }
      catch {close $aSocket}
      return
    }
}



#-
#- end of core/implementation/HttpServer.tcl
#-


#-
#- beginning of core/interface/RemoteInvocationServer.tcl
#-

@interface RemoteInvocationServer : Server {
  {
     pool 
     @commands 
  }
  - executeCommand: aCommand \
        fromSocket: aSocket \
           address: anAddress \
           andPort: aPort ;

  - permit: aCommand ;
  - deny: aCommand ;

  - permittedCommands ;
  - permittedCommands: permittedCommands ;

  - deniedCommands ;
  - deniedCommands: deniedCommands ;

}



#-
#- end of core/interface/RemoteInvocationServer.tcl
#-


#-
#- beginning of core/implementation/RemoteInvocationServer.tcl
#-

@implementation RemoteInvocationServer {
 - init {} {
     $super init
     set logfile RemoteInvocationServer.log
     set pool [AutoReleasePool new]
     set commands(permit) [set commands(deny) {}]
   }

 - handleRequest:fromSocket:address:andPort: \
     {aRequest aSocket anAddress aPort} {
      switch -regexp -- [string tolower $aRequest] {
        {^@rmi } {
          set cmd [join [lrange [split $aRequest] 1 end]]
          set err [$self executeCommand: $cmd \
                             fromSocket: $aSocket \
                                address: $anAddress \
                                andPort: $aPort]
          $self send: $err to: $aSocket
        }
        default {
          $self send: "unknown request" to: $aSocket
        }
      }
    }

  - executeCommand:fromSocket:address:andPort: \
      {aCommand aSocket anAddress aPort} {

      set result "command not allowed"

      set msg "$anAddress:$aPort -> "
      append msg $aCommand
      @log $logfile $msg

      foreach cmd $commands(deny) {
        if [regexp $cmd $aCommand] {
          return $result
        }
      }
      foreach cmd $commands(permit) {
        if [regexp $cmd $aCommand] {
          catch {eval $aCommand} result
          return $result
        }
      } 
      return $result
    }

  - dealloc {} {
      $pool release
      return [$super dealloc]
    }

  - permit: aCommand {
      if {![commands(permit) contains: $aCommand]} {
        lappend commands(permit) $aCommand
      }
    }

  - deny: aCommand {
      if {![commands(deny) contains: $aCommand]} {
        lappend commands(deny) $aCommand
      }
    }

  - permittedCommands {} {
      return $commands(permit)
    }

  - permittedCommands: permittedCommands {
      set commands(permit) $permittedCommands
    }

  - deniedCommands {} {
      return $commands(deny)
    }

  - deniedCommands: deniedCommands {
      set commands(deny) $deniedCommands
    }
}


#-
#- end of core/implementation/RemoteInvocationServer.tcl
#-


#-
#- beginning of core/interface/RemoteObject.tcl
#-

@interface RemoteObject {
  {
    remoteId connection
  }
  + newWithHost: aHost andCommand: aCommand ;
  + newWithHost: aHost port: aPort andCommand: aCommand ;
  + newWithConnection: aConnection andCommand: aCommand ;

  - forwardInvocation: anInvocation ;
  - dealloc ;
}


#-
#- end of core/interface/RemoteObject.tcl
#-


#-
#- beginning of core/implementation/RemoteObject.tcl
#-

@implementation RemoteObject {

  + newWithHost:andCommand: { aHost aCommand } {
      return [$self newWithConnection: [[Connection new] host: $aHost] \
                           andCommand: $aCommand]
    }

  + newWithHost:port:andCommand: { aHost aPort aCommand } {
      return [$self newWithConnection: \
              [[[Connection new] host: $aHost] \
                                 port: $aPort] \
                           andCommand: $aCommand]
    }

  + newWithConnection:andCommand: { aConnection aCommand } {
      #- make sure the connection is opened
      if {![$aConnection isAlive]} {
        if [catch {$aConnection open} err] {
          return -code error $err
        }
      }
      set remoteId [$aConnection remoteInvocation: $aCommand]

      if {![string length $remoteId] || ![regexp id* $remoteId]} {
        return -code error "couldn't create remote object ! \
          (got \"$remoteId\" from server)"
      }

      set id [@newInstanceFromClass $self]
      upvar #0 @$id ID
      set ID(connection) $aConnection
      set ID(remoteId) $remoteId
      return $id
    }

  - forwardInvocation: { { anInvocation {} } } {
      #- make sure the connection is opened
      if {![$connection isAlive] && ![$connection open]} {
        return -code error "couldn't open connection \
          with [$connection host] at port [$connection port]"
      }
      set remoteInvocation "$remoteId "
      append remoteInvocation $anInvocation
      return [$connection remoteInvocation: $remoteInvocation]
    }

  - release {} {
      return [$self dealloc]
    }

  - dealloc {} {
      if [$connection isAlive] {
        $connection remoteInvocation: "$remoteId dealloc"
      }
      $connection release

      upvar #0 @$self SELF
      rename $self {}
      unset SELF
      return
    }
}


#-
#- end of core/implementation/RemoteObject.tcl
#-


#-
#- beginning of core/interface/Sequence.tcl
#-

@interface Sequence : Object {
  { 
    value counter length
  }
  - next ;
  - value ;
  - value: aValue ;
}


#-
#- end of core/interface/Sequence.tcl
#-


#-
#- beginning of core/implementation/Sequence.tcl
#-

@implementation Sequence {

  - init {} {
      $super init
      set length [expr [string length [@maxInteger]] -1]
      $self value: 0
    }

  - next {} {
      set incrementing 1
      set index [expr [llength $counter] -1]
      while {$incrementing} {
        #- incrementing the current counter
        set count [expr [string trimleft [lindex $counter $index] 0] + 1]
        #- have we reached the limit ?
        if {[string length $count] == [expr $length + 1]} {
          #- reset it + report to next counter
          set counter [lreplace $counter $index $index [format %0${length}i 0]]
          incr index -1
          if {$index == -1} {
            #- next counter doesn't exist yet
            set counter [linsert $counter 0 0]
            set index 0
          }
        } else {
          #- no limit reached so, updating counter
          set counter \
            [lreplace $counter $index $index [format %0${length}i $count]]
          set incrementing 0
        }
      }
      set value [string trimleft [join $counter {}] 0]
      return $value
    }

  - value {} {
      return $value
    }

  - value: aValue {
      set value $aValue
      set counter [format %0${length}i $aValue]
    }
}


#-
#- end of core/implementation/Sequence.tcl
#-


#-
#- beginning of core/interface/Tree.tcl
#-

@interface Tree : Object {
  {
    root rootChilds separator @nodes @childs @properties
  }
  - root: rootName ;
  - root ;

  - add: aNode ;
  - remove: aNode ;
  - contains: aNode ;

  - parentOf: aNode ;
  - childsOf: aNode ;
  - allChildsOf: aNode ;

  - nodeNameOf: aNode ;
  - levelOf: aNode ;

  - loadFromTree: aTree ;

  - loadFromFile: aFile ;
  - saveToFile: aFile ;

  - property: propertyName of: aNode ;
  - property: propertyName of: aNode to: aValue ;
}


#-
#- end of core/interface/Tree.tcl
#-


#-
#- beginning of core/implementation/Tree.tcl
#-

@implementation Tree {
  - init {} {
      $super init 
      set separator /
      set root /
    }

  - root: rootName {
      #- changing existing child info
      if {[string length $rootName] && ("$rootName" != "$root")} {
        if [string length [set oldroot $root]] {
          foreach field [array names properties ${root},*] {
            set f [join [lreplace [split $field ,] 0 0] ,]
            set properties($rootName,$f) [set properties($field)]
            unset properties($field)
          }
        } 
        set root $rootName
      }
    }

  - root {} {
      return $root
    }

  - separator: aCharacter {
      if {"$separator" != "$aCharacter"} {
        #- changing existing info
        foreach array {nodes childs} {
          foreach field [array names $array] {
            if {[llength [split $field $separator]] > 1} {
              set ${array}([join [split $field $separator] $aCharacter]) [set ${array}($field)]
              unset ${array}($field)
            }
          }
        }
        foreach field [array names properties] {
          set propertyName [lindex [split $field ,] end] 
          set aPath [join [lreplace [split $field ,] end end] ,]
          if {[llength [split $aPath $separator]] > 1 && "$aPath" != "$aCharacter" && "$aPath" != "$separator"} {
            set ${array}([join [split $aPath $separator] $aCharacter],$propertyName) \
              [set properties($field)]
            unset properties($field)
          }
        }
        set separator $aCharacter
      }
    }

  - separator {} {
      return $separator
    }

  - add: aNode { 
      set aNode [$self format: $aNode]

      if {[$self contains: $aNode]} {
        return [error "([info level 0]) \"$aNode\" already exist"]
      }
     
      set nodeName [$self nodeNameOf: $aNode]
      set parent [$self parentOf: $aNode]

      #- checking parent existence
      if {"$parent" == "$root"} {
        lappend rootChilds $nodeName 
      } else {
        if {![$self contains: $parent]} {
          return [error "([info level 0]) \"$parent\" doesn't exist"]
        } else {
          lappend childs($parent) $nodeName
        }
      } 
      set nodes($aNode) $nodeName
    }

   - addFull: aNode {
      set aNode [$self format: $aNode]
       set parent $aNode
       set parents {}
       while 1 {
         set parent \
           [join [lreplace [split $parent $separator] end end] $separator]
         if {![string length $parent]} { break }
         set parents [linsert $parents 0 $parent]
       }
       foreach parent $parents {
         if {![$self contains: $parent]} {
           $self add: $parent
         }
       }
       if {![$self contains: $aNode]} {
          $self add: $aNode
       }
     }

  - remove: aNode {
      set aNode [$self format: $aNode]
      if {"$aNode" == "$root"} { $self clear ; $self root: / ; return $self }

      #- already gone ?
      if {![$self contains: $aNode]} { 
        return [error "([info level 0]) \"$aNode\" doesn't exist"]
      }
      
      set nodeName [$self nodeNameOf: $aNode]
      set parent [$self parentOf: $aNode]

      #- removing node link from parent childs list
      if {"$parent" != "$root"} {
        if [string length $parent] {
           childs($parent) remove: $nodeName
        }
      } else {
        rootChilds remove: $nodeName
      }

      #- removing node
      unset nodes($aNode)

      #- removing childs
      if [info exists childs($aNode)] {
        foreach node $childs($aNode) {
          $self remove: [join [list $aNode $node] $separator]
        }
        unset childs($aNode)
      }

      #- removing properties
      foreach field [array names properties $aNode,*] {
        unset properties($field)
      }
    }

  - parentOf: aNode {
      set aNode [$self format: $aNode]
      if {"$aNode" == "$root"} { return }

      set parent [join [lreplace [split $aNode $separator] end end] $separator]
      if {![string length $parent]} {
        return $root
      } else {
        return $parent
      }
    } 

  - childsOf: aNode {
      set aNode [$self format: $aNode]
      if {"$aNode" == "$root"} {
        return $rootChilds
      }

      if {![$self contains: $aNode]} {
        return [error "([info level 0]) \"$aNode\" doesn't exist"]
      }
      set childsOfNode {}
      if [info exists childs($aNode)] {
        foreach node $childs($aNode) {
          lappend childsOfNode [join [list $aNode $node] $separator]
        }
      }
      return $childsOfNode
    }

  - allChilds {} {
       return [array names nodes]
     }

  - allChildsOf: aNode {
      set aNode [$self format: $aNode]

      if {"$aNode" == "$root"} {
        return [array names nodes]
      }

      if [$self contains: $aNode] {
        if {"$aNode" != "$separator"} {
           return [array names nodes [join [list ${aNode} *] $separator]]
        } else {
          return [array names nodes ${aNode}?*]
        }
      }
      return
    }

  - contains: aNode {
      set aNode [$self format: $aNode]
      if {"$aNode" == "$root"} { return 1 }
      if [info exists nodes($aNode)] { return 1 } else { return 0 }
    }

  - loadFromTree: aTree {
      $self clear
      $self root: [$aTree root]
      $self separator: [$aTree separator]
      foreach child [$aTree allChilds] {
        $self add: $child 
      }
    }

  - loadFromFile: aFile {
      set f [open $aFile r]
      while {![eof $f]} {
        set line [string trim [gets $f]]
        if [string length $line] {
          $self add: $line
        }
      }
      close $f
    }

  - saveToFile: aFile {
      set f [open $aFile w]
      foreach node [lsort [$self allChildsOf: $root]] {
        puts $f $node
      }
      close $f
    }

  - clear {} {
      foreach array {nodes childs properties} {
        foreach field [array names $array] {
          unset ${array}($field)
        }
      }
      $self root: $root
    }

  - nodeNameOf: aNode {
      set aNode [$self format: $aNode]
      if {"$aNode" == "$separator" && "$aNode" == "$root"} {
        return $aNode
      }
      return [lindex [split $aNode $separator] end]
    }

  - levelOf: aNode {
      set aNode [$self format: $aNode]
      if {"$aNode" == "$root"} { return 0 }
      set nodeLevel [llength [split $aNode $separator]]
      #incr nodeLevel 1
      return $nodeLevel
    }

  - property:of: {propertyName aNode} {
      set aNode [$self format: $aNode]
      if [info exists properties($aNode,$propertyName)] {
        return $properties($aNode,$propertyName)
      } else {
        return [error "([info level 0]) \"$aNode\" doesn't have any \"$propertyName\" property"]
      }
    }

  - property:of:to: {propertyName aNode aValue} {
      set aNode [$self format: $aNode]
      set properties($aNode,$propertyName) $aValue
    }

  - loadFromDirectory: aDirectory {
      if {![$self contains: $aDirectory]} {
        $self addFull: $aDirectory
      } 
      foreach child [glob -nocomplain $aDirectory/*] {
        $self loadFromDirectory: $child
      }
    }

  - format: aNode {
      if {"$aNode" != "/"} {
        if [regexp [join [list $root (.*)] $separator] $aNode -> node] {
          if [string length $node] {
            set aNode $node
          }
        }
        set aNode [string trimleft $aNode /]
      }
      return $aNode
    }
}


#-
#- end of core/implementation/Tree.tcl
#-


#-
#- beginning of gui/interface/Widget.tcl
#-

@interface Widget : Object {
  {
    path @callbacks @defaults
  }

  + newWithPath: aPath ;
  - initWithPath: aPath ;

  - path ;

  - show ;
  - hide ;

  - cursor: aCursorName ;
  - cursor ;

  - on: event do: callback ;
  - do: event with: eventData ;

  - addCallback: callback forEvent: event ;
  - callbackForEvent: event atIndex: index ;
  - removeCallbackForEvent: event atIndex: index ;

  - callbackIndexesForEvent: event ;

  - enableEvent: eventList on: pathList ;
  - disableEvent: eventList on: pathList ;
}


#-
#- end of gui/interface/Widget.tcl
#-


#-
#- beginning of gui/implementation/Widget.tcl
#-

@implementation Widget {

  + newWithPath: aPath {
      return [[$self alloc] initWithPath: $aPath]
    }

  - initWithPath: aPath {
      set path $aPath
      return [$self init]
    }

  - init {} {
      $super init
      if {![string length $path]} {
        set path .$self
      }
      frame $path
      set callbacks(index) [Sequence new]
    }

  - path {} {
      return $path
    }

  - show {} {
      pack $path -fill both -expand 1
    }

  - hide {} {
      catch {[winfo manager $path] forget $path}
    }

  - dealloc {} {
      catch {destroy $path}
      return [$super dealloc]
    }

  - cursor: name {
      $path configure -cursor $name
      update idletasks
    }

  - cursor {} {
      return [$path cget -cursor]
    }

  - addCallback:forEvent: {callback event} {
      #- adding access to instance variables if not already done for this event
      if {![llength [array names callbacks $event,*]]} {
        set index [$callbacks(index) next]
        set callbacks($event,$index) [@getAccessToInstanceVariables $self]
      }
      set index [$callbacks(index) next]
      set callbacks($event,$index) $callback
      return $index
    }

  - removeCallbackForEvent:atIndex: {event index} {
      catch {unset callbacks($event,$index)}
    }

  - on:do: {event callback} {
      return [$self addCallback: $callback forEvent: $event]
    }

  - do:with: {event eventData} {
      if [string length [set fields [array names callbacks $event,*]]] {
        foreach field [lsort $fields] {
          eval $callbacks($field)
        }
      }
    }

  - callbackIndexesForEvent: event {
      set indexes {}
      set fields [array names callbacks $event,*]
      foreach field $field {
        lappend indexes [join [lreplace [split $field ,] 0 0] ,]
      }
      return $indexes
    }

  - callbackForEvent:atIndex: {event index} {
      if [info exists callbacks($event,$index)] {
        return $callbacks($event,$index)
      }
      return
    }

  - enableEvent:on: {eventList pathList} {
      foreach w $pathList {
        foreach e $eventList {
          bind $w <$e> "$self do: $e with: [@eventData]"
          eval bind $w [@removeBindingsOnDestroyEvent]
        }
      }
    }

  - disableEvent:on: {eventList pathList} {
      foreach w $pathList {
        foreach e $eventList {
          bind $w <$e> ""
        }
      }
    }
}


#-
#- end of gui/implementation/Widget.tcl
#-


#-
#- beginning of gui/interface/Box.tcl
#-

@interface Box : Widget {
  {
    frame label border
  }

  - label: aLabel ;
  - label ;

  - showLabel ;
  - hideLabel ;

  - showBorder ;
  - hideBorder ;

  - frame ;

  - background: aColor ;
  - background ;
}


#-
#- end of gui/interface/Box.tcl
#-


#-
#- beginning of gui/implementation/Box.tcl
#-

@implementation Box {
  - init {} {
      $super init

      set border [frame $path.border -relief groove -bd 2]
      set frame [frame $path.border.inner]
      set label [label $path.label]

      set font [$label cget -font]
      if {[set index [lsearch -exact $font bold]] != -1} {
        $label configure -font [lreplace $font $index $index]
      }

      set padxy [expr [winfo reqheight $label]/2]
      pack $border $frame -fill both -expand 1 \
        -padx $padxy -pady $padxy
    }

  - label: aLabel {
      $label configure -text $aLabel
      if [string length $aLabel] {
        $self showLabel
      } else {
        $self hideLabel
      }
    }
  - label {} {
      return [$label cget -text]
    }

  - showLabel {} {
      $self showBorder
      place $label -x 20 -rely -0.0
    }
  - hideLabel {} {
      place forget $label
    }

  - showBorder {} {
      set padxy [expr [winfo reqheight $label]/2]
      pack configure $frame -padx $padxy -pady $padxy
      pack configure $border -padx $padxy -pady $padxy
      $border configure -bd 2 -relief groove
    }
  - hideBorder {} {
      $self hideLabel
      $border configure -bd 0 -relief flat
      pack configure $frame -padx 1 -pady 1
      pack configure $border -padx 1 -pady 1
    }

  - frame {} {
      return $frame
    }

  - background: aColor {
      foreach w [list $path $border $frame $label] {
        $w configure -background $aColor
      }
    }

  - background {} {
      return [$path cget -background]
    }
}


#-
#- end of gui/implementation/Box.tcl
#-


#-
#- beginning of gui/interface/CanvasBox.tcl
#-

@interface CanvasBox : Box {
  {
    canvas 
    xscroll yscroll scrollLock 
    zoom zoomingId zoomFactor
  }
  - busy ;
  - unbusy ;
}


#-
#- end of gui/interface/CanvasBox.tcl
#-


#-
#- beginning of gui/implementation/CanvasBox.tcl
#-

@implementation CanvasBox {
  - init {} {
      $super init
      $super hideBorder 
      set zoom 0
      $self zoomFactor: 2

      set bg [[entry $path._e] cget -background]
      destroy $path._e

      $frame configure -relief sunken -bd 1

      set canvas [canvas $frame.canvas -width 1 -height 1 -bd 1 -relief flat]
      $canvas configure -bg $bg

      set xscroll [scrollbar $frame.xscroll -orient horizontal \
        -command "$canvas xview"]
      set yscroll [scrollbar $frame.yscroll -orient vertical \
        -command "$canvas yview"]

      grid rowconfigure $frame 0 -weight 1
      grid columnconfigure $frame 0 -weight 1
      grid $canvas $yscroll -sticky snew -padx 0 -pady 0
      grid $xscroll -sticky snew -padx 0 -pady 0

      $canvas configure -xscrollcommand "$xscroll set" \
        -yscrollcommand "$yscroll set"

      $self enableEvent: Configure on: $canvas
      $self on: Configure do: "$canvas configure -scrollregion \[$canvas bbox all\]"

      $self enableEvent: {Button-4 Button-5} on: $canvas
      $self on: Button-4 do: {$canvas yview scroll -1 units}
      $self on: Button-5 do: {$canvas yview scroll +1 units}
      
      $self enableEvent: {
        Alt-ButtonPress-1 Alt-ButtonPress-3 
        Alt-ButtonRelease-1 Alt-ButtonRelease-3
      } on: $canvas

      $self on: Alt-ButtonPress-1 do: {$self startZooming}
      $self on: Alt-ButtonPress-3 do: {$self startUnZooming}
      $self on: Alt-ButtonRelease-1 do: {$self stopZooming}
      $self on: Alt-ButtonRelease-3 do: {$self stopZooming}
    }

  - canvas {} {
      return $canvas
    }

  - zoomFactor: newFactor {
      set newFactor [expr 1.0 * $newFactor]
      if {($newFactor != 0.0) && ($newFactor != 1.0)} {
        $self zoomReset
        set zoomFactor $newFactor 
      }
    }

  - zoomFactor {} {
      return $zoomFactor
    }

  - zoom {} {
      $canvas scale all 0 0 $zoomFactor $zoomFactor
      $canvas configure -scrollregion [$canvas bbox all]
      incr zoom
    }

  - unzoom {} {
      set factor [expr 1.0/$zoomFactor]
      $canvas scale all 0 0 $factor $factor
      $canvas configure -scrollregion [$canvas bbox all]
      incr zoom -1
    }

  - zoomSync {} {
      #- to do after a canvas reset
      if {[set oldzoom $zoom] != 0} {
        set zoom 0
        if {$oldzoom > 0} {
          while {$zoom < $oldzoom} {
            $self zoom
          }
        } elseif {$oldzoom < 0} {
          while {$zoom > $oldzoom} {
            $self unzoom
          }
        }
      }
    }

  - zoomReset {} {
      if {$zoom != 0} {
        if {$zoom > 0} {
          while {$zoom > 0} { 
            $self unzoom
          }
        } elseif {$zoom < 0} {
          while {$zoom < 0} { 
            $self zoom
          }
        }
      }
    }

  - startZooming {} {
      $self stopZooming
      $self zooming
    }

  - zooming {} {
      $self zoom
      set cmd "$self zooming"
      set zoomingId [after 200 $cmd]
    }

  - startUnZooming {} {
      $self stopZooming
      $self unZooming
    }

  - unZooming {} {
      $self unzoom
      set cmd "$self unZooming"
      set zoomingId [after 200 $cmd]
    }

  - stopZooming {} {
      if {[string length $zoomingId]} {
        after cancel $zoomingId
        set zoomingId {}
      }
    }

  - canvasBackground: aColor {
      $canvas configure -bg $aColor
      $frame configure -bg $aColor
    }

  - canvasBackground {} {
      return [$canvas cget -background]
    }

  - busy {} {
      $canvas configure -cursor watch
    }
  - unbusy {} {
      $canvas configure -cursor {}
    }
}


#-
#- end of gui/implementation/CanvasBox.tcl
#-


#-
#- beginning of gui/interface/Choice.tcl
#-

@interface Choice : Widget {
  {
    value label image imageOn imageOff
  }

  - value ;
  - value: boolean ;

  - label ;
  - label: aLabel ;

  - type ;
  - type: radioOrSwitch ;

  - onChange: aCallback ;
  - onChange ;
}


#-
#- end of gui/interface/Choice.tcl
#-


#-
#- beginning of gui/implementation/Choice.tcl
#-

@implementation Choice {
  - init {} {
      $super init
      set value 0
      set type radio
      set imageOn [@image radioOn]
      set imageOff [@image radioOff]

      grid rowconfigure $path 0 -weight 1
      grid columnconfigure $path 0 -weight 1
      grid [set label [label $path.label -text {} -anchor e]] \
        -row 0 -column 0 -sticky snew
      grid [set image [label $path.image -image $imageOff -bd 1 -relief flat]] \
        -row 0 -column 1 -sticky ew -padx 2 -pady 2

      $self enableEvent: ButtonPress-1 on: [list $image $label]
      $self on: ButtonPress-1 do: {$self toogle}

      set callbacks(Change) {}
    }
  - toogle {} {
      if {$value} {
        $self value: 0
      } else {
        $self value: 1
      }
    }
  - value {} {
      return $value
    }
  - value: boolean {
      set value $boolean
      if {$value} {
        $image configure -image $imageOn
      } else {
        $image configure -image $imageOff
      }
      $self valueChange
    }
  - label {} {
      return [$label cget -text]
    }
  - label: aLabel {
      $label configure -text $aLabel
    }
  - type {} {
      return $type
    }
  - type: radioOrSwitch {
      switch -- [string tolower $radioOrSwitch] {
        radio {
          set type radio
          set imageOn [@image radioOn]
          set imageOff [@image radioOff]
          $image configure -relief flat -bd 1
          $self value: [$self value]
        }
        switch {
          set type switch
          set imageOn [@image switchOn]
          set imageOff [@image switchOff]
          $image configure -relief raised -bd 1
          $self value: [$self value]
        }
        * { return [error "$self: unknown type :\"$radioOrSwitch\""] }
      }
    }

  - onChange: aCallback {
      set callbacks(Change) $aCallback
    }
  - onChange {} {
      return $callbacks(Change)
    }
  - valueChange {} {
      if [string length $callbacks(Change)] {
        catch {eval $callbacks(Change)}
      }
    }

  - dealloc {} {
      catch {destroy $label $image}
      return [$super dealloc]
    }
}



#-
#- end of gui/implementation/Choice.tcl
#-


#-
#- beginning of gui/interface/ComboBox.tcl
#-

@interface ComboBox : Box {
  {
    entry selection listbox pushButton
  }
  - selection ;
  - selection: aValue ;

  - showListbox ;
  - hideListbox ;
}


#-
#- end of gui/interface/ComboBox.tcl
#-


#-
#- beginning of gui/implementation/ComboBox.tcl
#-

@implementation ComboBox {
  - init {} {
      $super init 
      $self hideBorder
      $frame configure -bd 1 -relief sunken
      grid \
        [set entry [entry $frame.entry -relief flat]] \
        [set pushButton [button $frame.button -padx 0m -pady 0m \
          -command "$self toggleVisibilityOfListbox"]] \
        -sticky snew
      $frame.button configure -image [@image down_arrow] -width 12
      grid rowconfigure $frame 0 -weight 1
      grid columnconfigure $frame 0 -weight 1
      set listbox [[ListBox newWithPath: .listbox$self] height: 5]
      [$listbox path] configure -bd 2 -relief groove
      $listbox onSelection: "
        $self selection: \[\$self selection\]
        \$self hide
      "
    }
  - forwardInvocation: msg {
      if [$listbox respondsTo: [@selector $msg]] {
        set res [eval $listbox $msg]
        if {"$res" == "$listbox"} { return $self } else { return $res }
      } else {
        return [$super $msg]
      }
    }
  - toggleVisibilityOfListbox {} {
      if [winfo ismapped .listbox$self] {
        $self hideListbox
      } else {
        $self showListbox
      }
    }
  - showListbox {} {
      place .listbox$self -in $frame -x 0 -rely 1.0 -y +2 -relwidth 1 \
        -bordermode outside
      raise .listbox$self
    }
  - hideListbox {} {
      catch {place forget .listbox$self}
    }
  - selection {} {
      return [$entry get]
    }
  - selection: aValue {
      $entry delete 0 end
      $entry insert end $aValue
    }
  - dealloc {} {
      $listbox dealloc
      catch {destroy $entry $pushButton}
      return [$super dealloc]
    }
  - show {} {
      grid $path
    }
}


#-
#- end of gui/implementation/ComboBox.tcl
#-


#-
#- beginning of gui/interface/DynamicListBox.tcl
#-

@interface DynamicListBox : Box {
  {
    listbox loadButton addButton deleteButton addEntry
  }
  - add ;
  - delete ;
  - insert ;
  - remove ;
  - loadFromFile ;
}



#-
#- end of gui/interface/DynamicListBox.tcl
#-


#-
#- beginning of gui/implementation/DynamicListBox.tcl
#-

@implementation DynamicListBox : Box {
  - init {} {
      $super init
      grid rowconfigure $frame 0 -weight 1
      grid columnconfigure $frame 1 -weight 1
      set column 0
      foreach f {actions listbox} {
        grid [frame $frame.$f] \
          -row 0 -column $column -sticky snew
        incr column
      }
      #- actions frame
      foreach action {add delete load} {
        pack [set ${action}Button [button $frame.actions.button$action \
          -text $action -command "$self $action"]] \
          -padx 4 -pady 2 -anchor n -fill x
        bind $frame.actions.button$action <Return> { %W invoke }
        eval bind $frame.actions.button$action [@removeBindingsOnDestroyEvent]
      }
      $loadButton configure -text "load from file" -command "$self loadFromFile"
      focus $addButton

      #- listbox
      set listbox [[ListBox newWithPath: $frame.listbox.listbox] show]

      set addEntry [entry $path.entry -width 15]
      bind $addEntry <Return> \
        "place forget $addEntry ; \
        focus $addButton ; \
        $self add: \[%W get\]"
      eval bind $addEntry [@removeBindingsOnDestroyEvent]
    }
  - add {} { 
      $addEntry delete 0 end
      place $addEntry -relx 0.0 -rely 0.0 -relwidth 1 \
        -y -[winfo reqheight $addEntry] -in [$listbox path]
      focus -force $addEntry
    }
  - delete {} {
      $listbox removeIndex: [$listbox indexOfSelection]
    }
  - insert {} {
      $self add
    }
  - remove {} {
      $self delete 
    }
  - loadFromFile {} {
      $listbox loadFromFile: [tk_getOpenFile]
    }
  - forwardInvocation: msg {
      if [$listbox respondsTo: [@selector $msg]] {
        return [eval $listbox $msg]
      } else {
        return [eval $super $msg]
      }
    }
  - dealloc {} {
      $listbox dealloc
      catch {destroy $toplevel $path} 
      return [$super dealloc]
    }
}


#-
#- end of gui/implementation/DynamicListBox.tcl
#-


#-
#- beginning of gui/interface/ListBox.tcl
#-

@interface ListBox : Box {
  {
    listbox xscroll yscroll
  }
  - list ; 
  - list: aList ;

  - add: anItem ;
  - remove: anItem ;
  - contains: anItem ;

  - selection ;
  - selection: anItem ;

  - indexOfSelection: anIndex ;
  - indexOfSelection ;

  - onSelection: aCallback ;
  - onSelection ;

  - clear ;
  - sort ;

  - width: width ;
  - width ;

  - height: height ;
  - height ;

  - listbox ;

  - loadFromFile: aFile ;
}


#-
#- end of gui/interface/ListBox.tcl
#-


#-
#- beginning of gui/implementation/ListBox.tcl
#-

@implementation ListBox {
  - init {} {
      $super init
      $self hideBorder
      set bg [$frame cget -bg]
      $frame configure -relief sunken -bd 1
      set listbox [listbox $frame.listbox -exportselection no -bd 1 \
        -relief flat]
      set xscroll [scrollbar $frame.xscroll -orient horizontal \
        -command "$listbox xview"]
      set yscroll [scrollbar $frame.yscroll -orient vertical \
        -command "$listbox yview"]

      $listbox configure -xscrollcommand "$xscroll set" \
        -yscrollcommand "$yscroll set"
      grid rowconfigure $frame 0 -weight 1 
      grid columnconfigure $frame 0 -weight 1
      grid $listbox $frame.yscroll -sticky snew -padx 0 -pady 0
      grid $frame.xscroll -sticky snew -padx 0 -pady 0
      set font [$listbox cget -font]
      if {[set index [lsearch -exact $font bold]] != -1} {
        $listbox configure -font [lreplace $font $index $index]
      }
      bind $listbox <ButtonRelease-1> "$self selectionChange"
      set callbacks(Selection) {}
    }

  - list {} {
      return [$listbox get 0 end]
    }

  - list: aList {
      foreach item $aList { 
        $listbox insert end $item
      }
    }

  - add: anItem {
      $listbox insert end $anItem
      $self update
    }
  - remove: anItem {
      if {[set i [lsearch -exact [$listbox get 0 end] $anItem]] == -1} {
        return
      }
      $listbox delete $i
      $self update
    }
  - removeIndex: anIndex {
      catch {$listbox delete $anIndex}
    }
  - contains: anItem {
      set aList [$listbox get 0 end]
      return [aList contains: $anItem]
    }

  - selection {} {
      if {![string length [set i [$self indexOfSelection]]]} { return }
      return [$listbox get $i]
    }
  - selection: anItem {
      if {[set i [lsearch -exact [$listbox get 0 end] $anItem]] == -1} {
        return
      }
      $listbox selection set $i
    }
  - clearSelection {} {
      $listbox selection clear 0 end
    }

  - indexOfSelection {} {
      return [$listbox curselection]
    }
  - indexOfSelection: anIndex {
      $listbox selection set $anIndex
    }

  - onSelection: aCallback {
      set callbacks(Selection) $aCallback
    }
  - onSelection {} {
      return $callbacks(Selection)
    }
  - selectionChange {} {
      catch {eval $callbacks(Selection)}
    }

  - clear {} {
      $listbox delete 0 end
    }
  - sort {} {
      set myList [$self list]
      myList sort
      $self list: $myList
    }

  - width: width {
      $listbox configure -width $width
    }
  - width {} {
      return [$listbox cget -width]
    }
  - height: height {
      $listbox configure -height $height
    }
  - height {} {
      return [$listbox cget -height]
    }

  - listbox {} {
      return $listbox
    }

  - loadFromFile: aFile {
      if [catch {set f [open $aFile r]} err] {
        return [error "$self loadFromFile: $aFile ($err)"]
      }
      while {[eof $f] == 0} {
        gets $f line
        $self add: $line
      }
      close $f
      $self update
    }

  - dealloc {} {
      catch {destroy $listbox}
      return [$super dealloc]
    }

  - update {} {
      eval [bind $listbox <Configure>]
    }
}


#-
#- end of gui/implementation/ListBox.tcl
#-


#-
#- beginning of gui/interface/Notebook.tcl
#-

@interface Notebook : Widget {
  { 
    views selection counter main bar shadow
  }

  - views: aList ;
  - views ;

  - add: aView ;
  - remove: aView ;

  - selection ;
  - selection: aView ;
  
  - contains: aView ;

  - frameOf: aView ;
}


#-
#- end of gui/interface/Notebook.tcl
#-


#-
#- beginning of gui/implementation/Notebook.tcl
#-

@implementation Notebook {
  - init {} {
      $super init
      $path configure -width 100 -height 100 -bd 2
      set counter 1
      label $path.test -text foo -bd 1 -relief raised -padx 3m -pady 1m
      set height [expr [winfo reqheight $path.test] -1]
      destroy $path.test

      set main [frame $path.main -height [expr $height +10] \
        -width 10 -relief raised -bd 1]
      set bar [frame $path.bar -height [expr $height + 4] -bd 0]
      set shadow [frame $path.shadow -height 2 -bd 0]

      place $bar -relx 0.0 -rely 0.0 -relwidth 1 -anchor nw
      pack propagate $bar 0
      place $main -x 0 -relwidth 1 -y $height \
        -relheight 1 -height -$height -anchor nw
      raise $main
    }

  - views: aList {
      foreach view $aList { $self add: $view }
      $self selection: [lindex $aList 0]
    }

  - views {} {
      return $views
    }

  - add: aView {
      if [$self contains: $aView] { return }
      pack [label $bar.$counter -text $aView -highlightthickness 0 \
        -relief raised -bd 1 -padx 3m -pady 1m] \
        -anchor s -side left -padx 1
      bind $bar.$counter <ButtonPress-1> "$self selection: \"$aView\"" 
      set f [frame $main.$counter -bd 10]
      incr counter
      lappend views $aView
      $self selection: $aView
      return $f
    }

  - remove: aView {
      if {![$self contains: $aView]} { return }
      views remove: $aView
      set index [$self indexOf: $aView]
      catch {destroy $bar.$index $main.$index}
      if {"$selection" == "$aView"} {
        $self selection: [lindex $views end] 
      }
    }

  - selection: aView {
      if {![$self contains: $aView]} { return }
      set index [$self indexOf: $aView]
      if [string length [set old [pack slaves $main]]] {
        set i [lindex [split $old .] end]
        $bar.$i configure -padx 3m
        pack configure $bar.$i -anchor s -padx 1
        pack forget $old
      }
      $bar.$index configure -padx 4m
      pack configure $bar.$index -anchor n -padx 0
      pack $main.$index -padx 1 -pady 1 -fill both -expand 1
      place $shadow -in $bar.$index -relwidth 1 \
        -relx 0.0 -rely 1.0 -bordermode inside
      place $shadow -y 0
      raise $main
      raise $main.$index
      update ; raise $shadow
      set selection $aView
    }

  - selection {} {
      return $selection
    }

  - contains: aView {
      if [views contains: $aView] {
        return 1
      } else {
        return 0
      }
    }

  - frameOf: aView {
      if [$self contains: $aView] {
        return $main.[$self indexOf: $aView]
      } else {
        return [error "$self doesn't contain \"$aView\""]
      }
    }
  - indexOf: aView {
      foreach view [pack slaves $bar] {
        if {"[$view cget -text]" == "$aView"} {
          return [lindex [split $view .] end]
        }
      }
      return 
    }

  - dealloc {} {
      catch {destroy $main $bar $shadow}
      return [$super dealloc]
    }
}


#-
#- end of gui/implementation/Notebook.tcl
#-


#-
#- beginning of gui/interface/ProgressBox.tcl
#-

@interface ProgressBox : Box {
  {
    canvas @items
    value @timer
  }

  - value: aValueFrom0To100 ;
  - value ;

  - reset ;

  - timer: aTimeValueInSecond ;
  - timer ;

  - timerStep: aTimeValueInMilliSeconds ;
  - timerStep ;

  - run ;
  - start ;
  - stop ;

  - onChange: aCallback ;
  - onEnd: aCallback ;

}


#-
#- end of gui/interface/ProgressBox.tcl
#-


#-
#- beginning of gui/implementation/ProgressBox.tcl
#-

@implementation ProgressBox {
  - init {} {
      $super init
      $super hideBorder
      set bg [[entry $frame.e] cget -selectbackground]
      set lightBg [$frame.e cget -background]
      destroy $frame.e
      set canvas [canvas $frame.bar -width 100 -height 20 -relief sunken -bd 1]
      $canvas configure -bg $lightBg
      set items(value) [$canvas create text 50 10 \
        -text 0 -font Fixed -anchor e -fill black -tag value]
      set items(unit) [$canvas create text 55 10 \
        -text % -font Fixed -anchor w -fill black]
      set items(bar) [$canvas create rectangle -100 0 0 20 \
        -fill $bg -outline $bg -tag bar]
      grid $canvas -sticky snew
      $canvas configure -scrollregion {0 0 100 20}
      set timer(value) 60
      set timer(step) 600
      set timer(id) 0
      set value 0
    }

  - value: aValueFrom0To100 {
      $canvas move $items(bar) [expr $aValueFrom0To100 - $value] 0
      $canvas itemconfigure $items(value) -text $aValueFrom0To100
      set value $aValueFrom0To100
      $canvas raise $items(value)
      $canvas raise $items(unit)
      if {[info exists callbacks(Change)] \
        && [string length $callbacks(Change)]} {
        catch {eval $callbacks(Change)}
      }
    }
  - value {} {
      return $value
    }

  - timerStep: aTimeValueInMs {
      if {$aTimeValueInMs == 0 } {
        $self stop 
      } else { 
        set timer(step) $aTimeValueInMs
        set timer(value) [expr $timer(step) * 100]
      }
    }

  - timerStep {} {
      return $timer(step)
    }

  - timer: aTimeValueInSec {
      if {$aTimeValueInSec == 0 } {
        $self stop
      } else {
        set timer(value) [expr $aTimeValueInSec * 1000]
        set timer(step) [expr $timer(value) / 100]
      }
    }

  - timer {} {
      return $timer(value)
    }

  - start {} {
      return [$self run]
    }

  - run {} {
      if {$value < 100} {
        $self value: [expr $value + 1]
        set timer(id) [after $timer(step) $self start]
      } else {
        set timer(id) 0
        if {[info exists callbacks(End)] \
          && [string length callbacks(End)]} {
          catch {eval $callbacks(End)}
        }
      }
    }

  - onChange: aCallback {
      set callbacks(Change) $aCallback
    }

  - onEnd: aCallback {
      set callbacks(End) $aCallback
    }

  - stop {} {
      if {$timer(id) != 0} {
        after cancel $timer(id)
        set timer(id) 0
      }

    }

  - reset {} {
      $self value: 0
    }
}


#-
#- end of gui/implementation/ProgressBox.tcl
#-


#-
#- beginning of gui/interface/RadioBox.tcl
#-

@interface RadioBox : Box {
  {
    selection itemsList 
    imageOn imageOff
  }
  - list ;
  - list: aList ;

  - add: anItem ;
  - remove: anItem ;
  - contains: anItem ;

  - selection ;
  - selection: anItem ;

  - indexOfSelection: anIndex ;
  - indexOfSelection ;

  - onSelection: aCallback ;
  - onSelection ;

  - clear ;
  - sort ;

  - loadFromFile: aFile ;
}




#-
#- end of gui/interface/RadioBox.tcl
#-


#-
#- beginning of gui/implementation/RadioBox.tcl
#-

@implementation RadioBox {
  - init {} {
      $super init
      $self hideBorder
      set imageOff [@imageCopy radioOff]
      set imageOn [@imageCopy radioOn]
      set callbacks(Selection) {}
    }
  - list {} {
      return $itemsList
    }
  - list: aList {
      foreach item $aList { $self add: $item }
    }
  - add: anItem {
      if [$self contains: $anItem] { return $self }
      lappend itemsList $anItem
      grid [label "$frame.l$anItem" -text $anItem -anchor e] \
        [label "$frame.i$anItem" -image $imageOff] -sticky ew
      bind "$frame.l$anItem" <ButtonPress-1> "$self selection: \"$anItem\""
      bind "$frame.i$anItem" <ButtonPress-1> "$self selection: \"$anItem\""
    }
  - remove: anItem {
      if {![$self contains: $anItem]} { return $self }
      itemsList remove: $anItem
      catch {destroy "$frame.i$anItem" "$frame.l$anItem"}
      if {"$selection" == "$anItem"} {
        set selection {}
        $self selection: [lindex $itemsList end]
      }
    }
  - contains: anItem {
      return [itemsList contains: $anItem]
    }

  - selection {} {
      return $selection
    }
  - selection: anItem {
      if {![$self contains: $anItem]} { return $self }
      if [string length $selection] {
        "$frame.i$selection" configure -image $imageOff
      }
      set selection $anItem
      "$frame.i$selection" configure -image $imageOn
      $self selectionChange
    }

  - indexOfSelection {} {
      return [lsearch $itemsList $selection]
    }
  - indexOfSelection: anIndex {
      $self selection: [lindex $itemsList $anIndex]
    }

  - onSelection: aCallback {
      set callbacks(Selection) $aCallback
    }
  - onSelection {} {
      return $callbacks(Selection)
    }
  - selectionChange {} {
      catch {eval $callbacks(Selection)}
    }

  - clear {} {
      foreach item [$self list] {
        $self remove: $item
      }
      set selection {}
    }
  - sort {} {
      foreach item $itemsList {
        grid forget $frame.l$item $frame.i$item
      }  
      itemsList sort
      foreach item $itemsList {
        grid $frame.l$item $frame.i$item -sticky ew
      }
    }

  - loadFromFile: aFile {
      if [catch {set f [open $aFile r]} err] {
        return -code error "$self loadFromFile: $aFile ($err)"
      }
      while {[eof $f] == 0} {
        gets $f line
        if [string length $line] { $self add: $line }
      }
      close $f
    }

  - dealloc {} {
      foreach item $itemsList {
        bind "$frame.l$item" <ButtonPress-1> ""
        bind "$frame.i$item" <ButtonPress-1> ""
        catch {destroy "$frame.l$item" "$frame.i$item" }
      }
      return [$super dealloc]
    }
}



#-
#- end of gui/implementation/RadioBox.tcl
#-


#-
#- beginning of gui/interface/ScrollableBox.tcl
#-

@interface ScrollableBox : Box {
  {
    xscroll yscroll innerFrame
  }
}


#-
#- end of gui/interface/ScrollableBox.tcl
#-


#-
#- beginning of gui/implementation/ScrollableBox.tcl
#-

@implementation ScrollableBox {
  - init {} {
      $super init
      $super hideBorder

      $frame configure -relief sunken -bd 1

      set canvas [canvas $frame.canvas -width 1 -height 1 -relief flat -bd 1]
      set xscroll [scrollbar $frame.xscroll -orient horizontal \
        -command "$frame.canvas xview"]
      set yscroll [scrollbar $frame.yscroll -orient vertical \
        -command "$frame.canvas yview"]
      $frame.canvas configure -xscrollcommand "$xscroll set" \
        -yscrollcommand "$yscroll set"

      grid rowconfigure $frame 0 -weight 1
      grid columnconfigure $frame 0 -weight 1
      grid $frame.canvas $yscroll -sticky snew -padx 0 -pady 0
      grid $xscroll -sticky snew -padx 0 -pady 0

      set i [$frame.canvas create window 0 0 \
        -window [frame $frame.canvas.frame] \
        -anchor nw]
      bind $frame.canvas <Configure> {
        %W configure -scrollregion [list 0 0 \
          [winfo reqwidth %W.frame] [winfo reqheight %W.frame]]
      }
      bind $frame.canvas.frame <Configure> {
        [winfo parent %W] configure -scrollregion [list 0 0 \
          [winfo reqwidth %W] [winfo reqheight %W]]
      }
      eval bind $frame.canvas [@removeBindingsOnDestroyEvent]
      set innerFrame $frame.canvas.frame
    }

  - frame {} {
      return $innerFrame
    }
}


#-
#- end of gui/implementation/ScrollableBox.tcl
#-


#-
#- beginning of gui/interface/SplitBox.tcl
#-

@interface SplitBox : Box {
  {
    top bottom left right
    separator bullet 
    orientation 
    split lastSplit
    highlightBackground
  }

  - split: fraction ;
  - split ;

  - orientation: horizontalOrVertical ;
  - orientation ;

  - top ;
  - bottom ;
  - left ;
  - right ;
}


#-
#- end of gui/interface/SplitBox.tcl
#-


#-
#- beginning of gui/implementation/SplitBox.tcl
#-

@implementation SplitBox {
  - init {} {
      $super init
      $super hideBorder
      set top [set left [frame $frame.frame1 -bd 0 -relief flat]]
      set bottom [set right [frame $frame.frame2 -bd 0 -relief flat]]
      set separator [frame $frame.separator -relief flat -bd 0]
      set split 0.5
      set orientation horizontal
      set bullet [label $frame.separator.bullet -image [@image bullet] \
        -padx 0 -pady 0 -bd 0]

      set e [entry $frame._entry]
      set highlightBackground [$e cget -selectbackground]
      destroy $e

      bind $frame.separator <B1-Motion> "$self slideFromX: %X andY: %Y"
      bind $frame.separator.bullet <B1-Motion> "$self slideFromX: %X andY: %Y"
      bind $frame.separator <ButtonPress-1> "$self highlight"
      bind $frame.separator <ButtonRelease-1> "$self unHighlight"
      bind $frame.separator.bullet <ButtonPress-1> "$self highlight"
      bind $frame.separator.bullet <ButtonRelease-1> "$self unHighlight"

      bind $frame.separator <Double-ButtonPress-1> "$self toggle"
      bind $frame.separator.bullet <Double-ButtonPress-1> "$self toggle"
      

      eval bind $frame.separator [@removeBindingsOnDestroyEvent]
      eval bind $frame.separator.bullet [@removeBindingsOnDestroyEvent]
      $self rebuild
    }

  - rebuild {} {
      if {"$orientation" == "horizontal"} {
        place $top -relx 0.0 -rely 0.0 -relheight $split \
          -height -2 -relwidth 1 -width 0 \
          -anchor nw -y 0
        place $bottom -relx 0.0 -rely 1.0 \
          -relheight [expr 1 - $split] -height -7 -width 0 \
          -relwidth 1 -anchor sw -y 0
        place $separator -relx 0.0 -rely $split \
          -relwidth 1 -height 7 -width 0 -x 0 -y -1 \
          -relheight {} -anchor nw
        place $bullet -relx 0.5 -rely 0 -y 0 -bordermode outside
      } else {
        place $left -relx 0.0 -rely 0.0 -relwidth $split \
          -width -2 -relheight 1 -anchor nw -height 0
        place $right -relx 1.0 -rely 0.0 -relwidth [expr 1 - $split] \
          -width -7 -height 0 \
          -relheight 1 -anchor ne 
        place $separator -relx $split -rely 0.0 -width 7 \
          -relheight 1 -height 0 -x 0 -y 0 \
          -relwidth {} -anchor nw
        place $bullet -rely 0.5 -relx 0 -x 0 -bordermode outside
      }
    }

  - split: fraction {
      if {"$fraction" != "0.98"} { set lastSplit $fraction } else { lappend lastSplit 0.98 }
      set split [format %.2f $fraction]
      if {"$orientation" == "horizontal"} {
        place $top -relheight $split
        place $bottom -relheight [expr {1.0 - $split}]
        place $separator -rely $split
      } else {
        place $left -relwidth $split
        place $right -relwidth [expr {1.0 - $split}]
        place $separator -relx $split
      }
    }
  - split {} {
      return $split
    }

  - maxSplit {} {
      $self split: 0.98
    }

  - lastSplit {} {
      set lastSplit [lindex $lastSplit 0]
      $self split: $lastSplit
    }
 
  - toggle {} {
      if {[llength $lastSplit] > 1} { $self lastSplit } else { $self maxSplit }
    }

  - highlight {} {
      $separator configure -bg $highlightBackground
      $bullet configure -bg $highlightBackground
    }
  - unHighlight {} {
      $separator configure -bg [$frame cget -bg]
      $bullet configure -bg [$frame cget -bg]
    }

  - slideFromX:andY: {x y} {
      if {"$orientation" == "horizontal"} {
        set i [expr ($y - [winfo rooty $frame])/double([winfo height $frame])]
      } else {
        set i [expr ($x - [winfo rootx $frame])/double([winfo width $frame])]
      } 
      $self split: $i
    }

  - orientation {} { return $orientation }
  - orientation: horizontalOrVertical {
      set orientation $horizontalOrVertical
      $self rebuild
    }

  - top {} {
      return $top
    }
  - bottom {} {
      return $bottom
    }
  - left {} {
      return $left
    }
  - right {} {
      return $right
    }
}



#-
#- end of gui/implementation/SplitBox.tcl
#-


#-
#- beginning of gui/interface/SwitchBox.tcl
#-

@interface SwitchBox : Box {
  {
    fieldValueList imageOff imageOn
  }
  - list ;
  - list: aFieldValueList ;

  - valueOf: anItem ;
  - valueOf: anItem to: boolean ;

  - switchOn: anItem ;
  - switchOff: anItem ;
  - toggle: anItem ;

  - onChange: aCallback ;
  - onChange ;

  - add: anItem ;
  - add: anItem withValue: boolean ;

  - remove: anItem ;
  - contains: anItem ;

  - clear ;
  - sort ;

  - loadFromFile: aFile ;
}



#-
#- end of gui/interface/SwitchBox.tcl
#-


#-
#- beginning of gui/implementation/SwitchBox.tcl
#-

@implementation SwitchBox {
  - init {} {
      $super init
      $self hideBorder
      set imageOff [@imageCopy switchOff]
      set imageOn [@imageCopy switchOn]
      set callbacks(Change) {}
    }

  - items {} {
      return [fieldValueList fields]      
    }
  - items: anItemsList {
      $self clear
      foreach item $anItemsList {
        $self add: $item
      }
    }

  - values {} {
      return [fieldValueList values]      
    }
  - values: aValuesList {
      foreach value $aValuesList field [$self items] {
        $self valueOf: $field to: $value
      }
    }

  - list {} {
      return $fieldValueList
    }
  - list: aFieldValueList {
      foreach {field value} $aFieldValueList {
        $self add: $field withValue: $value
      }
    }

  - valueOf: anItem {
      return [fieldValueList get: $anItem]
    }
  - valueOf:to: {anItem boolean} {
      switch $boolean {
        0 { set img $imageOff }
        1 { set img $imageOn }
        default { return -code error "unknown boolean ($boolean)" }
      }
      fieldValueList set: $anItem to: $boolean
      "$frame.i$anItem" configure -image $img
      $self onChange
    }

  - switchOn: anItem {
      $self valueOf: $anItem to: 1
    }
  - switchOff: anItem {
      $self valueOf: $anItem to: 0
    }

  - toggle: anItem {
      if [fieldValueList get: $anItem] {
        $self switchOff: $anItem
      } else {
        $self switchOn: $anItem
      }
    }

  - add: anItem {
      $self add: $anItem withValue: 0
    }
  - add:withValue: {anItem boolean} {
      if [$self contains: $anItem] { return $self }
      fieldValueList add: [list $anItem 0]
      grid [label "$frame.l$anItem" -text $anItem -anchor e] \
        [label "$frame.i$anItem" -image $imageOff -bd 1 -relief raised] \
        -sticky ew
      bind "$frame.l$anItem" <ButtonPress-1> "$self toggle: \"$anItem\""
      bind "$frame.i$anItem" <ButtonPress-1> "$self toggle: \"$anItem\""
      bind "$frame.l$anItem" <ButtonRelease-1> "$self valueChangeFor: \"$anItem\""
      bind "$frame.i$anItem" <ButtonRelease-1> "$self valueChangeFor: \"$anItem\""
      eval bind \"$frame.l$anItem\" [@removeBindingsOnDestroyEvent]
      eval bind \"$frame.i$anItem\" [@removeBindingsOnDestroyEvent]
      $self valueOf: $anItem to: $boolean
    }
  - remove: anItem {
      if {![$self contains: $anItem]} { return $self }
      fieldValueList remove: $anItem
      catch {destroy "$frame.l$anItem" "$frame.i$anItem"}
    }
  - contains: anItem {
      return [fieldValueList contains: $anItem]
    }

  - onChange: aCallback {
      set callbacks(Change) $aCallback
    }
  - onChange {} {
      return $callbacks(Change)
    }
  - valueChangeFor: selection {
      if [string length $callbacks(Change)] {
        catch {eval $callbacks(Change)}
      }
    }

  - clear {} {
      foreach {field value} $fieldValueList {
        $self remove: $field
      }
    }
  - sort {} {
      foreach field [set fields [$self items]] {
        grid forget "$frame.l$field" "$frame.i$field"
      }
      fieldValueList sortByField
      foreach {field value} $fieldValueList {
        grid "$frame.l$field" "$frame.i$field" -sticky ew
      }
    }

  - loadFromFile: aFile {
      if [catch {set f [open $aFile r]} err] {
        return -code error "$self loadFromFile: $aFile ($err)"
      }
      while {[eof $f] == 0} {
        gets $f line
        if [string length $line] {
          if [regexp "," $line] {
            set field [lindex [split $line ,] 0]
            set value [lindex [split $line ,] 1]
            $self add: $field withValue: $value
          } else {
            $self add: $line 
          }
        }
      }
      close $f
    }

  - dealloc {} {
      foreach {field value} $fieldValueList {
        catch {destroy "$frame.l$field" "$frame.i$field"}
      }
      return [$super dealloc]
    }
}



#-
#- end of gui/implementation/SwitchBox.tcl
#-


#-
#- beginning of gui/interface/TableBox.tcl
#-

@interface TableBox : Box {
  {
    main headers rowOfSelection counter xscroll yscroll separator
  }
  - columns: columns ;
  - columns ;
  - addColumn: column ;

  - rowOfSelection: row ;
  - rowOfSelection ;
  - selection ;

  - onSelection: aCallback ;
  - onSelection ;

  - separator ;
  - separator: separator ;

  - add: record ;
  - insert: record atRow: row ;
  - deleteRow: row ;
  - content ;

  - loadFromFile: aFile ;
}


#-
#- end of gui/interface/TableBox.tcl
#-


#-
#- beginning of gui/implementation/TableBox.tcl
#-

@implementation TableBox {
  - init {} {
      $super init
      set counter 1
      set separator {}
      $super hideBorder
      $frame configure -relief sunken -bd 1
      frame $frame.f

      canvas $frame.f.headers -bd 0 -relief flat -width 10 -height 10 
      pack $frame.f.headers -fill both -expand 1
      $frame.f.headers create window 0 0 \
        -window [set headers [frame $frame.f.headers.frame -bd 0]] \
        -anchor nw -tag headers
      canvas $frame.f.main -bd 0 -relief flat -width 10 -height 10 
      $frame.f.main create window 0 0 \
        -window [set main [frame $frame.f.main.frame -bd 0]] \
        -anchor nw -tag main

      set xscroll [scrollbar $frame.xscroll -orient horizontal \
        -command "$self xview"]
      set yscroll [scrollbar $frame.yscroll -orient vertical \
        -command "$self yview"]

      $frame.f.main configure -xscrollcommand [list $xscroll set]

      grid $frame.f $frame.yscroll -sticky snew -padx 0 -pady 0
      grid $frame.xscroll -sticky snew -padx 0 -pady 0
      grid rowconfigure $frame 0 -weight 1
      grid columnconfigure $frame 0 -weight 1

      grid $frame.f.headers -sticky snew -padx 0 -pady 0
      grid $frame.f.main -sticky snew -padx 0 -pady 0
      grid rowconfigure $frame.f 1 -weight 1
      grid columnconfigure $frame.f 0 -weight 1

      set callbacks(Selection) {}

      $self on: Configure do: {
        set w [eventData get: path]
        if {"$w" == "$frame.f.main.frame"} {
            $frame.f.main configure -scrollregion [set bbox [$frame.f.main bbox all]]
            $frame.f.headers configure -scrollregion $bbox
        } elseif {"$w" == "$frame.f.headers.frame"} {
            $frame.f.headers configure \
              -height [winfo height $frame.f.headers.frame] 
        } elseif {"$w" == "$frame.headers"} {
            $frame.f.headers configure \
              -height [winfo height $frame.f.headers.frame] \
              -scrollregion [$frame.f.headers bbox all]
        } elseif {"$w" == "$frame.f.main"} {
            $frame.f.main itemconfigure [$frame.f.main find withtag main] \
              -height [winfo height $frame.f.main]
            $frame.f.main configure -scrollregion [set bbox [$frame.f.main bbox all]]
            $frame.f.headers configure -scrollregion $bbox
        }
      }
      
      $self enableEvent: Configure \
                     on: [list $frame.f.main.frame $frame.f.headers.frame \
                               $frame.f.main $frame.f.headers]

      $self on: ButtonPress-1 do: {
        $self rowOfSelection: [[eventData get: path] nearest [eventData get: y]]
        $self selectionChange
      }
      $self on: B1-Motion do: {
        $self rowOfSelection: [[eventData get: path] nearest [eventData get: y]]
        $self selectionChange
      }
      $self on: Button-4 do: {
        foreach column [pack slaves $main] {
          if {"[eventData get: path]" != "$column"} {
            $column yview scroll -5 units
          }
        }
      }
      $self on: Button-5 do: {
        foreach column [pack slaves $main] {
          if {"[eventData get: path]" != "$column"} {
            $column yview scroll +5 units
          }
        }
      }
    }

  - columns: columns {
      if [string length $separator] {
        foreach column [split $columns $separator] { $self addColumn: $column }
      } else {
        foreach column $columns { $self addColumn: $column }
      }
    }
  - columns {} {
      set columns {}
      foreach column [pack slaves $headers] {
        lappend columns [$column get 0 end]
      }
      if [string length $separator] {
        return [join $columns $separator]
      } else {
        return $columns
      }
    }
  - rowOfSelection: row {
      foreach column [pack slaves $main] {
        $column selection clear 0 end
        if [string length $row] {
          $column select anchor $row
          $column select set anchor $row
          set rowOfSelection $row
        }
      }
    }
  - rowOfSelection {} {
      return $rowOfSelection
    }
  - selection {} {
      return [$self row: $rowOfSelection]
    }
  - row: aRow {
      set result {}
      set columns [pack slaves $main]
      if {[llength $columns] == 0} { return }
      foreach column $columns {
        lappend result [$column get $i]
      }
      if [string length $separator] {
        return [join [string trimright $result \n] $separator]
      } else {
        return [string trimright $result \n]
      }
    }
  - addColumn: column {
      global tcl_platform

      set length [expr [string length $column] + 3]
      pack [listbox $headers.$counter -relief raised -bd 1 -width $length \
        -height 1 -exportselection no \
        -background [$frame.f.headers cget -bg] \
        -selectbackground [$frame.f.headers cget -bg]] \
        -side left -fill y -expand 1 -padx 0 -pady 0
      $headers.$counter insert end $column
      pack [listbox $main.$counter -width $length -height 5 \
        -relief raised -bd 1 \
        -exportselection no -yscrollcommand [list $yscroll set]] \
        -side left -fill y -expand 1 -padx 0 -pady 0

      $headers configure -width [winfo reqwidth [winfo parent $headers]]

      $self enableEvent: {ButtonPress-1 B1-Motion Button-4 Button-5} \
                     on: $main.$counter
      incr counter
    }

  - add: record {
      set columns [pack slaves $main]
      if {[string length $separator]} {
        if {[llength [split $record $separator]] > [llength $columns]} {
          set record [join [lrange [split $record $separator] 0 [expr [llength $columns] -1]] $separator]
        }
        set items [split $record $separator]
      } else {
        if {[llength $record] > [llength $columns]} {
          set record [join [lrange $record 0 [expr [llength $columns] -1]]]
        }
        set items $record
      }

      foreach column $columns item $items {
        if {[set len [expr [string length $item] +3]] \
          > [$column cget -width]} {
          $column configure -width $len
          $headers.[lindex [split $column .] end] configure -width $len
        }
        $column insert end $item
      }
    }
  - insert:atRow: {record row} {
      set columns [pack slaves $main]
      if {[string length $separator]} {
        if {[llength [split $record $separator] > [llength $columns]} {
          set record [join [lrange [split $record $separator] 0 [expr [llength $columns] -1]] $separator]
        }
      } else {
        if {[llength $record] > [llength $columns]} {
          set record [join [lrange $record 0 [expr [llength $columns] -1]]]
        }
      }

      foreach column $columns item [split $record $separator] {
        if {[set len [expr [string length $item] +3]] \
          > [$column cget -width]} {
          $column configure -width $len
          $headers.[lindex [split $column .] end] configure -width $len
        }
        $column insert $row $item
      }
    }
  - deleteRow: row {
      foreach column [pack slaves $main] {
        $column delete $row $row
      }
    }
  - loadFromFile: aFile {
      if [catch {set f [open $aFile r]} err] {
        @error "$self loadFromFile: $aFile ($err)"
        return -1
      }
      gets $f line
      if {[string length $separator]} {
        set len [llength [split $line $separator]]
      } else {
        set len [llength $line]
      }
      if {![string length [$self columns]]} {
        for {set i 1} {$i <= $len} {incr i} {
          $self addColumn: $i
        }
      }

      while {[eof $f] == 0} {
        gets $f line
        if [string length $line] { 
          $self add: $line 
        }
      }
      close $f
    }
  - content {} {
      set result {}
      set columns [pack slaves $main]
      if {[llength $columns] == 0} { return }
      for {set i 0} {$i < [[lindex $columns 0] size]} {incr i} {
        set line {}
        foreach column $columns {
          append line [$column get $i],
        }
        append result [string trimright $line ,]\n
      }
      return [string trimright $result \n]
    }
  - xview {args} {
      eval [winfo parent $headers] xview $args
      eval [winfo parent $main] xview $args
    }
  - yview {args} {
      foreach l [pack slaves $main] {
        eval $l yview $args
      }
    }
  - onSelection: aCallback {
      set callbacks(Selection) $aCallback
    }
  - onSelection {} {
      return $callbacks(Selection)
    }
  - selectionChange {} {
      catch {eval $callbacks(Selection)}
    }
  - separator {} { return $separator }
  - separator: _separator { set separator $_separator }
  - dealloc {} { 
      catch {destroy $main $headers}
      return [$super dealloc]
    }
}


#-
#- end of gui/implementation/TableBox.tcl
#-


#-
#- beginning of gui/interface/TextBox.tcl
#-

@interface TextBox : Box {
  - text: aText ;
  - text ;

  - insert: aText ;

  - isEditable ;
  - enable ; 
  - disable ;

  - loadFromFile: aFile ;
}


#-
#- end of gui/interface/TextBox.tcl
#-


#-
#- beginning of gui/implementation/TextBox.tcl
#-

@implementation TextBox {
  - init {} {
      $super init
      $self hideBorder
      $frame configure -relief sunken -bd 1
      text $frame.text -wrap none -bd 1 -relief flat
      scrollbar $frame.xscroll -orient horizontal \
        -command "$frame.text xview"
      scrollbar $frame.yscroll -orient vertical \
        -command "$frame.text yview"

      grid rowconfigure $frame 0 -weight 1
      grid columnconfigure $frame 0 -weight 1
      grid $frame.text $frame.yscroll -sticky snew -padx 0 -pady 0
      grid $frame.xscroll -sticky snew -padx 0 -pady 0
      $frame.text configure -xscrollcommand "$frame.xscroll set" \
        -yscrollcommand "$frame.yscroll set"
    }

  - text {} {
      return [$frame.text get 0.0 end]
    }
  - text: aText {
      $self clear
      $self insert: $aText  
    }

  - insert: aText {
      $frame.text insert end $aText
    }

  - isEditable {} {
      if {"[$frame.text cget -state]" == "normal"} {
        return 1
      } else {
        return 0
      }
    }
  - enable {} {
      $frame.text configure -state normal 
    }
  - disable {} {
      $frame.text configure -state disabled 
    }

  - clear {} {
      $frame.text delete 0.0 end
    }

  - loadFromFile: aFile {
      if [catch {set f [open $aFile r]} err] {
        return -code error "$self loadFromFile: $aFile ($err)"
      }
      $self text: [read $f]
      close $f
    }

  - dealloc {} {
      catcr {destroy $frame.text}
      return [$super dealloc]
    }

  - background: aColor {
      $frame.text configure -background $aColor
    }

  - width: aWidth {
      $frame.text configure -width $aWidth
    }

  - height: anHeight {
      $frame.text configure -height $anHeight
    }

  - disable {} {
      $frame.text configure -state disabled
    }
  - enable {} {
      $frame.text configure -state normal
    }
}


#-
#- end of gui/implementation/TextBox.tcl
#-


#-
#- beginning of gui/interface/TreeBox.tcl
#-

@interface TreeBox : Tree,CanvasBox {
  {
    tree selection 
    @icons @buildData
  }

  - selection: aNode ;
  - selection ;

  - onSelection: aCallback ;
  - onSelection ;

  - open: aNode ;
  - toggle: aNode ;
  - close: aNode ;

  - add: aNode ;
  - remove: aNode ;

  - openedIcon: aBitmap ;
  - closedIcon: aBitmap ;
  - childIcon: aBitmap ;

  - openedIcon ;
  - closedIcon ;
  - childIcon ;

  - openedIconOfNode: aNode to: aBitmap ;
  - closedIconOfNode: aNode to: aBitmap ;
  - childIconOfNode: aNode to: aBitmap ;
  - iconOfNode: aNode to: aBitmap ;

  - openedIconOfNode: aNode ;
  - closedIconOfNode: aNode ;
  - childIconOfNode: aNode ;

  - itemOfNode: aNode ;

  - onLeftClick: aCallback ;
  - onMiddleClick: aCallback ;
  - onRightClick: aCallback ;

}


#-
#- end of gui/interface/TreeBox.tcl
#-


#-
#- beginning of gui/implementation/TreeBox.tcl
#-

@implementation TreeBox {
  - init {} {
      $super@ Tree init
      $super@ CanvasBox init

      set defaults(font) [$label cget -font]
      set defaults(status) closed
      set defaults(type) child 
      set defaults(fill) #a0a0a0

      set bg [[entry $path._e] cget -selectbackground] 
      set defaults(selectionBackground) $bg
      destroy $path._e

      set icons(empty) [@imageCopy empty]
      set icons(plus) [@imageCopy plus]
      $icons(plus) configure -background [$canvas cget -bg]
      set icons(minus) [@imageCopy minus]
      $icons(minus) configure -background [$canvas cget -bg]

      set icons(child) [@imageCopy white_sphere]
      set icons(opened) [@imageCopy sphere]
      set icons(closed) [@imageCopy sphere]

      set buildData(lastY) 0
      set buildData(selectionItem) {}

      $self property: status of: $root to: $defaults(status)
      $self draw: $root

      set callbacks(Selection) {}
    }

  - root: rootNode {
      $super@ Tree root: $rootNode
      $self redraw
    }

  - toggle: aNode {
      set aNode [$self format: $aNode]
      if [$self contains: $aNode] {
        if [string length [$self childsOf: $aNode]] {
          if {"[$self property: status of: $aNode]" == "opened"} {
            $self property: status of: $aNode to: closed
          } else {
            $self property: status of: $aNode to: opened
          }
          $self redraw
        }
      }
    }

  - open: aNode {
      set aNode [$self format: $aNode]
      if [$self contains: $aNode] {
        $self property: status of: $aNode to: opened
        $self redraw
      }
    }

  - close: aNode {
      set aNode [$self format: $aNode]
      if [$self contains: $aNode] {
        $self property: status of: $aNode to: closed
        $self redraw
      }
    }

  - add: aNode {
      set aNode [$self format: $aNode]
      $super add: $aNode
      $self property: status of: $aNode to: opened
      $self redraw
    }

  - addWithoutUpdate: aNode {
      set aNode [$self format: $aNode]
      $super add: $aNode
      $self property: status of: $aNode to: opened
    }

  - remove: aNode {
      set aNode [$self format: $aNode]
      $super remove: $aNode
      foreach field [array names icons $aNode,*] {
        unset icons($field)
      }
      $self redraw
    }

  - redraw {} {
      #- clearing current canvas content
      foreach item [$canvas find all] {
        if [string length [set c [bind $item]]] {
          @debug "$c"
        }
        $canvas delete $item
      }
      set buildData(lastY) 0
      $self draw: [$self root]
      $canvas configure -scrollregion [$canvas bbox all]
      if [string length $selection] {
        $self updateSelection
      }
      $self zoomSync
    }

  - draw: aNode {
      if {"$aNode" == "[$self root]"} {
        $canvas create image 0 0 -image $icons(empty)
      }

      if [$self contains: $aNode] {
        #- setting x,y for the drawing
        set x0 [expr ([$self levelOf: $aNode] * 20) + 40]
        set y0 $buildData(lastY)
        set x1 $x0
        set y1 [expr $y0 + 20]
        set x2 [expr $x0 + 20]
        set y2 $y1
        set x3 [expr $x2 + 10]
        set y3 $y2

        #- expanding branch if previous node of the same level is opened
        if {"$aNode" != "[$self root]"} {
          set childrenOfParent [$self childsOf: [$self parentOf: $aNode]]
          set i [childrenOfParent indexOf: $aNode]
          if {$i > 0} {
            incr i -1
            set previousNode [lindex $childrenOfParent $i]
            if {"[$self property: status of: $previousNode]" == "opened"} {
              #- getting Y
              set item [$canvas find withtag "node:$previousNode"]
              set y [expr round([lindex [$canvas coord $item] end])]
              $canvas lower [$canvas create line \
                $x0 $y $x0 $buildData(lastY) -fill $defaults(fill)]
            }
          }
        }

        #- setting icon following the current type and status of the node
        set _childs [$self childsOf: $aNode]
        set parent [$self parentOf: $aNode]

        if {![string length $_childs]} {
          if [info exists icons($aNode,child)] {
            set icon $icons($aNode,child) 
          } else {
            set icon $icons(child)
          }
          set expandIcon $icons(empty)
        } elseif {"[$self property: status of: $aNode]" == "opened"} {
          if [info exists icons($aNode,opened)] {
            set icon $icons($aNode,opened)
          } else {
            set icon $icons(opened)
          }
          if [string length $_childs] {
            set expandIcon $icons(minus)
          } else {
            set expandIcon $icons(empty)
          }
        } else {
          if [info exists icons($aNode,closed)] {
            set icon icons($aNode,closed)
          } else {
            set icon $icons(closed) 
          }
          if [string length $_childs] {
            set expandIcon $icons(plus)
          } else {
            set expandIcon $icons(empty)
          }
        }

        set item1 [$canvas create poly $x0 $y0 $x1 $y1 $x2 $y2 $x1 $y1 \
          -outline $defaults(fill)]
        set item2 [$canvas create image $x1 $y2 -image $expandIcon \
          -tag [list icon "expandIconOf:$aNode"]]
        set item3 [$canvas create image $x2 $y2 -image $icon \
          -tag "iconOf:$aNode" \
          -anchor c]
        set item4 [$canvas create text $x3 $y3 \
          -text [$self nodeNameOf: $aNode] -font $defaults(font) \
          -anchor w -tag [list parent:$parent node:$aNode]]

        $canvas bind $item4 <Double-1> "$self toggle: \"$aNode\""
        $canvas bind $item4 <ButtonPress-1> "$self selection: \"$aNode\""
        if {[string length $_childs]} {
          foreach i {1 2 3} {
            $canvas bind [set item$i] <ButtonPress-1> "$self toggle: \"$aNode\""
          }
        }
        $canvas lower $item1
        $canvas raise $item2
        $canvas raise $item3

        set buildData(lastY) [expr $buildData(lastY) + 20]
        if {[string length $_childs] \
          && "[$self property: status of: $aNode]" == "opened"} {
          foreach child $_childs {
            $self draw: $child
          }
        }
      } 
    }

  - clear {} {
      foreach item [$canvas find all] { 
        $canvas delete $item 
      }
      $self selection: {}
      $self clear
    }

  - selection: aNode {
      set aNode [$self format: $aNode]
      #- removing root name if in the path
      if [regexp [join [list [$self root] (.*)] [$self separator]] $aNode -> node] {
        if {![string length $node]} {
          set aNode $node
        }
      }

      if {"$selection" == "$aNode"} { $self selection: {} ; return $self }
      if {[string length $aNode] && ![$self contains: $aNode]} {
        return [error "$self doesn't contain \"$aNode\""]
      }
      #- clearing current selection
      if [string length [set i $buildData(selectionItem)]] {
        $canvas delete $i
        set buildData(selectionItem) {}
      } 
      set previousSelectionNode $selection
      set selection {}
      if {![string length $aNode]} { return $self }
      set selection $aNode
      $self updateSelection
      $self selectionChange
    }

  - updateSelection {} {
      if {![string length $selection]} { return $self }
      set bbox [$canvas bbox [$canvas find withtag node:$selection]]
       if {[llength $bbox] == 4} {
        foreach {x1 y1 x2 y2} $bbox {
          set i [$canvas create rectangle $x1 $y1 $x2 $y2 \
            -fill $defaults(selectionBackground) \
            -outline $defaults(selectionBackground)]
          #set i [$canvas create rectangle $x1 $y1 $x2 $y2 \
          #  -fill $defaults(selectionBackground) \
          #  -outline black]
          $canvas lower $i
          set buildData(selectionItem) $i
        }
      }
    }

  - selection {} {
      return $selection
    }

  - onSelection: aCallback {
       set callbacks(Selection) $aCallback
    }

  - onSelection {} {
      return $callbacks(Selection)
    }

  - selectionChange {} {
      catch {eval $callbacks(Selection)}
    }

  - itemOfNode: aNode {
      set aNode [$self format: $aNode]
      return [$canvas find withtag node:$aNode]
    }

  - openedIcon: aBitmap {
      set aNode [$self format: $aNode]
      set icons(opened) $aBitmap
      $self redraw
    }

  - closedIcon: aBitmap {
      set icons(closed) $aBitmap
      $self redraw
    }

  - childIcon: aBitmap {
      set icons(child) $aBitmap
      $self redraw
    }

  - openedIcon {} {
      return $icons(opened)
    }

  - closedIcon {} {
      return $icons(closed)
    }

  - childIcon {} {
      return $icons(child)
    }

  - openedIconOfNode:to: {aNode aBitmap} {
      set aNode [$self format: $aNode]
      set icons($aNode,opened) $aBitmap
      $self redraw
    }

  - closedIconOfNode:to: {aNode aBitmap} {
      set aNode [$self format: $aNode]
      set icons($aNode,closed) $aBitmap
      $self redraw
    }

  - childIconOfNode:to: {aNode aBitmap} {
      set aNode [$self format: $aNode]
      set icons($aNode,child) $aBitmap
      $self redraw
    }

  - iconOfNode:to: {aNode aBitmap} {
      set aNode [$self format: $aNode]
      set icons($aNode,opened) $aBitmap
      set icons($aNode,closed) $aBitmap
      set icons($aNode,child) $aBitmap
      $self redraw
    }

  - openedIconOfNode: aNode {
      set aNode [$self format: $aNode]
      if [info exists icons($aNode,opened)] {
        return $icons($aNode,opened)
      } else {
        return $icons(opened)
      }
    }

  - closedIconOfNode: aNode {
      set aNode [$self format: $aNode]
      if [info exists icons($aNode,closed)] {
        return $icons($aNode,closed)
      } else {
        return $icons(closed)
      }
    }

  - childIconOfNode: aNode {
      set aNode [$self format: $aNode]
      if [info exists icons($aNode,child)] {
        return $icons($aNode,child)
      } else {
        return $icons(child)
      }
    }

  - saveToFile: aFile {
      set f [open $aFile w]
      foreach node [lsort [$self allChilds]] {
        puts $f "$node"
      }
      close $f
    }

  - onLeftClick: aCallback {
      bind $canvas <ButtonRelease-1> "$self doLeftClick"
      set callbacks(LeftClick) $aCallback
    }

  - doLeftClick {} {
      catch {eval $callbacks(LeftClick)}
    } 

  - onMiddleClick: aCallback {
      bind $canvas <ButtonRelease-2> "$self doMiddleClick"
      set callbacks(MiddleClick) $aCallback
    }
  - doMiddleClick {} {
      catch {eval $callbacks(MiddleClick)} 
    } 

  - onRightClick: aCallback {
      bind $canvas <ButtonRelease-3> "$self doRightClick"
      set callbacks(RightClick) $aCallback
    }

  - doRightClick {} {
      catch {eval $callbacks(RightClick)}
    } 

  - add: aNode {
      set aNode [$self format: $aNode]
      $super add: $aNode
      $self property: status of: $aNode to: opened
      $self redraw
    }

  - seeEnd {} {
      $canvas yview moveto [lindex [$canvas yview] end]
      $canvas xview moveto [lindex [$canvas xview] end]
    }

  - forwardInvocation: anInvocation {
      if [$self respondsTo: [@selector $anInvocation]] {
        if [catch {set result [eval $self $anInvocation]} err] {
          return [error $err]
        }
        $self redraw
        if {"$result" != "$self"} {
          return $result
        }
      } else {
        return [eval $super $anInvocation]
      }
    }

  - dealloc {} {
      $self release
      return [$super dealloc]
    }

  - loadFromFile: aFile {
      $super@ Tree loadFromFile: $aFile
      foreach child [$self allChilds] {
        $self property: status of: $child to: closed
      }
      $self redraw
    }
 
  - loadFromDirectory: aDirectory {
      $super@ Tree loadFromDirectory: $aDirectory
      foreach child [$self allChilds] {
        $self property: status of: $child to: opened
      }
      $self redraw
    }

  - canvasBackground: aColor {
      $canvas configure -background $aColor
      $icons(plus) configure -background $aColor
      $icons(minus) configure -background $aColor
    }

}


#-
#- end of gui/implementation/TreeBox.tcl
#-

